// acis.cxx: ACIS class implementation.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: acis.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (acis, "$Id: acis.cc 1 2005-08-10 23:35:07Z frank $");

#ifdef HAVE_ACIS_5
#include <mmgrhusk/freelist.hxx>
#endif // HAVE_ACIS_5

#ifdef HAVE_ACIS_6
#include <baseutil/mmgr/freelist.hxx>
#endif // HAVE_ACIS_6

#include "stephusk/util/apiinit.hxx"
#include "stephusk/stepwrit/api/apiwrite.hxx"
#include "stephusk/stepread/api/apiread.hxx"

// Workaround for identifier conflict in the ACIS IGES husk.
#define List ACIS_List
#include <igeshusk/igs_api/routines.hxx>
#undef List

#include "interface.hxx"

#include "acis/acis.hxx"
#include "acis/silapi.hxx"
#include "acis/acis_module.hxx"
#include "acis/light.hxx"
#include "acis/at_material.hxx"

SIL_MODULE ()
{
  return new ACIS;
}

const char *ACIS::nm = "acis";

const char *ACIS::fl_formats[9] = { 
  "iges", "ACIS IGES Format",
  "step", "ACIS STEP Format",
  "sat_bin", "ACIS bin SAT Format",
  "sat", "ACIS text SAT Format",
  NULL
};

ACIS::ACIS ()
{
  // Modeller...
  init_freelist_size = check_free_lists();
  sil_acis_check (api_start_modeller (0),
		  "acis constructor", "start modeller");

  // Initialize ACIS components
  sil_acis_check (api_initialize_kernel(),
		  "acis constructor", "initialize kernel");
  sil_acis_check (api_initialize_spline(),
		  "acis constructor", "initialize spline");
  sil_acis_check (api_initialize_generic_attributes(),
		  "acis constructor", "initialize generic attributes");
  sil_acis_check (api_initialize_intersectors(),
		  "acis constructor", "initialize intersectors");
  sil_acis_check (api_initialize_faceter(),
		  "acis constructor", "initialize faceter");
  sil_acis_check (api_initialize_constructors(),
		  "acis constructor", "initialize constructors");
  sil_acis_check (api_initialize_euler_ops(),
		  "acis constructor", "initialize euler ops");
  sil_acis_check (api_initialize_booleans(),
		  "acis constructor", "initialize booleans");

  api_logging (FALSE);

  set_api_checking (TRUE);

  max_ll = 5.0;
}

ACIS::~ACIS()
{
  DELTA_STATE *ds = NULL;
  api_note_state (ds);
  if (ds)
    api_delete_ds (ds);

  // Terminate ACIS components
  sil_acis_check (api_terminate_booleans(),
		  "acis destructor", "terminate booleans");
  sil_acis_check (api_terminate_euler_ops(),
		  "acis destructor", "terminate euler ops");
  sil_acis_check (api_terminate_constructors(),
		  "acis destructor", "terminate constructors");
  sil_acis_check (api_terminate_faceter(),
		  "acis destructor", "terminate faceter");
  sil_acis_check (api_terminate_intersectors(),
		  "acis destructor", "terminate intersectors");
  sil_acis_check (api_terminate_generic_attributes(),
		  "acis destructor", "terminate generic attributes");
  sil_acis_check (api_terminate_spline(),
		  "acis destructor", "terminate spline");
  sil_acis_check (api_terminate_kernel(),
		  "acis destructor", "initialize kernel");

 // Shut down ACIS
  sil_acis_check (api_stop_modeller(),
		  "acis destructor", "stop modeller");
  int current_freelist_size = check_free_lists ();
  if (current_freelist_size != init_freelist_size) {
    sil_msg_error ("acis destructor", "possible memory leak");
    char str[255];
    snprintf (str, 255, "initial free lists: %d", init_freelist_size);
    sil_msg_error ("acis destructor", str);
    snprintf (str, 255, "free lists after stopping modeller: %d",
	      current_freelist_size);
    sil_msg_error ("acis destructor", str);
  }
}

void ACIS::add_symbol (void *object, char *name, bool visible)
{
  API_BEGIN;
  ENTITY *e = (ENTITY *)object;
  ATTRIB_GEN_NAME *a;
  sil_acis_check (api_find_named_attribute (e,
					    "SIL_NAME", a),
		  "acis add_symbol", "find named attribute SIL_NAME");
  while (a) {
    a->unhook ();
    sil_acis_check (api_del_entity (a),
		    "acis add_symbol", "delete SIL_NAME attribute");
    sil_acis_check (api_find_named_attribute (e,
					      "SIL_NAME", a),
		    "acis add_symbol", "find named attribute SIL_NAME");
  }
  sil_acis_check (api_add_generic_named_attribute (e,
						   "SIL_NAME",
						   name, SplitLose,
						   MergeKeepOne,
						   TransIgnore),
		  "acis add_symbol", "add generic named attribute SIL_NAME");
  sil_acis_check (api_find_named_attribute (e,
					    "SIL_VISIBLE", a),
		  "acis add_symbol", "find named attribute SIL_VISIBLE");
  while (a) {
    a->unhook ();
    sil_acis_check (api_del_entity (a),
		    "acis add_symbol", "delete SIL_VISIBLE attribute");
    sil_acis_check (api_find_named_attribute (e,
					      "SIL_VISIBLE", a),
		    "acis add_symbol", "find named attribute SIL_NAME");
  }
  if (visible) {
    sil_acis_check(api_add_generic_named_attribute (e, 
						    "SIL_VISIBLE",
						    SplitKeep, 
						    MergeKeepOne, 
						    TransIgnore),
		   "acis add_symbol",
		   "add generic named attribute SIL_VISIBLE");
  }
  API_END;
  return;
}

void ACIS::show (void *object)
{
  ENTITY *e = (ENTITY *)object;
  sil_acis_check (api_add_generic_named_attribute (e, 
						   "SIL_VISIBLE", SplitKeep, 
						   MergeKeepOne, 
						   TransIgnore),
		  "acis show", "add generic named attribute SIL_VISIBLE");
  return;
}

void ACIS::hide (void *object)
{
  API_BEGIN;
  ENTITY *e = (ENTITY *)object;
  ATTRIB_GEN_NAME *a;
  sil_acis_check (api_find_named_attribute (e,
					    "SIL_VISIBLE", a),
		  "acis hide", "find named attribute SIL_VISIBLE");
  while (a) {
    a->unhook ();
    sil_acis_check (api_del_entity (a),
		    "acis hide", "delete SIL_VISIBLE attribute");
    sil_acis_check (api_find_named_attribute (e,
					      "SIL_VISIBLE", a),
		    "acis hide", "find named attribute SIL_NAME");
  }
  API_END;
  return;
}

void ACIS::facet_init ()
{

  API_BEGIN;

  MESH_MANAGER *MM = NULL;
  sil_acis_check (api_get_mesh_manager (MM),
		  "acis facet_init", "get mesh manager");
  if (MM)
    delete MM;
  INDEXED_MESH_MANAGER *mm = new INDEXED_MESH_MANAGER;
  mm->SetTransform (NULL);
  sil_acis_check (api_set_mesh_manager (mm),
		  "acis facet_init", "set mesh manager");

  double surf_tol = 0.0;
  double normal_tol = 0.0;
  SiL_TriangMode t_mode = SIL_TRIANG_NONE;
  SiL_GridMode g_mode = SIL_GRID_NONE;
  int max_l = 0;
  int min_l = 0;
  double ar = 0.0;
  SiL_AdjustMode am = SIL_ADJUST_NON_GRID;

  sil_get_refinement (surf_tol, normal_tol, t_mode, g_mode, max_l, min_l,
  		      max_ll, ar, am);
  
  REFINEMENT *ref = NULL;
  sil_acis_check (api_get_default_refinement (ref),
		  "acis facet_init", "get default refinement");
  if (ref)
    sil_acis_check (api_del_entity (ref),
		    "acis facet_init", "delete refinement");
  ref = new REFINEMENT ();

  ref->set_surface_tol (static_cast<REFINEMENT_FLOAT> (surf_tol));
  ref->set_normal_tol (static_cast<REFINEMENT_FLOAT> (normal_tol));
  ref->set_grid_aspect_ratio (static_cast<REFINEMENT_FLOAT> (ar));
  ref->set_max_edge_length (static_cast<REFINEMENT_FLOAT> (max_ll));
  ref->set_min_v_grid_lines (min_l);
  ref->set_min_u_grid_lines (min_l);
  ref->set_max_grid_lines (max_l);
    
  switch (t_mode) {
  case SIL_TRIANG_NONE:
    ref->set_triang_mode (AF_TRIANG_NONE);
    break;
  case SIL_TRIANG_FRINGE_2:
    ref->set_triang_mode (AF_TRIANG_FRINGE_2);
    break;
  case SIL_TRIANG_FRINGE_3:
    ref->set_triang_mode (AF_TRIANG_FRINGE_3);
    break;
  case SIL_TRIANG_FRINGE_4:
    ref->set_triang_mode (AF_TRIANG_FRINGE_4);
    break;
  case SIL_TRIANG_ALL:
    ref->set_triang_mode (AF_TRIANG_ALL);
    break;
  default:
    ref->set_triang_mode (AF_TRIANG_FRINGE_1);
    break;
  }

  switch (g_mode) {
  case SIL_GRID_NONE:
    ref->set_grid_mode (AF_GRID_NONE);
    break;
  case SIL_GRID_TO_EDGES:
    ref->set_grid_mode (AF_GRID_TO_EDGES);
    break;
  default:
    ref->set_grid_mode (AF_GRID_INTERIOR);
    break;
  }
  
  switch (am) {
  case SIL_ADJUST_NON_GRID:
    ref->set_adjust_mode (AF_ADJUST_NON_GRID);
    break;
  case SIL_ADJUST_ALL:
    ref->set_adjust_mode (AF_ADJUST_ALL);
    break;
  default:
    ref->set_adjust_mode (AF_ADJUST_NONE);
    break;
  }
    
  ref->set_surf_mode (AF_SURF_ALL);

  // Grading should adjust facets to get better parametric aspect
  // ratio of a grid cell.
  // NOTE: However if it is true, it can cause a segmentation fault
  //       in the facetter.
  ref->set_grading_mode (0);
    
  sil_acis_check (api_set_default_refinement (ref),
		  "acis facet_init", "set default refinement");

  API_END;

  return;
}

void ACIS::facet_object (void *object)
{
  ENTITY *e = (ENTITY *)object;
  API_BEGIN;
  if (is_BODY (e)) {
    facet_body ((BODY *)e);
  } else if (is_LUMP(e)) {
    facet_lump ((LUMP *)e);
  } else if (is_SHELL(e)) {
    facet_shell ((SHELL *)e);
  } else if (is_FACE (e)) {
    facet_face ((FACE *)e);
  }
  API_END;
  return;
}

void ACIS::draw_init ()
{
  return;
}

void ACIS::draw_object (void *object, SimpleList<void> const &hirefs)
{
  ENTITY *e = (ENTITY *)object;
  if (!is_LIGHT (e)) {
    drawn.clear ();
    draw_entity (e, hirefs);
  }
  return;
}

void ACIS::draw_fixed_object (void *object, SimpleList<void> const &hirefs)
{
  ENTITY *e = (ENTITY *)object;
  if (is_LIGHT (e)) {
    drawn.clear ();
    ((LIGHT *)e)->gl_render ();
  }
  return;
}

void ACIS::del_object (void *object)
{
  ENTITY *e = (ENTITY *)object;
  del_symbol (e);
  ATTRIB_MATERIAL *am = (ATTRIB_MATERIAL *)
    find_attrib (e, ATTRIB_SIL_TYPE, ATTRIB_MATERIAL_TYPE);
  while (am) {
    am->unhook ();
    sil_acis_check (api_del_entity (am),
		    "acis del_object",
		    "delete material attribute in del_object");
    am = (ATTRIB_MATERIAL *)
      find_attrib (e, ATTRIB_SIL_TYPE, ATTRIB_MATERIAL_TYPE);
  }
  sil_acis_check (api_del_entity ((ENTITY *) object),
		  "acis del_object", "delete acis entity");
  return;
}

void ACIS::unfacet_body (BODY *b)
{
  LUMP *l = b->lump();
  LUMP *fl = l;
  while (l) {
    unfacet_lump (l);
    l = l->next ();
    if (fl == l)
      break;
  }
  return;
}

void ACIS::unfacet_lump (LUMP *l)
{
  SHELL *s = l->shell();
  SHELL *fs = NULL;
  while (s) {
    unfacet_shell (s);
    s = s->next ();
    if (fs == s)
      break;
  }
  return;
}

void ACIS::unfacet_shell (SHELL *s)
{
  FACE *f =s->face_list();
  FACE *ff = f;
  while (f) {
    unfacet_face (f);
    f = f->next_in_list ();
    if (ff == f)
      break;
  }
  return;
}

void ACIS::unfacet_face (FACE *f)
{
  af_delete_facets ((ENTITY *)f);
  ATTRIB_EYE *ae = (ATTRIB_EYE *)
    find_attrib (f, ATTRIB_EYE_TYPE);
  while (ae) {
    ae->unhook ();
    sil_acis_check (api_del_entity (ae),
		    "acis unfacet_face", "delete eye attribute");
    ae = (ATTRIB_EYE *)
      find_attrib (f, ATTRIB_EYE_TYPE);
  }
  return;
}

void ACIS::facet_body (BODY *b)
{
  LUMP *l = b->lump();
  LUMP *fl = l;
  while (l) {
    facet_lump (l);
    l = l->next ();
    if (fl == l)
      break;
  }
  return;
}

void ACIS::facet_lump (LUMP *l)
{
  SHELL *s = l->shell();
  SHELL *fs = NULL;
  while (s) {
    facet_shell (s);
    s = s->next ();
    if (fs == s)
      break;
  }
  return;
}

void ACIS::facet_shell (SHELL *s)
{
  FACE *f =s->face_list();
  FACE *ff = f;
  while (f) {
    facet_face (f);
    f = f->next_in_list ();
    if (ff == f)
      break;
  }
  return;
}

void ACIS::facet_face (FACE *f)
{
  sil_acis_check(api_facet_entity ((ENTITY *)f),
		 "acis facet_face", "facet entity face");
  return;
}

void ACIS::draw_entity (ENTITY *e, SimpleList<void> const &hirefs)
{
  API_BEGIN;
  ATTRIB_MATERIAL *aa = dynamic_cast<ATTRIB_MATERIAL*> 
    (find_attrib (e, ATTRIB_SIL_TYPE, ATTRIB_MATERIAL_TYPE));
  bool new_attr = false;
  Material *a = dynamic_cast<Material*> (aa);
  if (!a) {
    new_attr = true;
    a = new Material ();
  }
  if (is_BODY (e)) {
    draw_body ((BODY *)e, a, false, hirefs);
  } else if (is_LUMP(e)) {
    draw_lump ((LUMP *)e, a, false, hirefs);
  } else if (is_SHELL(e)) {
    draw_shell ((SHELL *)e, a, false, hirefs);
  } else if (is_FACE(e)) {
    draw_face ((FACE *)e, a, false, hirefs);
  } else if (is_WIRE(e)) {
    draw_wire ((WIRE *)e, a, false, hirefs);
  } else if (is_LOOP(e)) {
    draw_loop ((LOOP *)e, a, false, hirefs);
  } else if (is_COEDGE(e)) {
    draw_edge (((COEDGE *)e)->edge(), a, false, hirefs,
	       ((COEDGE *)e)->start (), ((COEDGE *)e)->end (),
	       ((COEDGE *)e)->sense ());
  } else if (is_EDGE(e)) {
    draw_edge ((EDGE *)e, a, false, hirefs);
  } else if (is_VERTEX(e)) {
    draw_vertex ((VERTEX *)e, a,false, hirefs);
  } else {
    sil_msg_error ("acis draw_entity", "drawing method for entity type not implemented");
  }
  if (new_attr)
    delete a;
  API_END;
  return;
}

void ACIS::draw_body (BODY *b, Material *a, bool inv,
		      SimpleList<void> const &hirefs)
{
  ATTRIB_MATERIAL *aa = dynamic_cast<ATTRIB_MATERIAL*> 
    (find_attrib (b, ATTRIB_SIL_TYPE, ATTRIB_MATERIAL_TYPE));
  if (aa)
    a = static_cast<Material*> (aa);
  LUMP *l = b->lump();
  LUMP *fl = l;
  if (!inv)
    inv = hirefs.is_member ((void *)b);
  while (l) {
    draw_lump (l, a, inv, hirefs,
	       (b->transform ())?(&b->transform () -> transform ()):NULL);
    l = l->next ();
    if (fl == l)
      break;
  }
  return;
}

void ACIS::draw_lump (LUMP *l, Material *a, bool inv,
		      SimpleList<void> const &hirefs,
		      transf const *trans)
{
  ATTRIB_MATERIAL *aa = dynamic_cast<ATTRIB_MATERIAL*> 
    (find_attrib (l, ATTRIB_SIL_TYPE, ATTRIB_MATERIAL_TYPE));
  if (aa)
    a = static_cast<Material*> (aa);
  if (!inv)
    inv = hirefs.is_member ((void *)l);
  SHELL *s = l->shell();
  SHELL *fs = s;
  while (s) {
    draw_shell (s, a, inv, hirefs, trans);
    s = s->next ();
    if (fs == s)
      break;
  }
  return;
}

void ACIS::draw_shell (SHELL *s, Material *a, bool inv,
		       SimpleList<void> const &hirefs,
		       transf const *trans)
{
  ATTRIB_MATERIAL *aa = dynamic_cast<ATTRIB_MATERIAL*> 
    (find_attrib (s, ATTRIB_SIL_TYPE, ATTRIB_MATERIAL_TYPE));
  if (aa)
    a = static_cast<Material*> (aa);
  if (!inv)
    inv = hirefs.is_member ((void *)s);
  FACE *f = s->face_list();
  FACE *ff = f;
  while (f) {
    draw_face (f, a, inv, hirefs, trans);
    f = f->next_in_list ();
    if (ff == f)
      break;
  }
  WIRE *w = s->wire_list ();
  WIRE *fw = w;
  while (w) {
    draw_wire (w, a, inv, hirefs, trans);
    w = w->next_in_list ();
    if (fw == w)
      break;
  }
  return;
}

void ACIS::draw_face (FACE *f, Material *a, bool inv,
		      SimpleList<void> const &hirefs,
		      transf const *trans)
{
  ATTRIB_MATERIAL *aa = dynamic_cast<ATTRIB_MATERIAL*> 
    (find_attrib (f, ATTRIB_SIL_TYPE, ATTRIB_MATERIAL_TYPE));
  if (aa)
    a = static_cast<Material*> (aa);
  INDEXED_MESH *m;
  sil_acis_check(api_get_indexed_mesh (f, m),
		 "acis draw_face", "get indexed mesh");
  bool iinv = false;
  if (!inv)
    inv = hirefs.is_member ((void *)f);
  else 
    iinv = true;
  if (m) {
    indexed_polygon *poly;
    position v;
    unit_vector n;
    int p_num = m->get_num_polygon ();
    int vertex_num;
    polygon_vertex **vertex_list, *V;
    if (sil_get_draw_solid ()) {
      a->gl_material (inv);
      int i = 0;
      while (i < p_num) {
	poly = m->get_polygon (i++);
	vertex_num = poly->num_vertex ();
	vertex_list = poly->get_vertices ();
	int j = 0;
	glBegin(GL_POLYGON);
	while (j < vertex_num) {
	  V = vertex_list[j++];
	  v = V -> get_position ();
	  n = V -> get_normal ();
	  if (trans) {
	    n *= *trans;
	    v *= *trans;
	  }
	  glNormal3d (n.x (), n.y (), n.z ());
	  glVertex3d (v.x (), v.y (), v.z ());
	}
	glEnd ();
      }
    } else {
      a->gl_material_bright (inv);
      int i = 0;
      while (i < p_num) {
	poly = m->get_polygon (i++);
	vertex_num = poly->num_vertex ();
	vertex_list = poly->get_vertices ();
	int j = 0;
	glBegin (GL_LINE_LOOP);
	while (j < vertex_num) {
	  V = vertex_list[j++];
	  v = V -> get_position ();
	  n = V -> get_normal ();
	  glNormal3d (n.x (), n.y (), n.z ());
	  glVertex3d (v.x (), v.y (), v.z ());
	}
	glEnd ();
      }
    }
  }
  // Draw edges if requested
  if (sil_get_draw_sub_parts ()) {
    LOOP *l = f->loop();
    LOOP *fl = l;
    while (l) {
      draw_loop (l, a, iinv, hirefs, trans);
      l = l->next ();
      if (fl == l)
	break;
    }
  }
  return;
}

void ACIS::draw_wire (WIRE *w, Material *a, bool inv,
		      SimpleList<void> const &hirefs,
		      transf const *trans)
{
  ATTRIB_MATERIAL *aa = dynamic_cast<ATTRIB_MATERIAL*> 
    (find_attrib (w, ATTRIB_SIL_TYPE, ATTRIB_MATERIAL_TYPE));
  if (aa)
    a = static_cast<Material*> (aa);
  if (!inv)
    inv = hirefs.is_member ((void *)w);
  COEDGE *c = w->coedge();
  COEDGE *fc = c;
  while (c) {
    draw_edge (c->edge(), a, inv, hirefs,
	       c->start (), c->end (), c->sense (), trans);
    c = c->next ();
    if (fc == c)
      break;
  }
  return;
}

void ACIS::draw_loop (LOOP *l, Material *a, bool inv,
		      SimpleList<void> const &hirefs, 
		      transf const *trans)
{
  ATTRIB_MATERIAL *aa = dynamic_cast<ATTRIB_MATERIAL*> 
    (find_attrib (l, ATTRIB_SIL_TYPE, ATTRIB_MATERIAL_TYPE));
  if (aa)
    a = static_cast<Material*> (aa);
  if (!inv)
    inv = hirefs.is_member ((void *)l);
  COEDGE *c = l->start();
  COEDGE *fc = c;
  while (c) {
    draw_edge (c->edge (), a, inv, hirefs,
    	       c->start (), c->end (), c->sense (), trans);
    c = c->next ();
    if (fc == c)
      break;
  }
  return;
}

void ACIS::draw_edge (EDGE *e, Material *a, bool inv,
		      SimpleList<void> const &hirefs,
		      VERTEX *start, VERTEX *end, bool sense,
		      transf const *trans)
{
  if (!drawn.add_first_check (e) && !inv)
    return;
  ATTRIB_MATERIAL *aa = dynamic_cast<ATTRIB_MATERIAL*> 
    (find_attrib (e, ATTRIB_SIL_TYPE, ATTRIB_MATERIAL_TYPE));
  if (aa)
    a = static_cast<Material*> (aa);
  bool iinv = false;
  if (!inv)
    inv = hirefs.is_member ((void *)e);
  else
    iinv = true;
  a->gl_material_bright (inv);
  if (!start)
    start = e->start ();
  if (!end)
    end = e->end ();
  position A = start->geometry () -> coords ();
  position B = end->geometry () -> coords ();
  CURVE *c = e->geometry ();
  double rb_l, rb_r, rb_n, rb_f, rb_b, rb_t;
  sil_get_render_box (rb_l, rb_r, rb_n, rb_f, rb_b, rb_t);
  if (is_STRAIGHT (c)) {
    STRAIGHT *s = (STRAIGHT *)c;
    const interval I = c->equation () . 
      param_range (box (interval (rb_l, rb_r),
			interval (rb_n, rb_f),
			interval (rb_b, rb_t)));
    glLineStipple (1, 0xFFFF);
    glBegin (GL_LINES);
    position P;
    double ps = I.start_pt ();
    double pe = I.end_pt ();
    ps = s->equation () . param (A, ps);
    pe = s->equation () . param (B, pe);
    s->equation () . eval (ps, P);
    if (trans)
      P *= *trans;
    glVertex3d (P.x (), P.y (), P.z ());
    s->equation () . eval (pe, P);
    if (trans)
      P *= *trans;
    glVertex3d (P.x (), P.y (), P.z ());
    glEnd ();
  } else if (is_UNDEFC (c)) {
    glLineStipple (1, 0xAAAA);
    UNDEFC *U = (UNDEFC *)c;
    position PS = U->start_point ();
    position DS = PS + U->start_dir ();
    position PE = U->end_point ();
    position DE = PS - U->end_dir ();
    if (trans) {
      PS *= *trans;
      DS *= *trans;
      PE *= *trans;
      DE *= *trans;
    }
    glBegin (GL_LINES);    
    glVertex3d (PS.x (), PS.y (), PS.z ());
    glVertex3d (DS.x (), DS.y (), DS.z ());
    glVertex3d (PE.x (), PE.y (), PE.z ());
    glVertex3d (DE.x (), DE.y (), DE.z ());
    glEnd ();
    glLineStipple (2, 0xAAAA);
    glBegin (GL_LINES);  
    glVertex3d (PS.x (), PS.y (), PS.z ());
    glVertex3d (PE.x (), PE.y (), PE.z ());
    glEnd ();
  } else {
    if (is_ELLIPSE (c) || is_COMPCURV (c) || is_INTCURVE (c)) {
      glLineStipple (1, 0xFFFF);
    } else {
      sil_msg_warn ("acis draw_edge",
		    "method for drawing edge type not implemented");
      glLineStipple (2, 0xAAAA);
    }
    const curve *C = &c->equation ();
    const interval I = C->param_range (box (interval (rb_l, rb_r),
					    interval (rb_n, rb_f),
					    interval (rb_b, rb_t)));
    double ps = I.start_pt ();
    double pe = I.end_pt ();
    ps = C->param (A, ps);
    pe = C->param (B, pe);
    if (pe < ps) {
      if (C->periodic ()) {
	double per = C->param_period ();
	while (pe < ps - per)
	  pe += per;
      }
      double h = ps;
      ps = pe;
      pe = h;
      position H = A;
      A = B;
      B = H;
    } else {
      if (C->periodic ()) {
	double per = C->param_period ();
	while (pe > ps + per)
	  pe -= per;
      }
    }
    if (e->sense ())
      sense = !sense;
    if (C->periodic () && sense) {
      // If sense is reversed and we have a periodic function, then
      // we exchange start and end position to get the correct 
      // orientation (negative).
      double h = pe - C->param_period ();
      pe = ps;
      ps = h;
      position H = A;
      A = B;
      B = H;
    }
    double step = (pe - ps) / 4.0;
    double length = C->length (ps, ps + step);
    while (length > max_ll && step > 2.0 * get_resabs ()) {
      step /= 2.0;
      length = C->length (ps, ps + step);
    }
    position P;
    glBegin (GL_LINE_STRIP);
    if (trans)
      A *= *trans;
    glVertex3d (A.x(), A. y(), A.z ());
    for (double p = ps + step; p < pe; p += step) {
      C->eval (p, P);
      if (trans)
	P *= *trans;
      glVertex3d (P.x (), P.y (), P.z ());
    }
    if (trans)
      B *= *trans;
    glVertex3d (B.x(), B.y(), B.z ());
    glEnd ();
  }
  if (sil_get_draw_sub_parts () &&
      e->start () && e->end () &&
      e->start () != e->end () &&
      (e->start () -> geometry () -> coords () - 
       e->end () -> geometry () -> coords ()).len () >= sil_get_epsilon ()) {
    draw_vertex (e->start (), a, iinv, hirefs, trans);
    draw_vertex (e->end (), a, iinv, hirefs, trans);
  }
  return;
}

void ACIS::draw_vertex (VERTEX *v, Material *a,  bool inv,
			SimpleList<void> const &hirefs,
			transf const *trans)
{
  if (!drawn.add_first_check (v) && !inv)
    return;
  ATTRIB_MATERIAL *aa = dynamic_cast<ATTRIB_MATERIAL*> 
    (find_attrib (v, ATTRIB_SIL_TYPE, ATTRIB_MATERIAL_TYPE));
  if (aa)
    a = static_cast<Material*> (aa);
  if (!inv)
    inv = hirefs.is_member ((void *)v);
  a->gl_material_bright (inv);
  position p = v->geometry () -> coords ();
  if (trans) 
    p *= *trans;
  glBegin (GL_POINTS);
  glVertex3d (p.x (), p.y (), p.z ());
  glEnd ();
  return;
}

void ACIS::load_file (FILE *fp, const char *format, bool ext, bool replace)
{
  int filetypes_id = 0;
  if (strcmp (format, "sat") == 0)
    filetypes_id = 1;
  else if (strcmp (format, "sat_bin") == 0)
    filetypes_id = 2;
  else if (strcmp (format, "iges") == 0)
    filetypes_id = 3;
  else if (strcmp (format, "step") == 0)
    filetypes_id = 4;
  else
    sil_msg_error ("acis open", "unsupported file format");
  API_BEGIN;
  ENTITY_LIST *loadlist = new ENTITY_LIST;
  if (filetypes_id == 4) {
    // STEP
    string fn = sil_tmp_filename ();
    FILE *tf = fopen (fn.c_str (), "w");
    if (tf) {
      char buf[512];
      size_t l;
      while ((l = fread (buf, sizeof (char), 511, fp)) > 0)
	fwrite (buf, sizeof (char), l, tf);
      fclose (tf);
      char *acis_path = getenv ("A3DT");
      if (acis_path == NULL || !*acis_path)
	acis_path = "/opt/acis/";
      char *step_path = new char[strlen (acis_path) + 18];
      strcpy (step_path, acis_path);
      strcat (step_path, "/step/tools/linux");
      sil_acis_check (api_set_str_option ("stp_tool_path", step_path),
		      "acis::load", "api_set_str_option");
      sil_acis_check (api_initialize_step (),
		      "acis::load", "api_initialize_step");
      sil_acis_check (api_step_convert_step_to_acisentlist 
		      ((char *)fn.c_str (), loadlist, NULL, NULL),
		      "acis::load", 
		      "api_step_convert_step_to_acisentlist");
      sil_acis_check (api_terminate_step (),
		      "acis::load", "api_terminate_step");
      delete []step_path;
    } else {
      sil_msg_error ("acis::load", "loading step file failed");
    }
    unlink (fn.c_str ());
  } else if (filetypes_id == 3) {
    // IGES
    char *acis_path = getenv ("A3DT");
    if (acis_path == NULL || !*acis_path)
      acis_path = "/opt/acis/";
    char *iges_path = new char[strlen (acis_path) + 23];
    strcpy (iges_path, acis_path);
    strcat (iges_path, "/iges/igac/iges_04.dbt");
    sil_acis_check (api_set_str_option ("iges_dbt_file", iges_path),
		    "acis::load", "api_set_str_option");
    FILE* db = fopen (iges_path, "rb");
    if(db) {
      sil_acis_check (api_initialize_iges (),
		      "acis::load", "api_initialize_iges");
      fclose(db);
      sil_acis_check (api_set_int_option ("iges_proc_msbo", 0),
		      "acis::load", "api_set_int_option");
      int success;
      sil_acis_check (api_iges_read (fp, NULL, 0, *loadlist, success, NULL),
		      "acis::load", "api_iges_read");
      sil_acis_check (api_terminate_iges(),
		      "acis::load", "api_terminate_iges");
    } else {
      sil_msg_error ("acis::load", "cannot locate IGES access library");
    }
    delete []iges_path;
  } else {
    // SAT
    sil_acis_check (api_restore_entity_list (fp, (filetypes_id==1),
					     *loadlist),
		    "acis open", "restore entity list");
  }
  loadlist->init ();
  ENTITY *e = loadlist->next ();
  ATTRIB_GEN_NAME *attrib;
  bool v = false;
  char *n = NULL;
  char nn[20];
  while (e) {
    if (ext || !is_LIGHT(e)) {
      if (!ext) {
	del_symbol (e);
	ATTRIB_MATERIAL *am = (ATTRIB_MATERIAL *)
	  find_attrib (e, ATTRIB_SIL_TYPE, ATTRIB_MATERIAL_TYPE);
	while (am) {
	  am->unhook ();
	  sil_acis_check (api_del_entity (am),
			  "acis open", "delete material attribute");
	  am = (ATTRIB_MATERIAL *)
	    find_attrib (e, ATTRIB_SIL_TYPE, ATTRIB_MATERIAL_TYPE);
	}
      }
      attrib = NULL;
      sil_acis_check (api_find_named_attribute (e, "SIL_NAME", attrib),
		      "acis open", "find named attribute SIL_NAME");
      if (attrib) {
	n = (char *)(((ATTRIB_GEN_STRING *)attrib)->value ());
	attrib = NULL;
	sil_acis_check(api_find_named_attribute (e, "SIL_VISIBLE", attrib),
		       "acis open", "find named attribute SIL_VISIBLE");
	if (attrib)
	  v = true;
	else
	  v = false;
	bool f = sil_add_symbol (n, (void *)e, "acis", v);
	if (!f) {
	  if (replace) {
	    sil_del_symbol (n);
	    char *str = new char[strlen (n) + 25];
	    strcpy (str, "replacing entity symbol ");
	    strcat (str, n);
	    sil_msg_warn("acis open", str);
	    delete []str;
	    if (!sil_add_symbol (n, e, "acis", v))
	      sil_msg_error ("acis open", "can't add replacement symbol");
	  } else {
	    char *str = new char[strlen (n) + 25];
	    strcpy (str, "entity symbol ");
	    strcat (str, n);
	    strcat (str, " exists.");
	    sil_msg_warn("acis open", str);
	    delete []str;
	  }
	}
      } else {
	unsigned long tn = sil_tmp_name ();
	snprintf (nn, 19, "Model_%lu", tn);
	while (!sil_add_symbol (nn, e, "acis", true)) {
	  sil_acis_check(api_remove_generic_named_attribute (e, "SIL_NAME"),
			 "acis open", "remove named attribute SIL_NAME");
	  unsigned long tnn = sil_tmp_name ();
	  if (tnn == tn) {
	    sil_msg_error ("acis open",
			   "no tmp_name symbols left for adding symbol");
	    break;
	  }
	  snprintf (nn, 19, "Model_%lu", tnn);
	}
      }
    }
    e = loadlist->next();
  }
  loadlist->clear ();
  delete loadlist;
  API_END;

  return;
}

void ACIS::save_file (FILE *fp, const char *format, bool ext)
{
  int filetypes_id = 0;
  if (strcmp (format, "sat") == 0)
    filetypes_id = 1;
  else if (strcmp (format, "sat_bin") == 0)
    filetypes_id = 2;
  else if (strcmp (format, "iges") == 0)
    filetypes_id = 3;
  else if (strcmp (format, "step") == 0)
    filetypes_id = 4;
  else {
    sil_msg_error ("acis save", "unsupported file format");
    return;
  }
  DELTA_STATE *ds1 = NULL;
  if (!ext) {
    api_logging (TRUE);
    api_note_state (ds1);
  }
  ENTITY_LIST *savelist = new ENTITY_LIST;
  const Symbol *s = sil_first_symbol ();
  while (s) {
    if (sil_symbol_is (s, "acis")) {
      if (ext || !is_LIGHT ((ENTITY *)sil_get_symbol_data (s))) {
	ENTITY *e = (ENTITY *)sil_get_symbol_data (s);
	if (!ext) {
	  del_symbol (e);
	  ATTRIB_MATERIAL *am = (ATTRIB_MATERIAL *)
	    find_attrib (e, ATTRIB_SIL_TYPE, ATTRIB_MATERIAL_TYPE);
	  while (am) {
	    am->unhook ();
	    sil_acis_check (api_del_entity (am),
			    "acis save", "delete material attribute");
	    am = (ATTRIB_MATERIAL *)
	      find_attrib (e, ATTRIB_SIL_TYPE, ATTRIB_MATERIAL_TYPE);
	  }
	}
	savelist->add (e);
      }
    }
    s = sil_next_symbol ();
  }
  if (filetypes_id == 4) {
    // STEP
    char *acis_path = getenv ("A3DT");
    if (acis_path == NULL || !*acis_path)
      acis_path = "/opt/acis/";
    char *step_path = new char[strlen (acis_path) + 18];
    strcpy (step_path, acis_path);
    strcat (step_path, "/step/tools/linux");
    sil_acis_check (api_set_str_option ("stp_tool_path", step_path),
		    "acis::save", "api_set_str_option");
    sil_acis_check (api_initialize_step (),
		    "acis::save", "api_initialize_step");
    string fn = sil_tmp_filename ();
    sil_acis_check (api_step_convert_acisentlist_to_step 
		    (*savelist, (char *)fn.c_str (), NULL, NULL),
		    "acis::save", 
		    "api_step_convert_acisentlist_to_step");
    FILE *tf = fopen (fn.c_str (), "r");
    if (tf) {
      char buf[512];
      size_t l;
      while ((l = fread (buf, sizeof (char), 511, tf)) > 0)
	fwrite (buf, sizeof (char), l, fp);
      fclose (tf);
    } else {
      sil_msg_error ("acis::save", "creating step file failed");
    }
    unlink (fn.c_str ());
    sil_acis_check (api_terminate_step (),
		    "acis::save", "api_terminate_step");
    delete []step_path;
  } else if (filetypes_id == 3) {
    // IGES
    char *acis_path = getenv ("A3DT");
    if (acis_path == NULL || !*acis_path)
      acis_path = "/opt/acis/";
    char *iges_path = new char[strlen (acis_path) + 23];
    strcpy (iges_path, acis_path);
    strcat (iges_path, "/iges/igac/iges_04.dbt");
    sil_acis_check (api_set_str_option ("iges_dbt_file", iges_path),
		    "acis::save", "api_set_str_option");
    FILE* db = fopen (iges_path, "rb");
    if(db) {
      sil_acis_check (api_initialize_iges (),
		      "acis::save", "api_initialize_iges");
      fclose(db);
      sil_acis_check (api_set_int_option ("iges_proc_msbo", 0),
		      "acis::save", "api_set_int_option");
      sil_acis_check (api_iges_write (fp, NULL, 0, savelist, NULL),
		      "acis::save", "api_iges_write");
      sil_acis_check (api_terminate_iges(),
		      "acis::save", "api_terminate_iges");
    } else {
      sil_msg_error ("acis::save", "cannot locate IGES access library");
    }
    delete []iges_path;
  } else {
    // SAT
    FileInfo *fi = new FileInfo;
    char *id = new char[strlen (VERSION) + 6];
    strcpy (id, "SiL V");
    strcat (id, VERSION);
    fi->set_product_id (id);
    fi->set_units (1.0);
    sil_acis_check(api_set_file_info (FileId|FileUnits, *fi),
		   "acis save", "set save file information");
    sil_acis_check(api_save_entity_list (fp, (filetypes_id==1), *savelist),
		   "acis save", "save entity list");
    delete fi;
    delete []id;
  }

  savelist->clear ();
  delete savelist;
  if (ext) {
    api_logging (FALSE);
    DELTA_STATE *ds2 = NULL;
    api_note_state (ds2);
    if (ds2) {
      api_change_state (ds2);
      api_delete_ds (ds2);
    }
    if (ds1)
      api_delete_ds (ds1);
  }
  return;
}

const char **ACIS::file_formats () const
{
  return fl_formats;
}

const char *ACIS::name () const
{
  return nm;
}

void ACIS::del_symbol (ENTITY *e)
{
  API_BEGIN;
  ATTRIB_GEN_NAME *a;
  sil_acis_check (api_find_named_attribute (e,
					    "SIL_NAME", a),
		  "acis del_symbols", "find named attribute SIL_NAME");
  while (a) {
    a->unhook ();
    sil_acis_check (api_del_entity (a),
		    "acis del_symbols", "delete SIL_NAME attribute");
    sil_acis_check (api_find_named_attribute (e,
					      "SIL_NAME", a),
		    "acis del_symbols", "find named attribute SIL_NAME");
  }
  sil_acis_check (api_find_named_attribute (e,
					    "SIL_VISIBLE", a),
		  "acis del_symbols", "find named attribute SIL_VISIBLE");
  while (a) {
    a->unhook ();
    sil_acis_check (api_del_entity (a),
		    "acis del_symbols", "delete SIL_VISIBLE attribute");
    sil_acis_check (api_find_named_attribute (e,
					      "SIL_VISIBLE", a),
		    "acis del_symbols", "find named attribute SIL_VISIBLE");
  }
  if (is_BODY (e)) {
    unfacet_body ((BODY *)e);
  } else if (is_LUMP(e)) {
    unfacet_lump ((LUMP *)e);
  } else if (is_SHELL(e)) {
    unfacet_shell ((SHELL *)e);
  } else if (is_FACE (e)) {
    unfacet_face ((FACE *)e);
  }
  API_END;
  return;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
