// interface.hxx: interface class.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: interface.hxx,v 1.47.2.1 2001/08/16 16:19:05 scmfcl Exp $

#ifndef INTERFACE_HXX
#define INTERFACE_HXX 1

#include <string>
#include "list.hxx"

#ifdef DEBUG
#define CHECK_INTERFACE \
  if (!_sil_interface) { \
    cerr << "FATAL: There is no SiL interface\n"; \
    exit (1); \
  }
#else // DEBUG
#define CHECK_INTERFACE
#endif // DEBUG

#define SIL_TIMERS 10

class App;
class GUI;
class Module;
class Symbols;
class Symbol;

typedef enum {
  SIL_TRIANG_NONE,
  SIL_TRIANG_FRINGE_1,
  SIL_TRIANG_FRINGE_2,
  SIL_TRIANG_FRINGE_3,
  SIL_TRIANG_FRINGE_4,
  SIL_TRIANG_ALL
} SiL_TriangMode;

typedef enum {
  SIL_GRID_NONE,
  SIL_GRID_INTERIOR,
  SIL_GRID_TO_EDGES
} SiL_GridMode;

typedef enum {
  SIL_ADJUST_NONE,
  SIL_ADJUST_NON_GRID,
  SIL_ADJUST_ALL
} SiL_AdjustMode;

class PluginInterface
{
public:
  // Symbols
  bool add_symbol (char *n, void *d, char *t, bool v);
  const Symbol *find_symbol (const char *n);
  void del_symbol (const char *n);
  const Symbol *first_symbol ();
  const Symbol *next_symbol ();
  void clear_symbols ();
  void symbols_updated () const;
  void *symbol_data (const Symbol *s) const;
  const char *symbol_type (const Symbol *s) const;
  const char *symbol_name (const Symbol *s) const;
  bool symbol_is (const Symbol *s, const char *type) const;
  bool symbol_is_visible (const Symbol *s) const;
  void show_symbol (Symbol *s);
  void hide_symbol (Symbol *s);
  // Configuration
  bool draw_sub_parts () const;
  void set_draw_sub_parts (bool f);
  bool draw_solid () const;
  void render_box (double &a, double &b, double &c, double &d,
		   double &e, double &f) const;
  void refinement (double &surf_tol,
		   double &normal_tol,
		   SiL_TriangMode &t_mode,
		   SiL_GridMode &g_mode,
		   int &max_l,
		   int &min_l,
		   double &max_ll,
		   double &ar,
		   SiL_AdjustMode &am) const;
  // Render
  void refacet ();
  void redraw ();
  void refresh ();
  void hilite (bool hi, char const *name, void *ref);
  bool is_hilited (const Symbol *s, const void *r);
  // Modules
  Module *module (char *name) const;
  // Messages
  void msg (string const &text);
  void msg (string const &name, string const &text);
  void msg_warn (string const &name, string const &text);
  void msg_error (string const &name, string const &text);
  // Tmp name
  unsigned long tmp_name ();
  string const tmp_filename () const;
  // SiL precision
  const double epsilon () const;
  const double epsilon2 () const;
  // Progress bar
  void progress_bar (bool flag, bool cancel_flag);
  void progress_bar_assign (gfloat base, gfloat range);
  void progress_bar_percentage (gfloat p);
  void progress_bar_text (string const &str);
  bool progress_cancel ();
  // Time
  void start_time (unsigned long t);
  void measure_time (unsigned long &secs, unsigned int &msecs,
		     unsigned long t) const;
protected:
  PluginInterface ();
  ~PluginInterface ();
  friend class App;
private:
#ifdef HAVE_TIMEVAL
  void timeval_to_secs (struct timeval const *tv, 
			unsigned long &secs, unsigned int &msecs) const;
  void diff_timeval (struct timeval &d,
		     struct timeval const &t1,
		     struct timeval const &t2) const;
  struct timeval time_before[SIL_TIMERS];
#endif // HAVE_GETRUSAGE
  // Pointer to main application.
  App *app;
  // Progress bar base and range.
  gfloat pb_b, pb_r;
};

extern PluginInterface *_sil_interface;

// module 

#define SIL_MODULE \
  Module *create_module

// plugin

#define SIL_PLUGIN \
  void plugin

#define SIL_CLOSE_PLUGIN \
  void plugin_close

extern void sil_register_tool (void (*f) (void), 
			       const char *n,
			       const char *v,
			       const char *a,
			       const char *c,
			       const char *d,
			       const char *cl,
			       const char *m);
extern void sil_register_tool (void (*f) (SimpleList<Symbol const> *), 
			       const char *n,
			       const char *v,
			       const char *a,
			       const char *c,
			       const char *d,
			       const char *cl,
			       const char *m);

extern void sil_register_update_handler (void (*u) (void));
extern void sil_deregister_update_handler (void (*u) (void));

// Functions and macros

inline bool sil_add_symbol (char *n, void *d, char *t, bool v) 
{
  CHECK_INTERFACE;
  return _sil_interface->add_symbol (n, d, t, v);
}

inline const Symbol *sil_find_symbol (const char *n)
{
  CHECK_INTERFACE;
  return _sil_interface->find_symbol (n);
}

inline void sil_del_symbol (const char *n)
{
  CHECK_INTERFACE;
  _sil_interface->del_symbol (n);
  return;
}

inline const Symbol *sil_first_symbol ()
{
  CHECK_INTERFACE;
  return _sil_interface->first_symbol ();
}

inline const Symbol *sil_next_symbol ()
{
  CHECK_INTERFACE;
  return _sil_interface->next_symbol ();
}

inline void sil_clear_symbols ()
{
  CHECK_INTERFACE;
  _sil_interface->clear_symbols ();
  return;
}

inline void sil_updated ()
{
  CHECK_INTERFACE;
  _sil_interface->symbols_updated ();
  return;
}

inline void *sil_get_symbol_data (const Symbol *s)
{
  CHECK_INTERFACE;
  return _sil_interface->symbol_data (s);
}

inline const char *sil_get_symbol_name (const Symbol *s)
{
  CHECK_INTERFACE;
  return _sil_interface->symbol_name (s);
}

inline const char *sil_get_symbol_type (const Symbol *s)
{
  CHECK_INTERFACE;
  return _sil_interface->symbol_type (s);
}

inline bool sil_symbol_is (const Symbol *s, const char *str)
{
  CHECK_INTERFACE;
  return _sil_interface->symbol_is (s, str);
}

inline bool sil_symbol_is_visible (const Symbol *s)
{
  CHECK_INTERFACE;
  return _sil_interface->symbol_is_visible (s);
}

inline void sil_show_symbol (Symbol *s)
{
  CHECK_INTERFACE;
  _sil_interface->show_symbol (s);
  return;
}

inline void sil_hide_symbol (Symbol *s)
{
  CHECK_INTERFACE;
  _sil_interface->hide_symbol (s);
  return;
}

inline bool sil_get_draw_sub_parts ()
{
  CHECK_INTERFACE;
  return _sil_interface->draw_sub_parts ();
}

inline void sil_set_draw_sub_parts (bool f)
{
  CHECK_INTERFACE;
  return _sil_interface->set_draw_sub_parts (f);
}

inline bool sil_get_draw_solid ()
{
  CHECK_INTERFACE;
  return _sil_interface->draw_solid ();
}

inline void sil_get_render_box (double &a, double &b, double &c,
				double &d, double &e, double &f)
{
  CHECK_INTERFACE;
  _sil_interface->render_box (a, b, c, d, e, f);
  return;
}

inline void sil_get_refinement (double &surf_tol,
				double &normal_tol,
				SiL_TriangMode &t_mode,
				SiL_GridMode &g_mode,
				int &max_l,
				int &min_l,
				double &max_ll,
				double &ar,
				SiL_AdjustMode &am)
{
  CHECK_INTERFACE;
  _sil_interface->refinement (surf_tol, normal_tol,
			      t_mode, g_mode, max_l, min_l,
			      max_ll, ar, am);
  return;
}

inline void sil_refacet ()
{
  CHECK_INTERFACE;
  _sil_interface->refacet ();
  return;
}

inline void sil_redraw ()
{
  CHECK_INTERFACE;
  _sil_interface->redraw ();
  return;
}

inline void sil_refresh ()
{
  CHECK_INTERFACE;
  _sil_interface->refresh ();
  return;
}

inline void sil_hilite (bool hi, char const *name, void *ref) 
{
  CHECK_INTERFACE;
  _sil_interface->hilite (hi, name, ref);
  return;
}

inline bool sil_is_hilited (const Symbol *s, const void *r)
{
  CHECK_INTERFACE;
  return _sil_interface->is_hilited (s, r);
}

inline Module *sil_get_module (char *name)
{
  CHECK_INTERFACE;
  return _sil_interface->module (name);
}

inline void sil_msg (string const &text)
{
  if (_sil_interface)
    _sil_interface->msg (text);
  else
    cout << text << endl;
  return;
}

inline void sil_msg (string const &name, string const &text)
{
  if (_sil_interface)
    _sil_interface->msg (name, text);
  else
    cout << name << ": " << text << endl;
  return;
}

inline void sil_msg_warn (string const &name, string const &text)
{
  if (_sil_interface)
    _sil_interface->msg_warn (name, text);
  else
    cout << name << ": " << text << endl;
  return;
}

inline void sil_msg_error (string const &name, string const &text)
{
  if (_sil_interface)
    _sil_interface->msg_error (name, text);
  else
    cout << name << ": " << text << endl;
  return;
}

inline unsigned long sil_tmp_name ()
{
  CHECK_INTERFACE;
  return _sil_interface->tmp_name ();
}

inline string const sil_tmp_filename ()
{
  CHECK_INTERFACE;
  return _sil_interface->tmp_filename ();
}

inline const double sil_get_epsilon ()
{
  CHECK_INTERFACE;
  return _sil_interface->epsilon ();
}

inline const double sil_get_epsilon2 ()
{
  CHECK_INTERFACE;
  return _sil_interface->epsilon2 ();
}

inline void sil_progress_bar (bool flag, bool cancel_flag = true) {
  CHECK_INTERFACE;
  _sil_interface->progress_bar (flag, cancel_flag);
  return;
}

inline void sil_progress_bar_assign (gfloat base, gfloat range) {
  CHECK_INTERFACE;
  _sil_interface->progress_bar_assign (base, range);
  return;
}

inline void sil_progress_bar_percentage (double p) {
  CHECK_INTERFACE;
  _sil_interface->progress_bar_percentage ((gfloat)p);
  return;
}

inline void sil_progress_bar_text (string const &str) {
  CHECK_INTERFACE;
  _sil_interface->progress_bar_text (str);
  return;
}

inline bool sil_cancel_request () {
  CHECK_INTERFACE;
  return _sil_interface->progress_cancel ();
}

inline void sil_start_time (unsigned long t = 0) {
  CHECK_INTERFACE;
  _sil_interface->start_time (t);
  return;
}

inline void sil_measure_time (unsigned long &secs, unsigned int &msecs,
			      unsigned long t = 0) {
  CHECK_INTERFACE;
  _sil_interface->measure_time (secs, msecs, t);
  return;
}

#endif // !INTERFACE_HXX

/// Local Variables:
/// exe-comm: "./sil"
/// compile-command: "make"
/// End:
