// analyse_positions.cc: Analyse a SiL model / positions.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: analyse_positions.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (analyse_positions,
       "$Id: analyse_positions.cc 1 2005-08-10 23:35:07Z frank $");

#include "list.hxx"
#include "xstring.hxx"
#include "kernel/analyse.hxx"
#include "kernel/combinatorics.hxx"
#include "kernel/geometry.hxx"

void Analyse::eval_positions ()
{
  // Intersection positions are also positions and are considered as
  // position for all regularities considered here.
  if (intersection_positions) {
    ListNode<Cluster> *rn = intersection_positions->first ();
    while (rn) {
      position_features.add_first 
	(new Position (rn ->data (),
		       *static_cast<Vector3*> (static_cast<Position*> 
					       (rn->data () -> value ()))));
      rn = rn->next ();
    }
  }

  if (position_features.length () < 1)
    return;

  // Cluster positions.
  sil_progress_bar_text ("Clustering " + 
			 as_string (position_features.length ()) +
			 " positions");
  List<Cluster> *clusters = new List<Cluster>;
  cluster (&position_features, tol_eq, tol_len_dist, NAN, true, clusters);

  simplify_cluster_hierarchy (clusters, tol_len_dist, max_len / 2);

  reg_list->add_first ((List<Regularity> *)clusters);
  
  if (f_positions_partial) {

    sil_progress_bar_text ("Checking for partial positions");
    
    // Check for partially equal positions.

    List<Cluster> dirs;

    // Collect planes and lines for projection.
    if (conical_angle_regularities) {
      // Check the orthogonal systems.
      ListNode<ConicalAngleRegular> *car = conical_angle_regularities->
	first ();
      while (car) {
	if (car->data () -> orthogonal ()) {
	  // Orthogonal system provides three planes and three lines.
	  Cluster *c= car->data () -> directions () -> first ();
	  while (c) {
	    dirs.add_last_check (c);
	    c = car->data () -> directions () -> next ();
	  }
	}
	car = car->next ();
      }
    }
    if (parallel_directions) {
      // Check for major axes. A major axis should have more than 1/3 of
      // the total faces as owners.
      List<Cluster> bcl;
      ListNode<Cluster> *c = parallel_directions->first ();
      while (c) {
	c->data () -> get_all_base_clusters (&bcl);
	c = c->next ();
      }
      c = bcl.first ();
      while (c) {
	if (!dirs.is_member (c->data ())) {
	  // Count face owners.
	  SimpleList<Entity> owners;
	  c->data () -> get_all_generators (&owners);
	  unsigned long n = 0;
	  Entity *e = owners.first ();
	  while (e) {
	    if (e->is (SIL_FACE_TYPE))
	      ++n;
	    e = owners.next ();
	  }
	  if (3 * n > static_cast<Real> (sld->faces () . length ()))
	    dirs.add_first (c->data ());
	}
	c = c->next ();
      }
    }

    List<Cluster> *pep = new List<Cluster>;
    List<Cluster> *lep = new List<Cluster>;

    // Each direction in dirs represents a plane and a line to project
    // the points on and check if they are equal.
    ListNode<Cluster> *c = dirs.first ();
    while (c) {
      check_plane_partially_equal (c->data (), clusters, pep);
      check_line_partially_equal (c->data (), clusters, lep);
      c = c->next ();
    }

    // Compare clusters: if there are clusters for different
    // directions which have the same points, then use the one with
    // the smaller tolerance.

    sil_progress_bar_text ("Check for ambiguous partially plane equal");

    // Check all plane clusters
    c = pep->first ();
    while (c) {
      ListNode<Cluster> *cnext = c->next ();
      // Compare with remaining clusters which are caused by a different
      // direction cluster.
      PartiallyPlaneEqual *cpe = static_cast<PartiallyPlaneEqual*> 
	(c->data () -> value ());
      ListNode<Cluster> *d = cnext;
      while (d) {
	ListNode<Cluster> *dnext = d->next ();
	PartiallyPlaneEqual *dpe = static_cast<PartiallyPlaneEqual*> 
	  (d->data () -> value ());
	if (cpe->direction () != dpe->direction ()) {

	  // Check if there are clusters in c which are also in d with
	  // respect to the features
	  SimpleList<Cluster> clist;
	  clist.add_first (c->data ());
	  Cluster *cc;
	  while ((cc = clist.unlink_first ())) {
	    
	    // Check if cc is in d and report d cluster
	    Cluster *cd = find_equal_pe_cluster (cc, d->data ());
	    //if (cd && (!cd->parent_cluster () || cc->parent_cluster ())) 
	    //  cd = NULL;
	    if (cd) {
	      // cc is in d as cd
	      if (cc->tolerance () < cd->tolerance ()) {
		// cc is better than cd, so delete cd and combine it
		// with parents / top list
		if (!cd->collapse ()) {
		  // This means we delete d.
		  clist.clear ();
		  if (cd->sub_clusters ()) {
		    Cluster *sc;
		    while ((sc = cd->sub_clusters () -> unlink_first ())) {
		      sc->parent_cluster () = NULL;
		      pep->add_last (sc);
		    }
		    if (!dnext)
		      dnext = d->next ();
		  }
		  if (cnext == d)
		    cnext = dnext;
		  pep->del (cd);
		}
		delete cd;
	      } else {
		// cd is better than cc, so delete cc and combine it
		// with parents / top list
		if (!cc->collapse ()) {
		  // That means we delete c.
		  dnext = NULL;
		  clist.clear ();
		  if (cc->sub_clusters ()) {
		    Cluster *sc;
		    while ((sc = cc->sub_clusters () -> unlink_first ())) {
		      sc->parent_cluster () = NULL;
		      pep->add_last (sc);
		    }
		    if (!cnext)
		      cnext = c->next ();
		  }
		  pep->del (cc);
		}
		delete cc;
	      }
	    } else if (cc->sub_clusters ()) {
	      // Add the sub-clusters, as cc is not in d
	      Cluster *sc = cc->sub_clusters () -> first ();
	      while (sc) {
		clist.add_first (sc);
		sc = cc->sub_clusters () -> next ();
	      }
	    }

	  }

	}
	d = dnext;
      }
      c = cnext;
    }

    sil_progress_bar_text ("Check for ambiguous partially line equal");

    // Check all line clusters
    c = lep->first ();
    while (c) {
      ListNode<Cluster> *cnext = c->next ();
      // Compare with remaining clusters which are caused by a different
      // direction cluster.
      PartiallyLineEqual *cpe = static_cast<PartiallyLineEqual*> 
	(c->data () -> value ());
      ListNode<Cluster> *d = c->next ();
      while (d) {
	ListNode<Cluster> *dnext = d->next ();
	PartiallyLineEqual *dpe = static_cast<PartiallyLineEqual*> 
	  (d->data () -> value ());
	if (cpe->direction () != dpe->direction ()) {

	  // Check if there are clusters in c which are also in d with
	  // respect to the features
	  SimpleList<Cluster> clist;
	  clist.add_first (c->data ());
	  Cluster *cc;
	  while ((cc = clist.unlink_first ())) {
	    
	    // Check if cc is in d and report d cluster
	    Cluster *cd = find_equal_pe_cluster (cc, d->data ());
	    if (cd) {
	      // cc is in d as cd
	      if (cc->tolerance () < cd->tolerance ()) {
		// cc is better than cd, so delete cd and combine it
		// with parents / top list
		if (!cd->collapse ()) {
		  // This means we delete d.
		  clist.clear ();
		  if (cd->sub_clusters ()) {
		    Cluster *sc;
		    while ((sc = cd->sub_clusters () -> unlink_first ())) {
		      sc->parent_cluster () = NULL;
		      lep->add_last (sc);
		    }
		    if (!dnext)
		      dnext = d->next ();
		  }
		  lep->del (cd);
		  if (cnext == d)
		    cnext = dnext;
		}
		delete cd;
	      } else {
		// cd is better than cc, so delete cc and combine it
		// with parents / top list
		if (!cc->collapse ()) {
		  // That means we delete c.
		  dnext = NULL;
		  clist.clear ();
		  if (cc->sub_clusters ()) {
		    Cluster *sc;
		    while ((sc = cc->sub_clusters () -> unlink_first ())) {
		      sc->parent_cluster () = NULL;
		      lep->add_last (sc);
		    }
		    if (!cnext)
		      cnext = c->next ();
		  }
		  lep->del (cc);
		}
		delete cc;
	      }
	    } else if (cc->sub_clusters ()) {
	      // Add the sub-clusters, as cc is not in d
	      Cluster *sc = cc->sub_clusters () -> first ();
	      while (sc) {
		clist.add_first (sc);
		sc = cc->sub_clusters () -> next ();
	      }
	    }


	  }

	}
	d = dnext;
      }
      c = cnext;
    }

    // NOTE: The ambiguity between line and plane equal could be
    //       removed, i.e. if plane equal then remove line equal if line is
    //       orthogonal to plane and there aren't more points in line equal.
    //       However, as we have to look for approximate orhtogonal, etc.
    //       this is not always exact and might remove some interesting
    //       relations, so we do not do it here (left for the hypothesizer).
    //       Note that especially the high tolerance level clusters could
    //       cause problems. Maybe there is a better way to eliminate
    //       those cases.
    
    if (pep->length () > 0)
      reg_list->add_first ((List<Regularity> *)pep);
    else
      delete pep;
    if (lep->length () > 0)
      reg_list->add_first ((List<Regularity> *)lep);
    else
      delete lep;
  }

  // NOTE: Very similarly to the axis grid and circle check, we could check
  //       if the position clusters are on one of the following structures:
  //       lines, 2D grids, 3D grids (with orthogonal/parallle directions
  //       and regular distances) and circles (with angle-regular arrangements)

  return;
}

void Analyse::check_plane_partially_equal (Cluster *dir, 
					   List<Cluster> *pts,
					   List<Cluster> *regs)
{
  // Project the point clusters on plane and create PartiallyPlaneEqual
  // features for the clusters.

  // Get the points.
  List<Cluster> bcl;
  ListNode<Cluster> *c = pts->first ();
  while (c) {
    c->data () -> get_all_base_clusters (&bcl);
    c = c->next ();
  }

  sil_progress_bar_text ("Clustering " + as_string (bcl.length ()) +
			 " 2d-partial positions for " +
			 as_string (*static_cast<Direction *> (dir->
							       value ())));

  // Project them on the plane and create the features for clustering.
  List<Feature> features;
  c = bcl.first ();
  while (c) {
    UnitVector3 d = *static_cast<Direction*> (dir->value ());
    Vector3 p = *static_cast<Position*> (c->data () -> value ());
    p = p - (p % d) * d;
    features.add_first (new PartiallyPlaneEqual (c->data (), p, dir));
    c = c->next ();
  }

  // Cluster the PartiallyPlaneEqual features.
  List<Cluster> pep;
  cluster (&features, tol_eq, tol_len_dist, NAN, true, &pep);

  // If cluster constains only points that are as close together as
  // their projectins, drop the cluster.

  // Start at base clusters
  bcl.clear ();
  c = pep.first ();
  while (c) {
    c->data () -> get_all_base_clusters (&bcl);
    c = c->next ();
  }
  // Check base-clusters and work upwards.
  while ((c = bcl.unlink_first ())) {
    // Get maximum distance between points
    Real dmax = -1.0;
    bool dmax_check = false;
    SimpleList<Feature> feat;
    SimpleList<PartiallyPlaneEqual> feat2;
    c->data () -> get_all_features (&feat);
    PartiallyPlaneEqual *f = static_cast<PartiallyPlaneEqual*> (feat.first ());
    while (f) {
      if (f->generators () . length () == 1) {
	Position *p = static_cast<Position*>
	  (static_cast<Cluster*> (f->generators () . first ()) -> value ());
	PartiallyPlaneEqual *f2 = feat2.first ();
	while (f2) {
	  Position *p2 = static_cast<Position*>
	    (static_cast<Cluster*> (f2->generators () . first ()) -> value ());
	  Real d = (*p2 - *p) . norm ();
	  if (d > dmax) {
	    dmax_check = true;
	    dmax = d;
	  }
	  f2 = feat2.next ();
	}
	feat2.add_first (f);
      } else {
	sil_msg_warn ("Analyse",
		      "partially plane equal feature without a single position generator");
      }
      f = static_cast<PartiallyPlaneEqual*> (feat.next ());
    }
    // Check maximum distance
    if (dmax_check &&
	dmax <= c->data () -> tolerance () + sil_get_epsilon ()) {
      // Points are too close, so collapse cluster
      Cluster *p = c->data () -> parent_cluster ();
      bool add_parent2 = (p && (!p->features () ||
				p->features () -> length () == 0));
      if (!c->data () -> collapse ()) {
	pep.del (c->data ());
      } else if (add_parent2) {
	bcl.add_last_check (p);
      }
      delete c->data ();      
    }
    delete c;
  }

  simplify_cluster_hierarchy (&pep, tol_len_dist, max_len / 2);

  // Delete clusters with a single feature.
  bcl.clear ();
  c = pep.first ();
  while (c) {
    bcl.add_first (c->data ());
    c = c->next ();
  }
  while ((c = bcl.unlink_first ())) {
    if (c->data () -> count_features () < 2) {
      if (!c->data () -> collapse ())
	pep.del (c->data ());
      delete c->data ();
    } else if (c->data () -> sub_clusters ()) {
      Cluster *sc = c->data () -> sub_clusters () -> first ();
      while (sc) {
	bcl.add_first (sc);
	sc = c->data () -> sub_clusters () -> next ();
      }
    }
    delete c;
  }

  regs->combine (&pep);

  return;
}

void Analyse::check_line_partially_equal (Cluster *dir, 
					  List<Cluster> *pts,
					  List<Cluster> *regs)
{
  // Project the point clusters on line and create PartiallyLineEqual
  // Features for the clusters.

  // Get the points.
  List<Cluster> bcl;
  ListNode<Cluster> *c = pts->first ();
  while (c) {
    c->data () -> get_all_base_clusters (&bcl);
    c = c->next ();
  }

  sil_progress_bar_text ("Clustering " + as_string (bcl.length ()) +
			 " 1d-partial positions for " +
			 as_string (*static_cast<Direction *> (dir->
							       value ())));

  // Project them on the line and create the features for clustering.
  List<Feature> features;
  c = bcl.first ();
  while (c) {
    UnitVector3 d = *static_cast<Direction*> (dir->value ());
    Vector3 p = *static_cast<Position*> (c->data () -> value ());
    p = (p % d) * d;
    features.add_first (new PartiallyLineEqual (c->data (), p, dir));
    c = c->next ();
  }

  // Cluster the PartiallyLineEqual features.
  List<Cluster> lep;
  cluster (&features, tol_eq, tol_len_dist, NAN, true, &lep);

  // If cluster constains only points that are as close together as
  // their projectins, drop the cluster.

  // Start at base clusters
  bcl.clear ();
  c = lep.first ();
  while (c) {
    c->data () -> get_all_base_clusters (&bcl);
    c = c->next ();
  }
  // Check base-clusters and work upwards.
  while ((c = bcl.unlink_first ())) {
    // Get maximum distance between points
    Real dmax = 0.0;
    SimpleList<Feature> feat;
    SimpleList<PartiallyLineEqual> feat2;
    c->data () -> get_all_features (&feat);
    PartiallyLineEqual *f = static_cast<PartiallyLineEqual*> (feat.first ());
    while (f) {
      if (f->generators () . length () == 1) {
	Position *p = static_cast<Position*>
	  (static_cast<Cluster*> (f->generators () . first ()) -> value ());
	PartiallyLineEqual *f2 = feat2.first ();
	while (f2) {
	  Position *p2 = static_cast<Position*>
	    (static_cast<Cluster*> (f2->generators () . first ()) -> value ());
	  Real d = (*p2 - *p) . norm ();
	  if (d > dmax)
	    dmax = d;
	  f2 = feat2.next ();
	}
	feat2.add_first (f);
      } else {
	sil_msg_warn ("Analyse",
		      "partially line equal feature without a single position generator");
      }
      f = static_cast<PartiallyLineEqual*> (feat.next ());
    }
    // Check maximum distance
    if (dmax <= c->data () -> tolerance () + sil_get_epsilon ()) {
      // Points are too close, so collapse cluster
      Cluster *p = c->data () -> parent_cluster ();
      bool add_parent2 = (p && (!p->features () ||
				p->features () -> length () == 0));
      if (!c->data () -> collapse ()) {
	lep.del (c->data ());
      } else if (add_parent2) {
	bcl.add_last_check (p);
      }
      delete c->data ();      
    }
    delete c;
  }

  simplify_cluster_hierarchy (&lep, tol_len_dist, max_len / 2);

  // Delete clusters with a single feature.
  bcl.clear ();
  c = lep.first ();
  while (c) {
    bcl.add_first (c->data ());
    c = c->next ();
  }
  while ((c = bcl.unlink_first ())) {
    if (c->data () -> count_features () < 2) {
      if (!c->data () -> collapse ())
	lep.del (c->data ());
      delete c->data ();
    } else if (c->data () -> sub_clusters ()) {
      Cluster *sc = c->data () -> sub_clusters () -> first ();
      while (sc) {
	bcl.add_first (sc);
	sc = c->data () -> sub_clusters () -> next ();
      }
    }
    delete c;
  }

  regs->combine (&lep);

  return;
}

Cluster *Analyse::find_equal_pe_cluster (Cluster *a, Cluster *b)
{
  // Find a cluster of partially equal features in tree b which is
  // caused by the same positions than the cluster of partially equal
  // feartures a.

  SimpleList<Entity> agens;
  a->get_all_generators (&agens);

  SimpleList<Cluster> bclus;
  bclus.add_first (b);
  Cluster *bc;
  while ((bc = bclus.unlink_first ())) {
    if (bc->count_features () == a->count_features ()) {
      // Feature number is equal, check if generators are as well.
      bool found = true;
      SimpleList<Entity> bgens;
      bc->get_all_generators (&bgens);
      if (bgens.length () == agens.length ()) {
	Entity *ag = agens.first ();
	while (ag) {
	  if (!bgens.is_member (ag)) {
	    found = false;
	    break;
	  }
	  ag = agens.next ();
	}
	if (found) {
	  // Sets are equal, report them
	  return bc;
	}
      }
      // Sets are not equal, but there might still be a sub-cluster
      if (bc->sub_clusters ()) {
	Cluster *sc = bc->sub_clusters () -> first ();
	while (sc) {
	  if (sc->count_features () >= a->count_features ())
	    bclus.add_first (sc);
	  sc = bc->sub_clusters () -> next ();
	}
      }
    } else if (bc->count_features () >= a->count_features () &&
	       bc->sub_clusters ()) {
      // Cluster bc can't be equal to a, but sub-clusters could. Add
      // those which have enough features to the list.
      Cluster *sc = bc->sub_clusters () -> first ();
      while (sc) {
	if (sc->count_features () >= a->count_features ())
	  bclus.add_first (sc);
	sc = bc->sub_clusters () -> next ();
      }
    }
    // Otherwise cluster has not enough features
  }
  return NULL;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
