// convert_acis.cc: ACIS to SiL conversion
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: convert_acis.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (convert_acis,
       "$Id: convert_acis.cc 1 2005-08-10 23:35:07Z frank $");

#include "interface.hxx"
#include "xstring.hxx"
#include "kernel/geometry.hxx"
#include "kernel/combinatorics.hxx"

#ifdef HAVE_ACIS

#include "acis/acis.hxx"
#include "acis/silapi.hxx"
#include <boolean/sg_husk/stitch/stitch.err>
#include <intersct/kernint/d3_ssi/ssi_inf.hxx>
#include <kernel/kernint/intsfsf/sfsfint.hxx>
#include <kernel/kernutil/d3_fn2/fn2.hxx>
#include <kernel/kerndata/geom/cnstruct.hxx>
#include <kernel/kerndata/geom/curve.hxx>
#include <kernel/kerngeom/curve/curdef.hxx>
#include <kernel/geomhusk/geom_utl.hxx>
#include <kernel/spline/bs3_crv/bs3curve.hxx>
#include <kernel/spline/sg_bs3c/sps3crtn.hxx>

extern double render_box_min;
extern double render_box_max;

unsigned long acis_face_count;

// ACIS vector to string.
string const as_string (vector const &v)
{
  return "[" + as_string (v.x ()) + ";" + as_string (v.y ()) + ";" + 
    as_string (v.z ()) + "]";
}

// ACIS position to string.
string const as_string (position const &p)
{
  return "[" + as_string (p.x ()) + ";" + as_string (p.y ()) + ";" + 
    as_string (p.z ()) + "]";
}

bool closed_edges_match (Edge *e1, EDGE *e2, transf const *trans,
			 bool &reversed_edge)
{
  // Check if geometry of closed edges e1 and e2 match.

  if (is_ELLIPSE (e2->geometry ())) {
    if (e1->curve () -> is (SIL_CIRCLE_TYPE)) {
      if (fabs (static_cast<ELLIPSE *> (e2->geometry ()) -> radius_ratio () -
		1.0) < sil_get_epsilon ()) {
	// Compare two circles
	Circle *c1 = static_cast<Circle *> (e1->curve ());
	unit_vector n1 = unit_vector (c1->normal () -> x (),
				      c1->normal () -> y (),
				      c1->normal () -> z ());
	position p1 = position (c1->centre () -> x (),
				c1->centre () -> y (),
				c1->centre () -> z ());
	Real r1 = c1->radius () -> value ();
	unit_vector n2 = static_cast<ELLIPSE *> (e2->geometry ()) -> normal ();
	if (trans)
	  n2 *= *trans;
	if (e2->sense () == REVERSED)
	  n2 = -n2;
	position p2 = static_cast<ELLIPSE *> (e2->geometry ()) -> centre ();
	if (trans)
	  p2 *= *trans;
	vector ma2 = static_cast<ELLIPSE *> (e2->geometry ()) -> major_axis ();
	if (trans)
	  ma2 *= *trans;
	Real r2 = ma2 . len ();
	Real n1n2 = n1 % n2;
	if (fabs (fabs (n1n2) - 1.0) < sil_get_epsilon () &&
	    (p1 - p2) . len () < sil_get_epsilon () &&
	    fabs (r1 - r2) < sil_get_epsilon ()) {
	  // Circles match. 
	  reversed_edge = (n1n2 < 0.0);
	  return true;
	}
      }
    } else if (e1->curve () -> is (SIL_ELLIPSE_TYPE)) {
      // Compare two ellipses
      Ellipse *ell = static_cast<Ellipse *> (e1->curve ());
      unit_vector n1 = unit_vector (ell->normal () -> x (),
	 			    ell->normal () -> y (),
		 		    ell->normal () -> z ());
      position c1 = position (ell->centre () -> x (),
	 		      ell->centre () -> y (),
		 	      ell->centre () -> z ());
      Real rr1 = ell->minor_radius () -> value () /
	ell->major_radius () -> value ();
      vector ma1 = vector (ell->major_direction () -> x (),
	 		   ell->major_direction () -> y (),
		 	   ell->major_direction () -> z ());
      Real r1 = ma1.len ();
      ma1 /= r1;
      unit_vector n2 = static_cast<ELLIPSE *> (e2->geometry ()) -> normal ();
      if (trans)
	n2 *= *trans;
      if (e2->sense () == REVERSED)
	n2 = -n2;
      position c2 = static_cast<ELLIPSE *> (e2->geometry ()) -> centre ();
      if (trans)
	c2 *= *trans;
      Real rr2 = static_cast<ELLIPSE *> (e2->geometry ()) -> radius_ratio ();
      vector ma2 = static_cast<ELLIPSE *> (e2->geometry ()) -> major_axis ();
      if (trans)
	ma2 *= *trans;
      Real r2 = ma2 . len ();
      ma2 /= r2;
      Real n1n2 = n1 % n2;
      if (fabs (fabs (n1n2) - 1.0) < sil_get_epsilon () &&
	  (c1 - c2) . len () < sil_get_epsilon () &&
	  fabs (fabs (ma1 % ma2) - 1.0) < sil_get_epsilon () &&
	  ((fabs (rr1 - rr2) < sil_get_epsilon () &&
	    fabs (r1 - r2) < sil_get_epsilon ()) ||
	   (fabs (r1 * rr1 - r2) < sil_get_epsilon () &&
	    fabs (r2 * rr2 - r1) < sil_get_epsilon ()))) {
	// Ellipses match.
	reversed_edge = (n1n2 < 0.0);
	return true;
      }
    }
  } else if (e1->curve () -> is (SIL_SPLINECURVE_TYPE) &&
	     is_INTCURVE (e2->geometry ())) {
    // Compare a spline curve with an intcurve.
 
    // Generate spline curve (ugly, but save way to ensure that it is
    // always compupted in the way ACIS does it).
    SplineCurve *sc = static_cast<SplineCurve *> (e1->curve ());
    position *ctrlpts = new position[sc->ctrlpts () -> length ()];
    Vector3 *cpt = sc->ctrlpts () -> first ();
    unsigned long k = 0;
    while (cpt) {
      ctrlpts[k++] = position (cpt->x (), cpt->y (), cpt->z ());
      cpt = sc->ctrlpts () -> next ();
    }
    bs3_curve bc = bs3_curve_from_ctrlpts (sc->degree (),
	 				   sc->rational (),
		 			   sc->closed (),
			 		   sc->periodic (),
				 	   sc->ctrlpts () -> length (),
					   ctrlpts,
 					   sc->weights (),
	 				   sc->ctrlpts_tol (),
		 			   sc->num_knots (),
			 		   sc->knots (),
				 	   sc->knot_tol (),
					   3);
    delete []ctrlpts;
    intcurve *c1 = new intcurve (bc, sc->ctrlpts_tol () , *(surface *)NULL_REF,
	 			 *(surface *)NULL_REF);

    // Check if e2 lies completely on e1 (c1).
    INTCURVE *C2 = static_cast<INTCURVE *> (e2->geometry ());
    // Parameter interval for 2nd curve.
    interval const I2 = C2->equation () .
      param_range (box (interval (render_box_min,
				  render_box_max),
			interval (render_box_min,
				  render_box_max),
			interval (render_box_min,
				  render_box_max)));
    double ps2 = I2.start_pt ();
    double pe2 = I2.end_pt ();
    double per2 = C2->equation () . param_period ();
    bool sense2 = (e2->sense () == REVERSED);
    if (fabs (per2) > get_resabs ()) {
      if (pe2 < ps2)
	sense2 = !sense2;
      pe2 = ps2 + per2;
    } else if (pe2 < ps2) {
      sense2 = !sense2;
      double h = pe2;
      pe2 = ps2;
      ps2 = h;
    }
    // Comparison step size.
    double l = C2->equation () . length (ps2, pe2) /
      (4.0 * sc->ctrlpts () -> length ());
    if (l < 2.0 * get_resabs ())
      l = 2.0 * get_resabs ();
    double step = (pe2 - ps2) / 4.0;
    double length = C2->equation () . length (ps2, ps2 + step);
    while (length > l && step > 2.0 * get_resabs ()) {
      step /= 2.0;
      length = C2->equation () . length (ps2, ps2 + step);
    }
    // Compare the two curves.
    double err = c1->approx_error ();
    double e2 = C2->equation () . approx_error ();
    if (e2 > err)
      err = e2;
    position P1, P2;
    for (double p = ps2; p <= pe2; p += step) {
      C2->equation () . eval (p, P2);
      if (trans)
	P2 *= *trans;
      c1->closest_point (P2, P1);
      if ((P2 - P1) . len () >= err) {
	delete c1;
	return false;
      }
    }

    // Curves match.

    // Check orientation.

    // Parameter interval for other curve.
    interval const I1 = c1->param_range (box (interval (render_box_min,
							render_box_max),
					      interval (render_box_min,
							render_box_max),
					      interval (render_box_min,
							render_box_max)));
    double ps1 = I1.start_pt ();
    double pe1 = I1.end_pt ();
    double per1 = c1->param_period ();
    bool sense1 = false;
    if (fabs (per1) > get_resabs ()) {
      pe1 = ps1 + per1;
      if (pe1 < ps1)
	sense1 = !sense1;
    } else if (pe1 < ps1) {
	double h = pe1;
	pe1 = ps1;
	ps1 = h;
	sense1 = !sense1;
    }
    // If the start points with repsect to sense1 and sense2 are the
    // same then the orientation of the edge is not reversed
    if (sense1)
      c1->eval (pe1, P1);
    else
      c1->eval (ps1, P1);
    if (sense2)
      C2->equation () . eval (pe2, P2);
    else
      C2->equation () . eval (ps2, P2);
    reversed_edge = !((P1 - P2) . len () < sil_get_epsilon ());

    delete c1;
    
    return true;
  }

  return false;
}

bool edges_match (Edge *e1, EDGE *e2, transf const *trans, bool &reversed_edge)
{
  // Check if geometry and vertices of e1 and e2 match (edges are not
  // closed) and the same part of the geometry is used for the
  // edge. Set reversed_edge to true if stat and end point of the two
  // edges are reversed.

  if (e1->geometry () -> is (SIL_CIRCLE_TYPE)) {
    if (is_STRAIGHT (e2->geometry ())) {
      // Compare two straight edges.
      position p1 = e2->start () -> geometry () -> coords ();
      position p2 = e2->end () -> geometry () -> coords ();
      if (trans) {
	p1 *= *trans;
	p2 *= *trans;
      }
      Position *P = e1->start_vertex () -> point () -> position ();
      position pp1 = position (P->x (), P->y (), P->z ());
      P = e1->end_vertex () -> point () -> position ();
      position pp2 = position (P->x (), P->y (), P->z ());
      if ((p1 - pp1) . len () < sil_get_epsilon () &&
	  (p2 - pp2) . len () < sil_get_epsilon ()) {
	reversed_edge = false;
	return true;
      } else  if ((p1 - pp2) . len () < sil_get_epsilon () &&
		  (p2 - pp1) . len () < sil_get_epsilon ()) {
	reversed_edge = true;
	return true;
      }
      return false;
    }
  } else if (is_ELLIPSE (e2->geometry ())) {
    if (e1->geometry () -> is (SIL_CIRCLE_TYPE) &&
  	fabs (static_cast<ELLIPSE *> (e2->geometry ()) ->
	      radius_ratio () - 1.0) < sil_get_epsilon ()) {
      // FIXME: compare two circles and check for reversed_edge.
      //        note that the edge has to be the same, not only the
      //        geometry.
    } else if (e1->geometry () -> is (SIL_ELLIPSE_TYPE)) {
      // FIXME: compare two ellipses and check for reversed_edge.
      //        note that the edge has to be the same, not only the
      //        geometry.
    }
  } else if (e1->geometry () -> is (SIL_SPLINECURVE_TYPE) && 
	     is_INTCURVE (e2->geometry ())) {
    // FIXME: compare spline curve with intcurve and check for reversed_edge.
    //        note that the edge has to be the same, not only the
    //        geometry.
  }
  return false;
}
  
Edge *match_edge (EDGE *E, Solid *solid, transf const *trans,
		  bool &reversed_edge)
{
  // Check if there is another edge in the model which is identical to
  // this one and if so, use that edge.
  
  // Only do this for edges with vertices (ACIS edges without vertices
  // appear to be partly undefined).
  if (!E->start () || !E->end ())
    return NULL;
 
  if (E->start () == E->end () ||
      (E->start () -> geometry () -> coords () - 
       E->end () -> geometry () -> coords ()) . len () < sil_get_epsilon ()) {
    // Closed edge.
    ListNode<Edge> *n = solid->edges () . first ();
    while (n) {
      if (!n->data () -> start_vertex () &&
	  !n->data () -> end_vertex () &&
	  closed_edges_match (n->data (), E, trans, reversed_edge)) {
	// Found another clsoed edge which matches this one.
	return n->data ();
      }
      n = n->next ();
    }
    return NULL;
  }

  ListNode<Edge> *n = solid->edges () . first ();
  while (n) {
    if (n->data () -> start_vertex () &&
	n->data () -> end_vertex () &&
	edges_match (n->data (), E, trans, reversed_edge)) {
      // Found another non-closed edge which mathes this one
      return n->data ();
    }
    n = n->next ();
  }
  return NULL;
}

Curve *convert_acis_curve (CURVE *C, transf const *trans, REVBIT sense)
{
  curve *c = C->equation () . copy_curve ();
  if (trans)
    *c *= *trans;
  if (sense == REVERSED)
    c->negate ();

  bool spline_straight = false;
  bool spline_planar = false;
  unit_vector n;
  if (is_intcurve (c)) {
    // Check if geometry of curve can be simplified.
    bs3_curve bc = static_cast<intcurve *> (c) -> cur ();
    if (bc) {
      int type = bs3_curve_planar (bc, n);
      // -2 = Curve is a point or a degenerate spline. 
      // -1 = Curve is a straight line, normal not set. 
      //  0 = Curve is nonplanar. 
      //  1 = Curve is planar, normal is set
      if (type == -1)
	spline_straight = true;
      else if (type == 1) {
	spline_planar = true;
	// FIXME: check if spline curve is circle or ellipse [based on 
	//        curve intersections if not free].
      }
    }
  }

  if (spline_straight || is_STRAIGHT (C)) {
    delete c;
    sil_progress_bar_text ("Straight curve");
    return new Straight ();
  } else if (is_ELLIPSE (C)) {
    ellipse *e = static_cast<ellipse *> (c);
    vector n = e->normal;
    position p = e->centre;
    vector ma = e->major_axis;
    if (fabs (e->radius_ratio - 1.0) < sil_get_epsilon ()) {
      Circle *cc = new Circle ();
      cc->normal () -> set (n.x (), n.y (), n.z ());
      cc->centre () -> set (p.x (), p.y (), p.z ());
      cc->radius () -> value () = ma.len ();
      sil_progress_bar_text ("Circle: " +
			     as_string (cc->radius () -> value ()) +
			     ", " + as_string (n) + ", " + as_string (p));
      delete c;
      return cc;
    } else {
      Ellipse *ee = new Ellipse ();
      ee->normal () -> set (n.x (), n.y (), n.z ());
      ee->centre () -> set (p.x (), p.y (), p.z ());
      ee->major_direction () -> set (ma.x (), ma.y (), ma.z ());
      ee->major_radius () -> value () = ma.len ();
      ee->minor_radius () -> value () = (ma.len () * e->radius_ratio);
      sil_progress_bar_text ("Ellipse: " +
			     as_string (ee->major_radius () -> value ()) +
			     ", " +
			     as_string (ee->minor_radius () -> value ()) +
			     ", " + as_string (n) + ", " + as_string (p) +
			     ", " + as_string (ma));
      delete c;
      return ee;
    }
  } else if (is_INTCURVE (C)) {
    bs3_curve bc = static_cast<intcurve *> (c) -> cur ();
    if (bc) {
      int num_pts = 0;
      position *ctrlpts = NULL;
      bs3_curve_control_points (bc, num_pts, ctrlpts);
      SimpleList<Vector3> *cpts = new SimpleList<Vector3>;
      for (int k = 0; k < num_pts; ++k)
	cpts->add_last (new Vector3 (ctrlpts[k] . x (), ctrlpts[k] . y (),
				     ctrlpts[k] . z ()));
      delete []ctrlpts;
      int num_w = 0;
      double *weights = NULL;
      bs3_curve_weights (bc, num_w, weights);
      int num_knots = 0;
      double *knots = NULL;
      bs3_curve_knots (bc, num_knots, knots);
      SplineCurve *sp = new SplineCurve ();
      sp->define (bs3_curve_degree (bc),
		  bs3_curve_rational (bc),
		  bs3_curve_closed (bc),
		  bs3_curve_periodic (bc),
		  cpts, weights,
		  get_resabs (),
		  num_knots, knots,
		  bs3_curve_knottol ());
      if (spline_planar) {
	sp->orientation () . set (n.x (), n.y (), n.z ());
	// FIXME: is orientation correctly returned by bs3_curve_planar?
	sp->set_plane_normal (UnitVector3 (n.x (), n.y (), n.z ()));
      }
      // FIXME: set an orientation if curve is not planar.
      sil_progress_bar_text ("Spline Curve");
      delete c;
      return sp;
    }
  }

  sil_progress_bar_text ("Unkown ACIS Curve");
  sil_msg_warn ("convert", "unkown acis curve");

  delete c;

  return new Curve ();
}

Surface *convert_acis_surface (SURFACE *S, transf const *trans, REVBIT sense)
{
  surface *s;
  if (trans) {
    s = S->trans_surface (*trans, sense);
  } else {
    s = S->equation () . copy_surf ();
    if (sense == REVERSED)
      s->negate ();
  }
  if (is_PLANE (S)) {
    plane *P = static_cast<plane *> (s);
    Plane *p = new Plane ();
    vector n = P->normal;
    p->direction () -> set (n.x (), n.y(), n. z());
    position rp = P->root_point;
    p->root_point () -> set (rp.x (), rp.y (), rp.z ());
    sil_progress_bar_text ("Plane: " + as_string (n) + ", " + 
			   as_string (rp));
    delete s;
    return p;
  } else if (is_CONE (S)) {
    cone *C = static_cast<cone *> (s);
    if (C->circular () && !C->flat ()) {
      // The signs of the trigonometric functions tell us which way the
      // cone is enlarging, and whether we are interested in the inside or
      // outside of the surface. The rules are:
      //              if cosine_angle is +ve, the surface direction is
      //                 outwards, away from the axis, so the surface is 
      //                 convex. Conversely if cosine_angle is negative.
      //              if sine_angle and cosine_angle have the same sign, then
      //                 the cone increases in cross-section in the direction
      //                 of the axis (the normal to the base ellipse), and
      //                 conversely if they have opposite sign.
      if (C->cylinder ()) {
	Cylinder *c = new Cylinder ();
	c->radius () -> value () = (C->base.major_axis.len ());
	vector n = C->base.normal;
	c->direction () -> set (n.x (), n.y (), n.z ());
	position rp = C->base.centre;
	c->root_point () -> set (rp.x (), rp.y (), rp.z ());
	c->convex () = (C->cosine_angle > 0.0);
	sil_progress_bar_text ("Cylinder: " + as_string (n) + ", " +
			       as_string (rp) + ", " + 
			       (c->convex () ? "*":"o"));
	delete s;
	return c;
      } else {
	Cone *c = new Cone ();
	c->angle () -> value () = atan (fabs (C->sine_angle / 
					      C->cosine_angle));
	unit_vector n = C->base.normal;
	if (C->sine_angle * C->cosine_angle < 0.0)
	  n = -n;
	c->direction () -> set (n.x (), n.y (), n.z ());
	position rp = C->get_apex ();
	c->root_point () -> set (rp.x (), rp.y(), rp.z ());
	c->convex () = (C->cosine_angle > 0.0);
	sil_progress_bar_text ("Cone: " + 
			       as_string (c->angle () -> value () / PI * 180) +
			       " deg, " + as_string (n) + ", " + 
			       as_string (rp) + ", " +
			       (c->convex ()?"*":"o"));
	delete s;
	return c;
      }
    }
  } else if (is_SPHERE (S)) {
    sphere *SS = static_cast<sphere *> (s);
    Sphere *ss = new Sphere ();
    ss->radius () -> value () = fabs (SS->radius);
    position rp = SS->centre;
    ss->root_point () -> set (rp.x (), rp.y (), rp.z ());
    ss->convex () = (SS->radius > 0.0);
    sil_progress_bar_text ("Sphere: " + as_string (ss->radius () -> value ()) +
			   ", " + as_string (rp) + ", " +
			   (ss->convex () ? "*":"o"));
    delete s;
    return ss;
  } else if (is_TORUS (S)) {
    torus *T = static_cast<torus *> (s);
    Torus *t = new Torus ();
    // Radius of the spine, normally positive, but negative (and
    // smaller in magnitude than the minor radius) for the inner
    // portion (i.e. a "lemon", as opposed to an "apple");
    t->major_radius () -> value () = fabs (T->major_radius);
    // Radius of cross-sectional circle of the ring, negated if torus
    // is a void.  Exactly zero to indicate "undefined".
    t->minor_radius () -> value () = fabs (T->minor_radius);
    t->apple () = (T->major_radius > 0.0);
    t->convex () = (T->minor_radius > 0.0);
    position rp = T->centre;
    t->root_point () -> set (rp.x (), rp.y (), rp.z ());
    vector n = T->normal;
    t->direction () -> set (n.x (), n.y (), n.z ());
    sil_progress_bar_text ("Torus: " + 
			   as_string (t->major_radius () -> value ()) + 
			   ", " +
			   as_string (t->minor_radius () -> value ()) +
			   ", " + as_string (n) + ", " + as_string (rp) + 
			   ", " +
			   (t->convex () ? "*":"o"));
    delete s;
    return t;
  }

  sil_progress_bar_text ("Unkown ACIS Surface");
  sil_msg_warn ("convert", "unkown acis surface");

  delete s;

  return new Surface ();
}

Vertex *convert_acis_vertex (VERTEX *V, Solid *solid, transf const *trans)
{
  if (!V)
    return NULL;

  // Check if vertex is already converted.
  ListNode<Vertex> *n = solid->vertices () . first ();
  while (n) {
    if (n->data () -> reference () == V)
      return n->data ();
    else
      n = n->next ();
  }

  // Create new vertex.
  Vertex *v = new Vertex;
  v->reference () = static_cast<void *> (V);
  solid->vertices () . add_first (v);

  // Convert geometry.
  position p = V->geometry () -> coords ();
  if (trans)
    p *= *trans;
  Point *P = new Point ();
  P->position () -> set (p.x (), p.y (), p.z ());
  v->set_point (P);

  return v;
}

Edge *convert_acis_edge (EDGE *E, Solid *solid, transf const *trans,
			 bool &reversed_edge)
{
  if (!E) {
    sil_msg_error ("convert", "undefined ACIS edge.");
    return NULL;
  }

  if (sil_cancel_request ())
    return NULL;

  // Check if edge is already converted.
  ListNode<Edge> *n = solid->edges () . first ();
  while (n) {
    if (n->data () -> reference () == E) {
      return n->data ();
    } else {
      n = n->next ();
    }
  }

  Edge *e = match_edge (E, solid, trans, reversed_edge);
  if (e)
    return e;

  // Create new edge.
  e = new Edge ();
  e->reference () = static_cast<void *> (E);
  solid->edges () . add_first (e);
  reversed_edge = false;

  // Convert geometry.
  e->set_curve (convert_acis_curve (E->geometry (), trans, E->sense ()));

  if (!E->start () || !E->end () ||
      (E->start () != E->end () &&
       (E->start () -> geometry () -> coords () - 
	E->end () -> geometry () -> coords ()) . len () >=
       sil_get_epsilon ())) {
    // Edge is not closed (or some vertices are not defined), so try
    // to add (existing) vertices.
    e->set_start_vertex (convert_acis_vertex (E->start (), solid, trans));
    e->set_end_vertex (convert_acis_vertex (E->end (), solid, trans));
  }

  return e;
}

Loop *convert_acis_loop (LOOP *L, Solid *solid, transf const *trans)
{
  if (!L || sil_cancel_request ())
    return NULL;

  sil_progress_bar_text ("Converting ACIS LOOP");

  // Create new loop.
  Loop *l = new Loop ();
  solid->loops () . add_first (l);

  COEDGE *C = L->start ();
  if (C) {
    // The first Edge in the Loop list decides about the direction the
    // Loop is traversed (the rest is defined by the sequence of edges in
    // the list). If the first is reversed in the ACIS model, we have to
    // set sense to true.
    if (C->sense () == REVERSED)
      l->sense () = true;
  }
  bool first_edge = true;
  COEDGE *FC = C;
  while (C) {
    bool reversed_edge = false;
    Edge *e = convert_acis_edge (C->edge (), solid, trans, reversed_edge);
    if (e) {
      if (first_edge) {
	first_edge = false;
	if (reversed_edge)
	  l->sense () = !l->sense ();
      }
      l->edges () . add_last (e);
      C = C->next ();
      if (FC == C)
	C = NULL;
    } else {
      delete l;
      return NULL;
    }
  }

  return l;
}

Face *convert_acis_face (FACE *F, Solid *solid, transf const *trans)
{
  if (!F || sil_cancel_request ())
    return NULL;

  sil_progress_bar_text ("Converting ACIS FACE");

  // Check if face was already converted.
  ListNode<Face> *n = solid->faces () . first ();
  while (n) {
    if (n->data () -> reference () == F) {
      return n->data ();
    } else {
      n = n->next ();
    }
  }

  // Create new face.
  Face *f = new Face ();
  f->reference () = static_cast<void *> (F);
  solid->faces () . add_first (f);

  sil_progress_bar_percentage (static_cast<Real> (solid->faces () .
						  length ()) /
			       static_cast<Real> (acis_face_count));

  // Convert geometry.
  f->set_surface (convert_acis_surface (F->geometry (), trans, F->sense ()));

  LOOP *L = F->loop ();
  LOOP *FL = L;
  while (L) {
    Loop * l = convert_acis_loop (L, solid, trans);
    if (l) {
      f->loops () . add_first (l);
      L = L->next ();
      if (FL == L)
	L = NULL;
    } else {
      delete f;
      return NULL;
    }
  }

  return f;
}

Shell *convert_acis_shell (SHELL *S, Solid *solid, transf const *trans)
{
  if (!S || sil_cancel_request ())
    return NULL;

  sil_progress_bar_text ("Converting ACIS SHELL");

  // Create new shell.
  Shell *s = new Shell ();
  solid->shells () . add_first (s);

  FACE *F = S->face ();
  FACE *FF = F;
  while (F) {
    Face *f = convert_acis_face (F, solid, trans);
    if (f) {
      s->faces () . add_first (f);
      F = F->next ();
      if (FF == F)
	F = NULL;
    } else {
      delete s;
      return NULL;
    }
  }

  return s;
}

Lump *convert_acis_lump (LUMP *L, Solid *solid, transf const *trans)
{
  if (!L || sil_cancel_request ())
    return NULL;

  // Check if we already processed the lump.
  XListNode<Lump> *n = solid->lumps () . first ();
  while (n) {
    if (n->data () -> reference () == L) {
      return n->data ();
    }
    n = n->next ();
  }

  sil_progress_bar_text ("Converting ACIS LUMP");

  // Create a new lump.
  Lump *l = new Lump;
  l->reference () = static_cast<void *> (L);

  // Add geometry.
  l->set_space (new Space ());

  SHELL *S = L->shell ();
  SHELL *FS = S;
  while (S) {
    if (S->wire_list ())
      sil_msg_warn ("convert", "ignoring shell wires");
    Shell *s = convert_acis_shell (S, solid, trans);
    if (s) {
      l->shells () . add_first (s);
      S = S->next ();
      if (FS == S)
	S = NULL;
    } else {
      delete l;
      return NULL;
    }
  }

  return l;
}

Solid *sil_convert (BODY *B, bool refs)
{
  if (!B)
    return NULL;
  
  sil_progress_bar_text ("Converting ACIS BODY");

  Solid *solid = new Solid ();

  if (B->wire ())
    sil_msg_warn ("convert", "ignoring body wires");

  ENTITY_LIST face_list;
  sil_acis_check (api_get_faces (B, face_list),
		  "sil_convert", "api_get_faces");
  acis_face_count = face_list.count ();

  LUMP *L = B->lump ();
  LUMP *FL = L;
  while (L) {
    Lump *l =convert_acis_lump (L, solid, B->transform () ? 
				&B->transform () -> transform () : NULL);
    if (l) {
      solid->lumps () . add_first (l);
    } else{
      delete solid;
      return NULL;
    }
    L = L->next ();
    if (FL == L)
      L = NULL;
  }

  if (!refs) {
    // Remove references on request.
    XListNode<Lump> *ln = solid->lumps () . first ();
    while (ln) {
      ln->data () -> reference () = NULL;
      ln = ln->next ();
    }
    ListNode<Face> *fn = solid->faces () . first ();
    while (fn) {
      fn->data () -> reference () = NULL;
      fn = fn->next ();
    }
    ListNode<Edge> *en = solid->edges () . first ();
    while (en) {
      en->data () -> reference () = NULL;
      en = en->next ();
    }
    ListNode<Vertex> *vn = solid->vertices () . first ();
    while (vn) {
      vn->data () -> reference () = NULL;
      vn = vn->next ();
    }
  }

  return solid;
}

#endif // HAVE_ACIS

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
