// features.hxx: solid model features.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: features.hxx,v 1.17.2.1 2001/08/16 16:19:09 scmfcl Exp $

#ifndef FEATURES_HXX
#define FEATURES_HXX 1

#include "xmath.hxx"
#include "kernel/entity.hxx"
#include "kernel/vector.hxx"

class Cluster;
class SubRegularity;

// Feature of a solid model describing the shape of a cell and its
// position and orientation in space without topological and
// inside/outside orientation related information.
class Feature : public Entity {
public:
  Feature (Entity *g) : Entity () {
    Entity::type (1, SIL_FEATURE_TYPE);
    if (g)
      gen.add_first (g);
    d = 0;
    fn = 0;
  }
  virtual ~Feature () {
  }
  virtual string const type_id () const {
    return tid;
  }
  // Entities responsible for generating the feature
  SimpleList<Entity> &generators () {
    return gen;
  }
  SimpleList<Entity> generators () const {
    return gen;
  }
  // Feature id for the generator.
  unsigned long &id () {
    return d;
  }
  unsigned long id () const {
    return d;
  }
  virtual void hilite (bool hi, string const &name);
  virtual string const description () const = 0;
  virtual Real sim (Feature const *f) const = 0;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const = 0;
  virtual Feature *copy (Entity *g) const = 0;
  virtual bool regular () const {
    return false;
  }
  virtual unsigned long &count_indices () {
    return fn;
  }
  virtual unsigned long count_indices () const {
    return fn;
  }
  virtual unsigned long *indices () {
    return fi;
  }
private:
  static string const tid;
  SimpleList<Entity> gen;
  unsigned long d, fn, fi[2];
};

// Direction feature.
class Direction : public Feature, public UnitVector3 {
public:
  Direction (Entity *g) : Feature (g), UnitVector3 () {
    Entity::type (2, SIL_DIRECTION_TYPE);
  }
  Direction (Entity *g, Vector3 const &v) : Feature (g), UnitVector3 (v) {
    Entity::type (2, SIL_DIRECTION_TYPE);
  }
  Direction (Entity *g, UnitVector3 const &v) : Feature (g), UnitVector3 (v) {
    Entity::type (2, SIL_DIRECTION_TYPE);
  }
  virtual ~Direction () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy (Entity *g) const {
    return new Direction (g, static_cast<UnitVector3> (*this));
  }
private:
  static string const tid;
};

// Position feature.
class Position : public Feature, public Vector3 {
public:
  Position (Entity *g) : Feature (g), Vector3 () {
    Entity::type (2, SIL_POSITION_TYPE);
  }
  Position (Entity *g, Vector3 const &p) : Feature (g), Vector3 (p) {
    Entity::type (2, SIL_POSITION_TYPE);
  }
  virtual ~Position () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy (Entity *g) const {
    return new Position (g, static_cast<Vector3> (*this));
  }
private:
  static string const tid;
};

// Abstract scalar feature.
class Scalar : public Feature {
public:
  explicit Scalar (Entity *g, Real v) : Feature (g) {
    Entity::type (2, SIL_SCALAR_TYPE);
    val = v;
  }
  virtual ~Scalar () {
  }
  virtual string const type_id () const {
    return tid;
  }
  Real value () const {
    return val;
  }
  Real &value () {
    return val;
  }
private:
  static string const tid;
  Real val;
};

// Angle feature.
class Angle : public Scalar {
public:
  explicit Angle (Entity *g, Real v = NAN) : Scalar (g, v) {
    Entity::type (3, SIL_ANGLE_TYPE);
  }
  virtual ~Angle () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy (Entity *g) const {
    return new Angle (g, value ());
  }
private:
  static string const tid;
};

// Lenght feature.
class Length : public Scalar {
public:
  explicit Length (Entity *g, Real v = NAN) : Scalar (g, v) {
    Entity::type (3, SIL_LENGTH_TYPE);
  }
  virtual ~Length () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy (Entity *g) const {
    return new Length (g, value ());
  }
private:
  static string const tid;
};

// Direction plane feature.
class DirectionPlane : public Direction {
public:
  DirectionPlane (Entity *g) : Direction (g) {
    Entity::type (3, SIL_DIRECTIONPLANE_TYPE);
  }
  DirectionPlane (Entity *g, Vector3 const &v) : Direction (g, v) {
    Entity::type (3, SIL_DIRECTIONPLANE_TYPE);
  }
  DirectionPlane (Entity *g, UnitVector3 const &v) : Direction (g, v) {
    Entity::type (3, SIL_DIRECTIONPLANE_TYPE);
  }
  virtual ~DirectionPlane () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy (Entity *g) const {
    return new DirectionPlane (g, static_cast<UnitVector3> (*this));
  }
private:
  static string const tid;
};

// Direction cone feature.
class DirectionCone : public Direction {
public:
  DirectionCone (Entity *g) : Direction (g) {
    Entity::type (3, SIL_DIRECTIONCONE_TYPE);
    angl = NAN;
    reg = false;
  }
  DirectionCone (Entity *g, Vector3 const &v, Real an) : Direction (g, v) {
    Entity::type (3, SIL_DIRECTIONCONE_TYPE);
    angl = an;
    reg = false;
  }
  DirectionCone (Entity *g, UnitVector3 const &v, Real an) : Direction (g, v) {
    Entity::type (3, SIL_DIRECTIONCONE_TYPE);
    angl = an;
    reg = false;
  }
  virtual ~DirectionCone () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy (Entity *g) const {
    return new DirectionCone (g, static_cast<UnitVector3> (*this), angl);
  }
  virtual bool regular () const {
    return reg;
  }
  bool &regular () {
    return reg;
  }
  Real &angle () {
    return angl;
  }
  Real angle () const {
    return angl;
  }
private:
  static string const tid;
  Real angl;
  bool reg;
};

// Intersection position feature.
class IntersectionPosition : public Position {
public:
  IntersectionPosition (Entity *g) : Position (g) {
    Entity::type (3, SIL_INTERSECTIONPOSITION_TYPE);
    t = NAN;
  }
  IntersectionPosition (Entity *g, Vector3 const &p,
			Real tol) : Position (g, p) {
    Entity::type (3, SIL_INTERSECTIONPOSITION_TYPE);
    t = tol;
  }
  virtual ~IntersectionPosition () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy (Entity *g) const {
    return new IntersectionPosition (g, static_cast<Vector3> (*this), t);
  }
  Real &intersection_tolerance () {
    return t;
  }
  Real intersection_tolerance () const {
    return t;
  }
private:
  static string const tid;
  Real t;
};

// Aligned axes feature.
class AlignedAxes : public Position {
public:
  AlignedAxes (Entity *g) : Position (g) {
    Entity::type (3, SIL_ALIGNEDAXES_TYPE);
    f = false;
    dir = NULL;
  }
  AlignedAxes (Entity *g, Position *p, Cluster *d) : Position (g, *p) {
    Entity::type (3, SIL_ALIGNEDAXES_TYPE);
    dir = d;
    pf = p;
    f = false;
  }
  AlignedAxes (Entity *g, Vector3 const &v, Cluster *d) : Position (g, v) {
    Entity::type (3, SIL_ALIGNEDAXES_TYPE);
    dir = d;
    pf = NULL;
    f = false;
  }
  virtual ~AlignedAxes () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy (Entity *g) const {
    return new AlignedAxes (g, static_cast<Vector3> (*this), dir);
  }
  Cluster *direction () const {
    return dir;
  }
  Position *position_feature () const {
    return pf;
  }
  bool flag () const {
    return f;
  }
  bool &flag () {
    return f;
  }
private:
  static string const tid;
  Cluster *dir;
  Position *pf;
  bool f;
};

// Parallel line feature.
class ParallelLine : public Feature, public Vector3 {
public:
  ParallelLine (Entity *g) : Feature (g), Vector3 () {
    Entity::type (2, SIL_PARALLELLINE_TYPE);
  }
  ParallelLine (Entity *g, Vector3 const &d) : Feature (g), Vector3 (d) {
    Entity::type (2, SIL_PARALLELLINE_TYPE);
  }
  virtual ~ParallelLine () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy (Entity *g) const {
    return new ParallelLine (g, static_cast<Vector3> (*this));
  }
private:
  static string const tid;
};

// Axis circle feature.
class PointCircle : public Feature {
public:
  PointCircle (Entity *g) : Feature (g) {
    Entity::type (2, SIL_POINTCIRCLE_TYPE);
    c = Vector3 (0.0, 0.0, 0.0);
    r = 0.0;
    reg = false;
  }
  PointCircle (Entity *g, Vector3 const &cc, Real rr) : Feature (g) {
    Entity::type (2, SIL_POINTCIRCLE_TYPE);
    c = cc;
    r = rr;
    reg = false;
  }
  virtual ~PointCircle () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy (Entity *g) const {
    return new PointCircle (g, c, r);
  }
  virtual bool regular () const {
    return reg;
  }
  bool &regular () {
    return reg;
  }
  Vector3 &centre () {
    return c;
  }
  Vector3 centre () const {
    return c;
  }
  Real &radius () {
    return r;
  }
  Real radius () const {
    return r;
  }
private:
  static string const tid;
  Vector3 c;
  Real r;
  bool reg;
};

// Partially equal with respect to a plane
class PartiallyPlaneEqual : public Position {
public:
  PartiallyPlaneEqual (Entity *g) : Position (g) {
    Entity::type (3, SIL_PARTIALLYPLANEEQUAL_TYPE);
    d = NULL;
  }
  PartiallyPlaneEqual (Entity *g, Vector3 const &p, 
		       Cluster *dir) : Position (g, p) {
    Entity::type (3, SIL_PARTIALLYPLANEEQUAL_TYPE);
    d = dir;
  }
  virtual ~PartiallyPlaneEqual () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy (Entity *g) const {
    return new PartiallyPlaneEqual (g, *static_cast<Vector3 const*> 
				    (static_cast<Position const*> (this)), d);
  }
  Cluster *direction () const {
    return d;
  }
private:
  static string const tid;
  Cluster *d;
};

// Partially equal with respect to a lines
class PartiallyLineEqual : public Position {
public:
  PartiallyLineEqual (Entity *g) : Position (g) {
    Entity::type (3, SIL_PARTIALLYLINEEQUAL_TYPE);
    d = NULL;
  }
  PartiallyLineEqual (Entity *g, Vector3 const &p,
		      Cluster *dir) : Position (g, p) {
    Entity::type (3, SIL_PARTIALLYLINEEQUAL_TYPE);
    d = dir;
  }
  virtual ~PartiallyLineEqual () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy (Entity *g) const {
    return new PartiallyLineEqual (g, *static_cast<Vector3 const *> 
				   (static_cast<Position const *> (this)),
				   d);
  }
  Cluster *direction () const {
    return d;
  }
private:
  static string const tid;
  Cluster *d;
};

// Loop polygon feature.
class LoopPolygon : public Feature {
public:
  LoopPolygon (Entity *g) : Feature (g) {
    Entity::type (2, SIL_LOOPPOLYGON_TYPE);
    pts = NULL;
    a = b = NULL;
    num_coeff = 0;
  }
  LoopPolygon (Entity *g, SimpleXList<Position> *points) : Feature (g) {
    Entity::type (2, SIL_LOOPPOLYGON_TYPE);
    pts = points;
    a = b = NULL;
    num_coeff = 0;
  }
  virtual ~LoopPolygon () {
    if (pts)
      delete pts;
    if (a)
      delete a;
    if (b)
      delete b;
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy (Entity *g) const;
  void set (SimpleXList<Position> *points);
  void ft (unsigned long nc);
private:
  static string const tid;
  SimpleXList<Position> *pts;
  Real *a, *b;
  unsigned long num_coeff;
};

#endif // !FEATURES_HXX

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
