// kernel.cxx: Kernel module implementation.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: kernel.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (kernel, "$Id: kernel.cc 1 2005-08-10 23:35:07Z frank $");

#include "kernel/kernel_module.hxx"
#include "kernel/kernel.hxx"
#include "kernel/entity.hxx"
#include "kernel/combinatorics.hxx"
#include "kernel/memory.hxx"
#include "kernel/analyse.hxx"

#include "interface.hxx"

// For f2c library.
extern "C" {
  int (*MAIN__) (int, char **) = NULL;
}

SIL_MODULE ()
{
  return new KernelModule;
}

const char *KernelModule::nm = "kernel";

KernelModule::KernelModule ()
{
  MM = new MemoryManager;
  drawn = NULL;
}

KernelModule::~KernelModule ()
{
  if (MM)
    delete MM;
  if (drawn)
    delete drawn;
}

void KernelModule::add_symbol (void *object, char *name, bool visible)
{
  take (static_cast<Entity *> (object));
}

void KernelModule::show (void *object)
{
  Entity *e = (Entity *)object;
  e->visible () = true;
  return;
}

void KernelModule::hide (void *object)
{
  Entity *e = (Entity *)object;
  e->visible () = false;
  return;
}

void KernelModule::facet_init ()
{
  return;
}

void KernelModule::draw_init ()
{
  if (drawn != NULL)
    delete drawn;
  drawn = new SimpleList<Entity>;

  double surf_tol, normal_tol, ar;
  int max_l, min_l;
  SiL_TriangMode t_mode;
  SiL_GridMode g_mode;
  SiL_AdjustMode am;
  sil_get_refinement (surf_tol, normal_tol, t_mode, g_mode, max_l, min_l,
		      max_line_length, ar, am);

  return;
}

void KernelModule::facet_object (void *object)
{
  return;
}

void KernelModule::draw_object (void *object, SimpleList<void> const &hirefs)
{
  Entity *e = static_cast<Entity*> (object);
  Cell *c = dynamic_cast<Cell*> (e);
  if (c) {
    draw_cell (e, hirefs);
  } else {
    Boundary *b = dynamic_cast<Boundary*> (e);
    if (b) {
      draw_boundary (e, hirefs);
    } else {
      sil_msg_warn ("kernel", "cannot draw " + e->type_id ());
    }
  }
  return;
}

void KernelModule::draw_fixed_object (void *object,
				      SimpleList<void> const &hirefs)
{
  return;
}

void KernelModule::del_object (void *object)
{
  drop (static_cast<Entity *> (object));
  return;
}

const char *KernelModule::name () const
{
  return nm;
}

void KernelModule::draw_boundary (Entity *b, SimpleList<void> const &hirefs)
{
  if (b->is (SIL_SOLID_TYPE)) {
    Solid *s = static_cast<Solid*> (b);
    Material *m = s->material ();
    if (!m) {
      m = new Material ();
      s->material () = m;
    }
    bool inv = hirefs.is_member (s);
    m->gl_material (inv);
    XListNode<Lump> *ln = s->lumps () . first ();
    while (ln) {
      draw_lump (ln->data (), m, inv, hirefs);
      ln = ln->next ();
    }
  } else if (b->is (SIL_SHELL_TYPE)) {
    Shell *s = static_cast<Shell*> (b);
    XListNode<Face> *fn = s->faces () . first ();
    while (fn) {
      draw_cell (fn->data (), hirefs);
      fn = fn->next ();
    }
  } else if (b->is (SIL_LOOP_TYPE)) {
    Loop *l = static_cast<Loop*> (b);
    XListNode<Edge> *en = l->edges () . first ();
    while (en) {
      draw_cell (en->data (), hirefs);
      en = en->next ();
    }
  }
  return;
}

void KernelModule::draw_cell (Entity *c, SimpleList<void> const &hirefs)
{
  Cell *C = (Cell *)c;
  Material *m = C->material ();
  if (!m) {
    m = new Material ();
    C->material () = m;
  }
  bool inv = hirefs.is_member (c);
  m->gl_material (inv);

  if (c->is (SIL_VERTEX_TYPE)) {
    draw_vertex (static_cast<Vertex*> (c), m, inv, 
		 hirefs);
  } else if (c->is (SIL_EDGE_TYPE)) {
    draw_edge (static_cast<Edge*> (c), m, inv, hirefs);
  } else if (c->is (SIL_FACE_TYPE)) {
    draw_face (static_cast<Face*> (c), m, inv, hirefs);
  } else if (c->is (SIL_LUMP_TYPE)) {
    draw_lump (static_cast<Lump*> (c), m, inv, hirefs);
  } else {
    sil_msg_warn ("kernel", "cannot draw " + c->type_id ());
  }
  
  return;
}

void KernelModule::draw_vertex (Vertex *v, Material *m, bool inv,
				SimpleList<void> const &hirefs)
{
  if (!drawn->add_first_check (v) && !inv)
    return;

  if (v->point ()) {

    if (v->material ())
      m = v->material ();
    if (!inv)
      inv = hirefs.is_member (v);
    m->gl_material_bright (inv);
    
    if (v->point ()) {
      glBegin (GL_POINTS);
      glVertex3d (v->point () -> position () -> x (),
		  v->point () -> position () -> y (),
		  v->point () -> position () -> z ());
      glEnd();
    }

  }

  return;
}

void KernelModule::draw_edge (Edge *e, Material *m, bool inv,
			      SimpleList<void> const &hirefs)
{
  if (!drawn->add_first_check (e) && !inv)
    return;

  if (e->material ())
    m = e->material ();
  if (!inv)
    inv = hirefs.is_member (e);
  m->gl_material_bright (inv);

  if (e->curve ()) {
    
    if (e->curve () -> is (SIL_STRAIGHT_TYPE)) {
      if (e->start_vertex () && e->end_vertex () &&
	  e->start_vertex () -> point () && e->end_vertex () -> point ()) {
	glBegin (GL_LINES);
	Vertex *v = e->start_vertex ();
	glVertex3d (v->point () -> position () -> x (),
		    v->point () -> position () -> y (),
		    v->point () -> position () -> z ());
	v = e->end_vertex ();
	glVertex3d (v->point () -> position () -> x (),
		    v->point () -> position () -> y (),
		    v->point () -> position () -> z ());
	glEnd();
      }
    } else if (e->curve () -> is (SIL_CIRCLE_TYPE)) {
      // x = p + r d1 cos (t) + r d2 sin (t)
      Circle *c = static_cast<Circle*> (e->curve ());
      Vector3 rd1, rd2;
      double r = c->radius () -> value ();
      if (e->start_vertex () && e->start_vertex () -> point ())
	rd1 = r * (*e->start_vertex () -> point () -> position () -
		   *c->centre ()) . normalise ();
      else
	rd1 = r * (make_ortho (*c->normal ()) . normalise ());
      rd2 = r * ((*c->normal () * rd1) . normalise ());
      Real tmax = 2.0 * PI;
      if (e->end_vertex () && e->end_vertex () -> point ()) {
	Vector3 ev = *e->end_vertex () -> point () -> position () -
	  *c->centre();
	Vector3 n1 = rd1 * ev;
	Vector3 n2 = rd1 * rd2;
	if ((n1 % *c->normal ()) * (n2 % n1) < 0.0)
	  rd2 = -rd2;
	Real a = (rd1 % ev) / (rd1.norm () * ev.norm ());
	if (a >= 1.0)
	  a = 0.0;
	else if (a <= -1.0)
	  a = -1.0;
	else
	  a = acos (a);
	if (rd2 % ev < 0.0)
	  a = 2.0 * PI - a;
	tmax = a;
      }
      double step = max (min (min (max_line_length / r, PI / 12.0),
			      tmax / 2.0),
			 2.0 * sil_get_epsilon ());
      glBegin (GL_LINE_STRIP);
      Real t = 0.0;
      while (t <= tmax) {
	Vector3 p = *c->centre () + rd1 * cos (t) + rd2 * sin(t);
	glVertex3d (p.x (), p.y(), p.z ());
	t += step;
      }
      glEnd ();
    }

    // FIXME: draw ellipses
    // FIXME: draw spline curves

  }

  if (sil_get_draw_sub_parts ()) {
    if (e->start_vertex ())
      draw_vertex (e->start_vertex (), m, inv, hirefs);
    if (e->end_vertex ())
      draw_vertex (e->end_vertex (), m, inv, hirefs);
  }

  return;
}

void KernelModule::draw_face (Face *f, Material *m, bool inv,
			      SimpleList<void> const &hirefs)
{
  if (f->material ())
    m = f->material ();
  if (!inv)
    inv = hirefs.is_member (f);
  m->gl_material (inv);

  // FIXME: draw plane
  // FIXME: draw sphere
  // FIXME: draw cylinder
  // FIXME: draw cone
  // FIXME: draw torus

  if (sil_get_draw_sub_parts ()) {
    XListNode<Loop> *ln = f->loops () . first ();
    while (ln) {
      Loop *l = ln->data ();
      XListNode<Edge> *en = l->edges () . first ();
      while (en) {
	draw_edge (en->data (), m, inv, hirefs);
	en = en->next ();
      }
      ln = ln->next ();
    }
  }

  return;
}

void KernelModule::draw_lump (Lump *l, Material *m, bool inv,
			      SimpleList<void> const &hirefs)
{
  if (l->material ())
    m = l->material ();
  if (!inv)
    inv = hirefs.is_member (l);
  m->gl_material (inv);

  XListNode<Shell> *sn = l->shells () . first ();
  while (sn) {
    Shell *s = sn->data ();
    XListNode<Face> *fn = s->faces () . first ();
    while (fn) {
      draw_face (fn->data (), m, inv, hirefs);
      fn = fn->next ();
    }
    sn = sn->next ();
  }

  return;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
