// vector.cxx: Vector class implementation.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: vector.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (vector, "$Id: vector.cc 1 2005-08-10 23:35:07Z frank $");

#include "xstring.hxx"
#include "interface.hxx"
#include "kernel/vector.hxx"

Vector3::Vector3 (UnitVector3 const &vec)
{
  a = vec.a;
  b = vec.b;
  c = vec.c;
}

Vector3 operator- (Vector3 const &v1)
{
  return Vector3 (-v1.a, -v1.b, -v1.c);
}

Vector3 operator+ (Vector3 const &v1, Vector3 const &v2)
{
  return Vector3 (v1.a + v2.a, v1.b + v2.b, v1.c + v2.c);
}

Vector3 const &Vector3::operator+= (Vector3 const &v1)
{
  a += v1.a;
  b += v1.b;
  c += v1.c;
  return *this;
}

Vector3 operator- (Vector3 const &v1, Vector3 const &v2)
{
  return Vector3 (v1.a - v2.a, 
		  v1.b - v2.b,
		  v1.c - v2.c);
}

Vector3 const &Vector3::operator-= (Vector3 const &v1)
{
  a -= v1.a;
  b -= v1.b;
  c -= v1.c;
  return *this;
}

Vector3 operator* (Real s, Vector3 const &v1)
{
  return Vector3 (s * v1.a, s * v1.b, s * v1.c);
}

Vector3 operator* (Vector3 const &v1, Real s)
{
  return Vector3 (s * v1.a, s * v1.b, s * v1.c);
}

Vector3 const &Vector3::operator*= (Real s)
{
  a *= s;
  b *= s;
  c *= s;
  return *this;
}

Vector3 operator/ (Vector3 const &v1, Real s)
{
  assert (fabs (s) > sil_get_epsilon ());
  return Vector3 (v1.a / s, v1.b / s, v1.c / s);
}

Vector3 const &Vector3::operator/= (Real s)
{
  assert (fabs (s) > sil_get_epsilon ());
  a /= s;
  b /= s;
  c /= s;
  return *this;
}

Vector3 operator* (Vector3 const &v1, Vector3 const &v2)
{
  return Vector3 (v1.b * v2.c - v1.c * v2.b,
		  v1.c * v2.a - v1.a * v2.c,
		  v1.a * v2.b - v1.b * v2.a);
}

Real operator% (Vector3 const &v1, Vector3 const &v2)
{
  return v1.a * v2.a + v1.b * v2.b + v1.c * v2.c;
}

bool operator== (Vector3 const &v1, Vector3 const &v2) 
{
  if (fabs (v1.a - v2.a) >= sil_get_epsilon ())
    return false;
  if (fabs (v1.b - v2.b) >= sil_get_epsilon ())
    return false;
  if (fabs (v1.c - v2.c) >= sil_get_epsilon ())
    return false;
  return true;
}

bool operator!= (Vector3 const &v1, Vector3 const &v2) 
{
  if (fabs (v1.a - v2.a) < sil_get_epsilon () ||
      fabs (v1.b - v2.b) < sil_get_epsilon () ||
      fabs (v1.c - v2.c) < sil_get_epsilon ())
    return false;
  return true;
}

ostream &operator<< (ostream &out, Vector3 const &v)
{
  out << "[" << v.a << ";" << v.b << ";" << v.c << "]";
  return out;
}

string const as_string (Vector3 const &v)
{
  return "[" + as_string (v.a) + ";" + as_string (v.b) + ";" + 
    as_string (v.c) + "]";
}

UnitVector3 Vector3::normalise () const
{
  return UnitVector3 (a, b, c);
}

Vector3 const &Vector3::operator= (Vector3 const &v1)
{
  if (this != &v1) {
    a = v1.a;
    b = v1.b;
    c = v1.c;
  }
  return *this;
}

Real &Vector3::operator[] (int i)
{
  switch (i) {
  case 0:
    return a;
  case 1:
    return b;
  case 2:
    return c;
  }
  cerr << "FATAL: kernel/vector.cc: illegal Vector3 index " << i << endl;
  exit (1);
}

UnitVector3 operator- (UnitVector3 const &v1)
{
  return UnitVector3 (-v1.x (), -v1.y (), -v1.z ());
}

UnitVector3 const &UnitVector3::operator= (UnitVector3 const &v1)
{
  if (this != &v1)
    set (v1.x (), v1.y (), v1.z ());
  return *this;
}

Vector3 make_ortho (Vector3 const &v)
{
  Vector3 o (0.0, 0.0, 0.0);
  if (fabs (v.x ()) > fabs (v.y ())) {
    if (fabs (v.y ()) > fabs (v.z ())) {
      o = Vector3 (v.y (), -v.x (), 0.0);
    } else {
      o = Vector3 (v.z (), 0.0, -v.x ());
    }
  } else {
    if (fabs (v.x ()) > fabs (v.z ())) {
      o = Vector3 (-v.y (), v.x (), 0.0);
    } else {
      o = Vector3 (0.0, v.z (), -v.y ());
    }
  }
  return o;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
