// vector.hxx: Vector classes.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: vector.hxx,v 1.13.2.1 2001/08/16 16:19:09 scmfcl Exp $

#ifndef VECTOR_HXX
#define VECTOR_HXX 1

#include "list.hxx"
#include "kernel/entity.hxx"
#include "interface.hxx"

class UnitVector3;

// Vector3
class Vector3 {
public:
  Vector3 () {
  }
  Vector3 (Vector3 const &v) {
    a = v.a;
    b = v.b;
    c = v.c;
  }
  Vector3 (UnitVector3 const &vec);
  Vector3 (Real x, Real y, Real z) {
    a = x;
    b = y;
    c = z;
  }
  virtual ~Vector3 () {
  }
  void set (Real x, Real y, Real z) {
    a = x;
    b = y;
    c = z;
    return;
  }
  void set (Vector3 const &v) {
    a = v.a;
    b = v.b;
    c = v.c;
    return;
  }
  void set_x (Real x) {
    a = x;
    return;
  }
  void set_y (Real x) {
    b = x;
    return;
  }
  void set_z (Real x) {
    c = x;
    return;
  }
  Real x () const {
    return a;
  }
  Real y () const {
    return b;
  }
  Real z () const {
    return c;
  }
  Real &x () {
    return a;
  }
  Real &y () {
    return b;
  }
  Real &z () {
    return c;
  }
  Real norm () const {
    return sqrt (norm_sq ());
  }
  Real norm_sq () const {
    return a * a + b * b + c * c;
  }
  UnitVector3 normalise () const;
  Vector3 const &operator= (Vector3 const &v1);
  Real &operator[] (int i);
  friend Vector3 operator- (Vector3 const &v1);
  friend Vector3 operator+ (Vector3 const &v1, Vector3 const &v2);
  Vector3 const &operator+= (Vector3 const &v1);
  friend Vector3 operator- (Vector3 const &v1, Vector3 const &v2);
  Vector3 const &operator-= (Vector3 const &v1);
  friend Vector3 operator* (Real s, Vector3 const &v1);
  friend Vector3 operator* (Vector3 const &v1, Real s);
  Vector3 const &operator*= (Real s);
  friend Vector3 operator/ (Vector3 const &v1, Real s);
  Vector3 const &operator/= (Real s);
  friend Vector3 operator* (Vector3 const &v1, Vector3 const &v2);
  friend Real operator% (Vector3 const &v1, Vector3 const &v2);
  friend bool operator== (Vector3 const &v1, Vector3 const &v2);
  friend bool operator!= (Vector3 const &v1, Vector3 const &v2);
  friend ostream &operator<< (ostream &out, Vector3 const &v);
  friend string const as_string (Vector3 const &v);
private:
  Real a, b, c;
};

// UnitVector3
class UnitVector3 : public Vector3 {
public:
  UnitVector3 () : Vector3 () {
  }
  UnitVector3 (Real x, Real y, Real z) : Vector3 () {
    set (x, y, z);
  }
  UnitVector3 (Vector3 const &vec) : Vector3 () {
    set (vec.x (), vec.y (), vec.z ());
  }
  UnitVector3 (UnitVector3 const &vec) : Vector3 () {
    set (vec.x (), vec.y (), vec.z ());
  }
  virtual ~UnitVector3 () {
  }
  void set (Real x, Real y, Real z) {
    Real l = sqrt (x * x + y * y + z * z);
    if (l > sil_get_epsilon ()) {
      Vector3::set (x / l, y / l, z / l);
    } else {
      Vector3::set (0.0, 0.0, 0.0);
    }
    return;
  }
  void set (UnitVector3 const &v) {
    Vector3::set (v);
    return;
  }
  void set_x (Real x) {
    Real l = sqrt (x * x + y() * y() + z () * z());
    if (l > sil_get_epsilon ()) {
      Vector3::set (x / l, y () / l, z () / l);
    } else {
      Vector3::set (0.0, 0.0, 0.0);
    }
    return;
  }
  void set_y (Real y) {
    Real l = sqrt (y * y + x() * x() + z () * z());
    if (l > sil_get_epsilon ()) {
      Vector3::set (x () / l, y / l, z () / l);
    } else {
      Vector3::set (0.0, 0.0, 0.0);
    }
    return;
  }
  void set_z (Real z) {
    Real l = sqrt (z * z + x () * x () + y () * y ());
    if (l > sil_get_epsilon ()) {
      Vector3::set (x () / l, y () / l, z / l);
    } else {
      Vector3::set (0.0, 0.0, 0.0);
    }
    return;
  }
  friend UnitVector3 operator- (UnitVector3 const &v1);
  UnitVector3 const &operator= (UnitVector3 const &v1);
private:
};

Vector3 make_ortho (Vector3 const &v);

#endif // !VECTOR_HXX

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
