// list.hxx: List templates.
// 
// Copyright (C) 1998-2001 Frank Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: list.hxx,v 1.21.2.1 2001/08/16 16:19:05 scmfcl Exp $
//

#ifndef LIST_HXX
#define LIST_HXX 1

template<class T> class List;

template<class T>
class ListNode {
public:
  ListNode (T *d, List<T> *l) {
    dt = d;
    lst = l;
    prv = nxt = NULL;
  }
  ListNode<T> *previous () const {
    return prv;
  }
  ListNode<T> *next () const {
    return nxt;
  }
  ListNode<T> *circ_next () const {
    if (nxt)
      return nxt;
    return list->first ();
  }
  ListNode<T> *circ_previous () const {
    if (prv)
      return prv;
    return list->last ();
  }
  void unlink ();
  void add_after (T *e);
  void add_before (T *e);
  T *data () const {
    return dt;
  }
  List<T> *list () const {
    return lst;
  }
protected:
  friend class List<T>;
  List<T> *lst;
  ListNode<T> *nxt, *prv;
  T *dt;
};

template<class T>
class List {
public:
  inline List () {
    frst = lst = NULL;
    len = 0;
  };
  ~List ();
  void add_first (T *d);
  bool add_first_check (T *d);
  void add_last (T *d);
  bool add_last_check (T *d);
  bool del (T const *d);
  ListNode<T> *first () const {
    return frst;
  }
  ListNode<T> *last () const {
    return lst;
  }
  ListNode<T> *unlink_first () {
    ListNode<T> *h = frst;
    if (h)
      frst->unlink ();
    return h;
  }
  ListNode<T> *find (T const *d) const;
  bool is_member (T const *d) const {
    return find (d) != NULL;
  }
  void clear ();
  void purge ();
  bool empty () const {
    return (frst == NULL);
  }
  unsigned long length () const {
    return len;
  }
  void combine (List<T> *list);
protected:
  friend class ListNode<T>;
  ListNode<T> *frst, *lst;
  unsigned long len;
};

template<class T>
List<T>::~List ()
{
  ListNode<T> *w = frst;
  while (w) {
    ListNode<T> *ww = w->next ();
    delete w;
    w = ww;
  }
}

template<class T>
void List<T>::clear ()
{
  ListNode<T> *w = frst;
  while (w) {
    ListNode<T> *ww = w->next ();
    delete w;
    w = ww;
  }
  lst = frst = NULL;
  len = 0;
  return;
}

template<class T>
void List<T>::purge ()
{
  ListNode<T> *w = frst;
  while (w) {
    ListNode<T> *ww = w->next ();
    delete w->data ();
    delete w;
    w = ww;
  }
  lst = frst = NULL;
  len = 0;
  return;
}

template<class T>
void List<T>::add_first (T *e)
{
  if (!frst) {
    frst = lst = new ListNode<T> (e, this);
    ++len;
    return;
  }
  frst->add_before (e);
  return;
}

template<class T>
bool List<T>::add_first_check (T *e)
{
  if (!frst) {
    frst = lst = new ListNode<T> (e, this);
    ++len;
    return true;
  }
  ListNode<T> *w = frst;
  while (w) {
    if (e == w->data ())
      return false;
    w = w->next ();
  }
  frst->add_before (e);
  return true;
}

template<class T>
void List<T>::add_last (T *e)
{
  if (!frst) {
    frst = lst = new ListNode<T> (e, this);
    ++len;
    return;
  }
  lst->add_after (e);
  return;
}

template<class T>
bool List<T>::add_last_check (T *e)
{
  if (!frst) {
    frst = lst = new ListNode<T> (e, this);
    ++len;
    return true;
  }
  ListNode<T> *w = frst;
  while (w) {
    if (e == w->data ())
      return false;
    w = w->next ();
  }
  lst->add_after (e);
  return true;
}

template<class T>
bool List<T>::del (T const *e)
{
  ListNode<T> *w = frst;
  bool res = false;
  while (w) {
    ListNode<T> *ww = w->next ();
    if (e == w->data ()) {
      w->unlink ();
      delete w;
      res = true;
    }
    w = ww;
  }
  return res;
}

template<class T>
void List<T>::combine (List<T> *list) 
{
  if (list->frst) {
    if (lst) {
      lst->nxt = list->frst;
      list->frst->prv = lst;
    } else {
      frst = list->frst;
    }
    lst = list->frst;
    while (lst) {
      lst->lst = this;
      lst = lst->nxt;
    }
    lst = list->lst;
    len += list->len;
    list->len = 0;
    list->frst = list->lst = NULL;
  }
  return;
}

template<class T>
ListNode<T> *List<T>::find (T const *e) const
{
  ListNode<T> *w = frst;
  while (w) {
    if (e == w->data ())
      return w;
    w = w->next ();
  }
  return NULL;
}

template<class T>
void ListNode<T>::add_after (T *e) 
{
  ListNode<T> *n = new ListNode<T>(e, lst);
  ++lst->len;
  if (nxt) {
    nxt->prv = n;
    n->nxt = nxt;
    nxt = n;
    n->prv = this;
  } else {
    nxt = n;
    n->prv = this;
    lst->lst = n;
  } 
  return;
}

template<class T>
void ListNode<T>::add_before (T *e)
{
  ListNode<T> *n = new ListNode<T>(e, lst);
  ++lst->len;
  if (prv) {
    prv->nxt = n;
    n->prv = prv;
    prv = n;
    n->nxt = this;
  } else {
    prv = n;
    n->nxt = this;
    lst->frst = n;
  }
  return;
}

template<class T>
void ListNode<T>::unlink ()
{
  --lst->len;
  if (prv) {
    prv->nxt = nxt;
  } else {
    lst->frst = nxt;
  }
  if (nxt) {
    nxt->prv = prv;
  } else {
    lst->lst = prv;
  }
  nxt = prv = NULL;
  return;
}

template<class T> class SimpleList;

template<class T>
class SimpleListNode {
protected:
  T *element;
  SimpleListNode<T> *next;
  friend class SimpleList<T>;
};

template<class T>
class SimpleList {
public:
  inline SimpleList () {
    head = tail = window = NULL;
    l = 0;
  }
  ~SimpleList ();
  void add_first (T *e);
  bool add_first_check (T *e);
  void add_last (T *e);
  bool add_last_check (T *e);
  bool del (T const *e);
  void clear ();
  void purge ();
  T *first ();
  T *last ();
  T *next ();
  T *circ_next ();
  T *unlink_first ();
  void reverse ();
  bool is_member (T const *d) const;
  void combine (SimpleList<T> *l);
  inline bool empty () const {
    return head == NULL;
  }
  inline unsigned long length () const {
    return l;
  }
  void sort () {
    xsort (&head, &tail);
    return;
  }
private:
  void xsort (SimpleListNode<T> **h, SimpleListNode<T> **l);
  SimpleListNode<T> *head, *tail;
  SimpleListNode<T> *window;
  unsigned long l;
};

template<class T>
SimpleList<T>::~SimpleList ()
{
  while (head) {
    window = head->next;
    delete head;
    head = window;
  }
}

template<class T>
void SimpleList<T>::add_first (T *e)
{
  SimpleListNode<T> *w = new SimpleListNode<T>;
  w->element = e;
  w->next = head;
  head = w;
  if (!tail)
    tail = w;
  ++l;
  return;
}

template<class T>
bool SimpleList<T>::add_first_check (T *e)
{
  SimpleListNode<T> *w = head;
  while (w) {
    if (e == w->element)
      return false;
    w = w->next;
  }
  w = new SimpleListNode<T>;
  w->element = e;
  w->next = head;
  head = w;
  if (!tail)
    tail = w;
  ++l;
  return true;
}

template<class T>
void SimpleList<T>::add_last (T *e)
{
  SimpleListNode<T> *w = new SimpleListNode<T>;
  w->element = e;
  w->next = NULL;
  if (tail) {
    tail->next = w;
    tail = w;
  } else {
    head = tail = w;
  }
  ++l;
  return;
}

template<class T>
bool SimpleList<T>::add_last_check (T *e)
{
  SimpleListNode<T> *w = head;
  while (w) {
    if (e == w->element)
      return false;
    w = w->next;
  }
  w = new SimpleListNode<T>;
  w->element = e;
  w->next = NULL;
  if (tail) {
    tail->next = w;
    tail = w;
  } else {
    head = tail = w;
  }
  ++l;
  return true;
}

template<class T>
bool SimpleList<T>::del (T const *e)
{
  SimpleListNode<T> *w = head;
  SimpleListNode<T> *p = NULL;
  while (w) {
    if (e == w->element) {
      if (p)
	p->next = w->next;
      else
	head = w->next;
      if (w == tail)
	tail = p;
      delete w;
      --l;
      return true;
    }
    p = w;
    w = w->next;
  }
  return false;
}

template<class T>
void SimpleList<T>::clear ()
{
  SimpleListNode<T> *w = head;
  SimpleListNode<T> *p;
  while (w) {
    p = w->next;
    delete w;
    w = p;
  }
  head = tail = window = NULL;
  l = 0;
  return;
}

template<class T>
void SimpleList<T>::purge ()
{
  SimpleListNode<T> *w = head;
  SimpleListNode<T> *p;
  while (w) {
    p = w->next;
    delete w->element;
    delete w;
    w = p;
  }
  head = tail = window = NULL;
  l = 0;
  return;
}

template <class T>
T *SimpleList<T>::first ()
{
  window = head;
  if (!window)
    return NULL;
  return window->element;
}

template <class T>
T *SimpleList<T>::last ()
{
  return tail->element;
}

template<class T>
T *SimpleList<T>::next ()
{
  if (window && window->next) {
    window = window->next;
    return window->element;
  }
  window = NULL;
  return NULL;
}

template<class T>
T *SimpleList<T>::circ_next ()
{
  if (window && window->next)
    window = window->next;
  else
    window = head;
  return window->element;
}

template<class T>
T *SimpleList<T>::unlink_first ()
{
  if (!head)
    return NULL;
  SimpleListNode<T> *f = head;
  if (head == tail)
    tail = NULL;
  head = head->next;
  if (window == f)
    window = head;
  T *e = f->element;
  delete f;
  --l;
  return e;
}

template<class T>
void SimpleList<T>::reverse ()
{
  SimpleListNode<T> *newhead = NULL;
  SimpleListNode<T> *auxnode;
  tail = head;
  while (head) {
    auxnode = head->next;
    head->next = newhead;
    newhead = head;
    head = auxnode;
  }
  head = newhead;
  return;
}

template<class T>
bool SimpleList<T>::is_member (T const *e) const
{
  SimpleListNode<T> *w = head;
  while (w) {
    if (e == w->element)
      return true;
    w = w->next;
  }
  return false;
}

template<class T>
void SimpleList<T>::combine (SimpleList<T> *list)
{
  if (list) {
    if (tail)
      tail->next = list->head;
    else
      head = list->head;
    if (list->tail)
      tail = list->tail;
    l += list->l;
    list->head = list->tail = NULL;
    list->l = 0;
  }
  return;
}

template<class T>
void SimpleList<T>::xsort (SimpleListNode<T> **h,
			   SimpleListNode<T> **l)
{
  if (*h && (*h)->next) {
    T *cmp = (*h)->element;
    SimpleListNode<T> *h1 = NULL;
    SimpleListNode<T> *h2 = NULL;
    SimpleListNode<T> *r = (*h)->next;
    while (r) {
      SimpleListNode<T> *n = r->next;
      if (*(r->element) <= *cmp) {
	if (h1) {
	  r->next = h1;
	  h1 = r;
	} else {
	  r->next = NULL;
	  h1 = r;
	}
      } else {
	if (h2) {
	  r->next = h2;
	  h2 = r;
	} else {
	  r->next = NULL;
	  h2 = r;
	}
      }
      r = n;
    }
    SimpleListNode<T> *l1;
    SimpleListNode<T> *l2;
    xsort (&h1,&l1);
    xsort (&h2,&l2);
    if (l2)
      *l = l2;
    else
      *l = *h;
    (*h)->next = h2;
    if (l1) {
      l1->next = *h;
      *h = h1;
    }
  } else {
    *l = *h;
  }
  return;
}

#endif /* LIST_HXX */

/// Local Variables:
/// exe-comm: "./sil"
/// compile-command: "make"
/// End:
