// module.hxx: Module class.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: module.hxx,v 1.13.2.1 2001/08/16 16:19:05 scmfcl Exp $

#ifndef MODULE_HXX
#define MODULE_HXX 1

#include "list.hxx"

// The Module class defines the interface between SiL and the lodable
// modules. All Modules have to be implemented as classes derived from
// the Module class where most (as far as mapplicable) virtual functions
// of the Module class are replaced. 
//
// Depending on the type of a symbol added to Sil's symbol table the
// corresponding module (if one is found) is used to modifz the
// symbol. If there is a type for which tehre is no module, SiL
// assumes that it is handled by the main Module class (in this case
// the type is set to "unkown"). But note that Module only defines the
// interface between SiL and the module. The plugins which might
// create, modify, etc. a symbol have to use the functions provided by
// the module or provide their own code. Normalli a module should
// define an API for this.
class Module {
public:
  // Initialize the modle with the constructor.
  Module () {};
  // The destructor is called when the module is permanently removed
  // from SiL. It has to cleanup everything created by the module
  // functions (if not already done).
  virtual ~Module () {};
  // Possibly add additonal information to the object as it is added
  // to the symbol table with the given name and visibility state.
  virtual void add_symbol (void *object, char *name, bool visible);
  // The object's visibility status should be changed to visible.
  virtual void show (void *object);
  // The objects visibility status should be changed to invisible.
  virtual void hide (void *object);
  // Initialize the facet functions for the symbols handled
  // by the module. That includes setting the refinement as
  // provided by the renderer.
  virtual void facet_init ();
  // Initialize the drawing functions for the symbols handled
  // by the module. That includes setting the refinement as
  // provided by the renderer.
  virtual void draw_init ();
  // Create the facets for rendering the object. More generally
  // this can be used to prepare the object for rendering.
  virtual void facet_object (void *object);
  // Draw the object using OpenGL if the object does not have
  // a fixed position related to rotation (!Lights).
  virtual void draw_object (void *object, SimpleList<void> const &hirefs);
  // Draw the object using OpenGL if the object has a a fixed position
  // related to rotation (Lights).
  virtual void draw_fixed_object (void *object, 
				  SimpleList<void> const &hirefs);
  // Delete the object permanently.
  virtual void del_object (void *object);
  // Load the contents of the file using the given format. If ext
  // is true also load special SiL extensions to the file format.
  // If replace is true, replace the symbols if they are already
  // defined.
  virtual void load_file (FILE *fp, const char *format, bool ext,
			  bool replace);
  // Save the symbols from the symbol table the file using the given
  // format (only those that can be saved). If ext is true also save
  // any special SiL extensions.
  virtual void save_file (FILE *fp, const char *format, bool ext);
  // Get list of provided file formats:
  //   FORMAT, DESCRIPTION, FORMAT, DESCRIPTION, ..., NULL
  virtual const char **file_formats () const;
  // Get the type identifier/name for the module.
  virtual const char *name () const;
private:
  // Static pointer to the "unkown" type identifier.
  static const char *nm;
};

#endif // !MODULE_HXX

/// Local Variables:
/// exe-comm: "./sil"
/// compile-command: "make"
/// End:
