// closeness.cc: Identiy measure for surfaces.
//
// Copyright (C) 1999-2000 Bruce I Mills <B.I.Mills@cs.cf.ac.uk>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: closeness.cc 1 2005-08-10 23:35:07Z frank $
//
// * Original code desinged and implemented by Bruce I Mills 
//   <B.I.Mills@cs.cf.ac.uk> for sil 0.1.4.
// * Modifications for sil 0.1.5 and cleanups by Frank C Langbein 
//   <frank@langbein.org>.

#include "config.h"

RCSID (closeness, "$Id: closeness.cc 1 2005-08-10 23:35:07Z frank $");

#include "acis/acis.hxx"

#include "interface.hxx"
#include "xmath.hxx"
#include "list.hxx"
#include "acis/silapi.hxx"

// Various routines relating to the analysis of a solid model 

/*

Measures of identity:
=====================

This long comment is a discussion of the justification for the 
routines that follow. The basic problem is to determine some 
floating point number computed from the data for two surfaces
such that the number is 0 for identicle surfaces, and gradually
larger for surfaces that are more and more distinct. What would be
good in this context would be an appropriate metric on the space
of these surfaces.


Surfaces:
---------

We want a measure of how closely two faces, F1 and F2, of an object 
are related by an isometric transformation, T. We do this by creating
a metric on the space of faces, and then finding the distance between 
F1 and T(F2). As long as the metric between faces is not affected by
the application of an isometric transformation to both faces, that is,
dist(A,B) = dist(T(A),T(B)), then this resulting measure is symmetric
We could make this a bit more robust by taking the square root of the 
sum of the squares of dist(A,B) and dist(B,A) if this was required.

Being part of the surface of a solid gives a distinction between the
two unit normals to the surface at a given point. One is the inward
unit normal, pointing locally into the material of the object, and the
other is the outward normal, which points locally into space.

In dealing with the question of whether or not F1 ~ T(F2) the question
of which side the object is on should be taken into account. So if 
T brings F2 into coincidence with F1, then we should also check to
see whether it brings the inward and outward directions into 
correspondence. If not, then T is not a symmetry of the body, even if
it does map one face onto another.


Planar surfaces:
----------------

In ACIS the family of planes in space is parameterised by
giving a point p on the plane, and a unit normal n to the plane.
The unit normal points out from the object that the plane is 
a part of. We are interested, as explained above, in finding
a measure for the closeness of two oriented planes. Now a
description of the plane is {x : n.x=n.p}. The value of 
f(x) = (n.x-n.p) = n.(x-p) is positive if x is on the side of 
the plane to which n is pointing. So this expression also holds
the information about the orientation of the plane. Thus the
vector (n : n.p), is a vector of 4 real numbers which represent
the oriented plane in a continuous manner. So we can use the
distance between these as a measure of closeness of two planes.

closeness((n1,p1),(n2,p2)) = |(n1-n2 : n1.p1 - n2.p2)|

It should be fine to use the square of this, which is 
slightly simpler to compute, and monotonic with respect to
the norm since the norm is always non negative. 


Cylinders and cones:
--------------------

In ACIS, the class CONE handles cones and also handles cylinders
as a special case. However the full family of surfaces that can
be represented by an instance of the CONE class is eliptical cones.
It has been stated that for the moment we are only going to handle
circular cross sections. There is a major/minor axis ratio entry in
the class CONE, and if this is exactly the unity floating point number
then the cross section is circular. There is no need to approximate
this. Since it requires a lot more work to determine the closeness of
two eliptical cones, than of two cones, we will assume here that 
this is the case. If a non unity maj/minor axis ratio is found, then
a -1 will be returned to express the point that this surface is outside
the range currently being compared.

nb: to compare eliptical cone we could start by comparing elipses.
these can be characterised by the equation (ax+b)^2+(cy+d)^2 = 0
The numbers {a,b/a,c,d/c} are unique, and can be used to test
equality, or the distance between two vectors made up from them
in a consistant order can be used as a continous measure. This can
be wrangled into an apropriate measure on eliptical cones, or we
could take the equation of the cone itself and work with that, but
stability and complexity of the formulas is an issue. The fact that
equivalence of two circles in 2-space is not affected by rotation 
makes the whole thing a lot simpler.

d  = ((CONE *)(lg[j]))->direction();
m  = ((CONE *)(lg[j]))->major_axis();
p  = ((CONE *)(lg[j]))->root_point();
sa = ((CONE *)(lg[j]))->sine_angle();
ca = ((CONE *)(lg[j]))->cosine_angle();
rr = ((CONE *)(lg[j]))->radius_ratio();

We assume that radius_ratio = 1.00000....

sa = 0 && ca > 0 .... cylinder with solid on concave side

sa = 0 && ca < 0 .... cylinder with solid on convex side

sa > 0 && ca > 0 .... cone, getting bigger in given direction
                      with solid on concave side

sa < 0 && ca < 0 .... cone, getting bigger in given direction
                      with solid on convex side

sa < 0 && ca > 0 .... cone, getting smaller in given direction
                      with solid on concave side

sa > 0 && ca > 0 .... cone, getting smaller in given direction
                      with solid on convex side

sa < 0 && ca = 0 .... not acceptable in ACSIS

sa > 0 && ca = 0 .... not acceptable in ACSIS

It is unclear at this stage whether ACSIS uses an algebraic
formula for defining the cone in terms of the parameters, or
whether it uses logical tests to sort things out. However even
if there is a formula, the approach in general does not seem
to show a purely algebraic effect. For example if the radius
of a sphere is negative we get the outside of the sphere rather
than the inside. (Algebraicly, using x^2+y^2+z^2-r^2 as the
basic formula, the sign of the radius is irrelevant, rather 
it is the sign of r^2 (sort of), so we'd need an imaginary
radius to do this (but also imaginary scaling). The general
conclusion being that there is no point in looking for 
algebraic simplicity in the ACIS solid modeler.

In this context, backed up by empirical test of the modeler
where I set up the data structure so that it would have an
apex at an existing location on the surface, I find that
if we try to extend the surface through the apex then the
result is a model that is logically inconsistant about which
is the inside and which is the outside. So the best thing is
to assume that in ACIS, the underlying notion of a cone is
something that is topologically a single sheet, like the 
probverbial blanket with a ball bearing weighing it down 
in the middle. The sheet can be cut by planes to make a
frustrum, or as a special case it can be turned into a
cylinder. But it never a non manifold surface.

For a given position, not at the apex, there are two ways in
which a given cone can be defined, one for each direction along
the axis. It looks like being best to just try both options for
matching the cones. One on the assumption that the directions
are the same, and the other for the directions assumed opposite.

For the directions being the same we require that

dirn_1 = dirn_2
(ca_1, sa_1) = (ca_2, sa_2)
radius_2 = ||posn_2 - posn_1|| sa_1 + radius_1
radius_1 = ||posn_1 - posn_1|| sa_2 + radius_2

For the directions being opposite we require that

dirn_1 = - dirn_2
(ca_1, sa_1) = -(ca_2, sa_2)
radius_2 = ||posn_2 - posn_1|| sa_1 + radius_1
radius_1 = ||posn_1 - posn_1|| sa_2 + radius_2

We can use the sum of the squares of the diference between the
left hand side and the right hand side as a measure of how close
the cones are to each other. Compute this for the presumption of
directions the same, and directions opposite, and then take the
minimum of these to indicate the closeness of the cones.


Toroidal surfaces:
------------------

Different shapes covered by the ACIS data type TORUS:

inside doughnut ......  major >  minor > 0
outside doughnut .....  major > -minor > 0

inside apple .........  minor >  major > 0
outside apple ........ -minor >  major > 0

inside lemon .........  minor > -major > 0
outside lemon ........ -minor > -major > 0

*/

// Determine the sided surface intended by an instance of TORUS.
int torus_option (float major, float minor)
{
  if (( major >  minor) && ( minor > 0))
    return 1; // the inside of a doughnut
  if (( major > -minor) && (-minor > 0))
    return 2; // the outside of a doughnut
  if (( minor >  major) && ( major > 0))
    return 3; // the inside of a apple
  if ((-minor >  major) && ( major > 0))
    return 4; // the outside of a apple
  if (( minor > -major) && (-major > 0))
    return 5; // the inside of a lemon
  if ((-minor > -major) && (-major > 0))
    return 6; // the outside of a lemon
  return 0;  // not a valid torus option
}

double plane_plane_measure (SURFACE *s1, SURFACE *s2)
{
  unit_vector dirn1 = ((PLANE *)s1)->normal ();
  unit_vector dirn2 = ((PLANE *)s2)->normal ();
  position posn1 = ((PLANE *)s1)->root_point ();
  position posn2 = ((PLANE *)s2)->root_point ();
  double d1 = dirn1.x () * posn1.x () + dirn1.y () * posn1.y () + 
    dirn1.z () * posn1.z ();
  double d2 = dirn2.x () * posn2.x () + dirn2.y () * posn2.y () + 
    dirn2.z () * posn2.z ();
  return (dirn1 - dirn2) . len () + (d1 - d2) * (d1 - d2);
}

double cone_cone_measure (SURFACE *s1, SURFACE *s2)
{
  // first check to see if we have a circular cross section
  double rr1 = ((CONE *)(s1)) -> radius_ratio ();
  double rr2 = ((CONE *)(s2)) -> radius_ratio ();

  if (rr1 != 1.0 || rr2 != 1.0) 
    return -1.0; // can't compare these surfaces

  // extract the desired parameters from surface s1
  unit_vector dirn1 = ((CONE *)(s1)) -> direction ();
  double rad1       = ((CONE *)(s1)) -> major_axis () . len ();
  position posn1    = ((CONE *)(s1)) -> root_point ();
  double sa1        = ((CONE *)(s1)) -> sine_angle ();
  double ca1        = ((CONE *)(s1)) -> cosine_angle ();

  // extract the desired parameters from surface s1
  unit_vector dirn2 = ((CONE *)(s2)) -> direction ();
  double rad2       = ((CONE *)(s2)) -> major_axis () . len ();
  position posn2    = ((CONE *)(s2)) -> root_point ();
  double sa2        = ((CONE *)(s2)) -> sine_angle ();
  double ca2        = ((CONE *)(s2)) -> cosine_angle ();

  // compute metric for direct match of cones
  double f1 = (dirn1 - dirn2) . len ();
  double f2 = ca1 - ca2; 
  double f3 = sa1 - sa2;
  double f4 = rad2 - (posn2 - posn1) . len () * sa1 - rad1;
  double f5 = rad1 - (posn1 - posn2) . len () * sa2 - rad2;

  double dm = f1 * f1 + f2 * f2 + f3 * f3 + f4 * f4 + f5 * f5;

  // compute metric for indirect match of cones
  f1 = (dirn1 + dirn2) . len ();
  f2 = ca1 + ca2;
  f3 = sa1 + sa2;
  f4 = rad2 - (posn2 - posn1) . len () * sa1 - rad1;
  f5 = rad1 - (posn1 - posn2) . len () * sa2 - rad2;

  double im = f1 * f1 + f2 * f2 + f3 * f3 + f4 * f4 + f5 * f5;

  return min (dm, im);
}

double sphere_sphere_measure (SURFACE *s1, SURFACE *s2)
{
  position posn1 = ((SPHERE *)s1) -> centre ();
  double rad1 = ((SPHERE *)s1) -> radius ();

  position posn2 = ((SPHERE *)s2) -> centre ();
  double rad2 = ((SPHERE *)s2) -> radius ();
  
  double f1 = (posn1 - posn2) . len ();
  double f2 = rad1 - rad2;

  return f1 * f1 + f2 * f2;
}

double torus_torus_measure (SURFACE *s1, SURFACE *s2)
{
  double major1 = ((TORUS *)s1) -> major_radius ();
  double minor1 = ((TORUS *)s1) -> minor_radius ();

  double major2 = ((TORUS *)s2) -> major_radius ();
  double minor2 = ((TORUS *)s2) -> minor_radius ();

  int opt1 = torus_option (major1, minor1);
  int opt2 = torus_option (major2, minor2);

  if (opt1 == 0 || opt2 == 0 || opt1 != opt2)
    return -1.0;

  position posn1 = ((TORUS *)s1) -> centre();
  unit_vector dirn1 = ((TORUS *)s1) -> normal();

  position posn2 = ((TORUS *)s2) -> centre();
  unit_vector dirn2 = ((TORUS *)s2) -> normal();

  double f1 = (posn1 - posn2) . len ();

  double f2 = (dirn1 * dirn2) . len ();

  double f3 = major1 - major2;
  double f4 = minor1 - minor2;

  return f1 * f1 + f2 * f2 + f3 * f3 + f4 * f4;
}


// This routine computes a general notion of closeness for two
// arbitrary surfaces, however it does not actually compare different
// types such as cone and plane, although they could in principle be
// compared with in some bounding region. This would be a fairly
// natural extension of this basic idea.
// 
// It returns a closeness value, roughly speaking, this is 0 for
// identical things, and more and more positive for things that are
// not very similar. A -1 is returned if the two surface are deamed
// incomperable.
double closeness (SURFACE *s1, SURFACE *s2)
{
  if (strcmp (s1->type_name (), s2->type_name ()))
    return -1.0;

  if (!strcmp (s1->type_name (), "plane"))
    return plane_plane_measure (s1, s2);
  if (!strcmp (s1->type_name (), "cone"))
    return cone_cone_measure (s1, s2);
  if (!strcmp (s1->type_name (), "sphere"))
    return sphere_sphere_measure (s1, s2);
  if (!strcmp (s1->type_name (), "torus"))
    return torus_torus_measure (s1, s2);

  return -1.0; // unknown type of surface, so can't compare
}


/*

The exact case:
---------------

The next thing we need is a method for finding all the isometries
that map one surface to another. For example, given two planes we
have a rotary translation that maps one to the other, and then we
can apply any element of the symmetry group of the plane.

Suppose I have two subsets S1 and S2 of space. We wish to find the
set of isometries {T : T(S1) = S2}. This set may be empty. Call
this set Map(S1,S2). Now if T is in Map(S1,S2) and T' is a 
symmetry of S2, then T'oT is also an element of Map(S1,S2).

The other way around, if T1 and T2 are both in Map(S1,S2) then
T = T2^-1 o T1 must be a symmetry of S1. So T1 = T2 o T. Thus
all elements of Map(S1,S2) can be expressed as the composition
of some selcted elements, with a symmetry of S1.

This is the exact case however, and there are questions of 
how much this is true approximately, if we deal with the
set of elements that approximately map S1 to S2.

The approximate case:
---------------------

We start with a class, S, of surfaces in space, such that S is closed
under the application of isometries of space. Then we construct a
metric, d, on S which generates the same topology as the Hausdorf metric.
This is the measure of closeness mentioned earlier. Now, given two
specific surfaces s1 and s2, we can construct a map from the set of
isometries of space onto the real numbers by asigning to each isometry,
T, the number d(T(s1),s2). Intuitively, we are measuring how closely
T maps s1 to s2. We want then to identify the set of all isometries
that bring s1 to within a distance x of s2.

Now, suppose that the metric, d, is not affected by isometries.
By this we mean that for any isometry T, d(T(s1),T(s2) = d(s1,s2).
If T maps s1 to within x of s2, and R1 is a symmetry of s1 and  
R is a symmetry of s2, then d(R2Ts1,R2s2) = d(R2Ts1,s2) = d(R2TR1 s1,s2)

note: take the direct product of the group of symmetries of s1 and the
group of symmetries of s2. Now we can define an action of this product
on the space of isometries, ie (r1,r2) * T = r1Tr2. Not only is this
an action, but it is a faithful action. However would typically have
multiple orbits, since the aplication of a symmetry of s1 and a symetry
of s2, for bounded s1 and s2, would not be able to construct enough
translations to map every isometry to every other isometry. 


We consider an example. Let S be the class of subsets of 3-space that
contain exactly a single point. The distance between two elements, s1
and s2 of S is just the distance between the two points. Now given 
s1 and s2 that are distinct, the set of isometries that map s1 to 
exactly a given distance x from s2 is the set of isometries that 
map s1 to the surface of the sphere of radius x about s2.

*/

// Figure out what entities a body is constructed from
void deconstruct (ENTITY *e)
{
  if (!is_BODY (e)) {
    cout << "I don't know what this is\n";
    return;
  }
  cout << "This is a body\n";

  BODY *b = (BODY *)e;
  sil_acis_check (api_change_body_trans (b, NULL),
		  "closeness", "api_change_body_trans");

  LUMP *sl = b->lump ();
  LUMP *l = sl;
  int i = 0;
  while ((l = l->next ()) && (sl != l))
    ++i;
  ++i;
  cout << "This body has " << i << " lumps\n";

  SHELL *ss = sl->shell ();
  SHELL *s = ss;
  i = 0;
  while ((s = s->next ()) && (ss != s))
    ++i;
  ++i;
  cout << "This lump has " << i << " shells\n";

  FACE *sf = ss->face ();
  FACE *f = sf;
  i = 0;
  while ((f = f->next ()) && (sf != f))
    ++i;
  FACE *lf[i];
  f = sf;
  i = 0;
  while ((f = f->next ()) && (sf != f))
    lf[i++] = f;
  cout << "This shell has " << i << " faces\n";

  SURFACE *lg[i];
  for(int j = 0; j < i; ++j)
    lg[j] = lf[j]->geometry ();

  cout << "about to list all the faces\n";

  for(int j = 0; j < i; ++j) {
    cout << lg[j]->type_name() << endl;
    if (!strcmp (lg[j]->type_name (), "plane")) {
      position p = ((PLANE *)(lg[j])) -> root_point ();
      unit_vector v = ((PLANE *)(lg[j])) -> normal ();
      printf ("p = (%3.2f,%3.2f,%3.2f)\n", p.x (), p.y (), p.z ());
      printf ("v = (%3.2f,%3.2f,%3.2f)\n", v.x (), v.y (), v.z ());
    } else if (!strcmp(lg[j]->type_name (), "cone")) {
      unit_vector d  = ((CONE *)(lg[j]))->direction ();
      vector m       = ((CONE *)(lg[j]))->major_axis ();
      position p     = ((CONE *)(lg[j]))->root_point ();
      double sa      = ((CONE *)(lg[j]))->sine_angle ();
      double ca      = ((CONE *)(lg[j]))->cosine_angle ();
      double rr      = ((CONE *)(lg[j]))->radius_ratio ();
      printf ("p = (%3.2f,%3.2f,%3.2f)\n", p.x (), p.y (), p.z ());
      printf ("d = (%3.2f,%3.2f,%3.2f)\n", d.x (), d.y (), d.z ());
      printf ("m = (%3.2f,%3.2f,%3.2f)\n", m.x (), m.y (), m.z ());
      printf ("sa = %3.2f\n", sa);
      printf ("ca = %3.2f\n", ca);
      printf ("rr = %3.2f\n", rr);
    } else if (!strcmp(lg[j]->type_name (), "sphere")) {
      position p  = ((SPHERE *)(lg[j]))->centre ();
      double r  = ((SPHERE *)(lg[j]))->radius ();
      printf ("p = (%3.2f,%3.2f,%3.2f)\n", p.x (), p.y (), p.z ());
      printf ("r = %3.2f\n", r);
    } else if(!strcmp(lg[j]->type_name (), "torus")) {
      position p = ((TORUS *)(lg[j]))->centre ();
      unit_vector v = ((TORUS *)(lg[j]))->normal ();
      double mi = ((TORUS *)(lg[j]))->minor_radius ();
      double ma = ((TORUS *)(lg[j]))->major_radius ();
      printf("p = (%3.2f,%3.2f,%3.2f)\n",p.x(),p.y(),p.z());
      printf("v = (%3.2f,%3.2f,%3.2f)\n",v.x(),v.y(),v.z());
      printf("mi = %3.2f\n",mi);
      printf("ma = %3.2f\n",ma);
    }
  }

  char *(name[4]) = {"cone","sphere","plane","torus"};

  for (int nc = 0; nc < 4; ++nc) {
    for (int j1 = 0; j1 < i; ++j1) {
      for (int j2 = 0; j2 < i; ++j2) {
	float c = closeness (lg[j1], lg[j2]);
	if ((c > -0.5) && (!strcmp (name[nc], lg[j2]->type_name ()))) {
	  if (j1 < 10)
	    printf (" ");
	  printf ("%d:",j1);
	  int count = printf ("%s", lg[j1]->type_name ());
	  while (count++ < 8)
	    printf(" ");
	  if (j2 < 10) 
	    printf(" ");
	  printf ("%d:", j2);
	  count = printf("%s", lg[j2]->type_name ());
	  while (count++ < 8)
	    printf(" ");
	  printf(" .... %9.2f\n", c);
	}
      }
    }
  }

  return;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

#define MAX_PARTIAL_INJECTION_SIZE 200
#define PARTIAL_INJECTION_DEBUG 1

class partial_injection
{
public:
  int size;                                   // keep track of size of map 
  int forward[MAX_PARTIAL_INJECTION_SIZE];    // forward map
  int reverse[MAX_PARTIAL_INJECTION_SIZE];    // reverse map

  void identity (int n);
  void empty ();
  int is_empty ();
  int addpair (int x, int y);
  int subpair (int x, int y);
  int mulpair (int x, int y);
  void show ();
  void show (FILE *f);
};

void partial_injection::identity(int n)
{
  int limit = min (n, MAX_PARTIAL_INJECTION_SIZE);
  this->size = limit - 1;
  int i = 0;
  while (i < limit) {
    this->forward[i] = i;
    this->reverse[i] = i;
    ++i;
  }
  while (i < MAX_PARTIAL_INJECTION_SIZE) {
    this->forward[i] = -1;
    this->reverse[i] = -1;
    ++i;
  }
  return;
}

void partial_injection::empty()
{
  this->identity (0);
  return;
}

int partial_injection::is_empty()
{
  return this->size == 0;
}

// attempt to just add the pair (x,y) into the partial injection 
// if adding the member (x,y) to the underlying relation of the
// injection produces a non injection then then an error condition
// is returned 

int partial_injection::addpair (int x, int y)
{
#ifdef PARTIAL_INJECTION_DEBUG
  if (x > MAX_PARTIAL_INJECTION_SIZE)
    return 1;
  if (y > MAX_PARTIAL_INJECTION_SIZE)
    return 1;
  if (this->forward[x] != -1)
    return 2;
  if (this->reverse[y] != -1)
    return 2;
#endif // PARTIAL_INJECTION_DEBUG

  this->forward[x] = y;
  this->reverse[y] = x;
  this->size++;

  return 0;
}

int partial_injection::subpair (int x, int y)
{
#ifdef PARTIAL_INJECTION_DEBUG
  if (x > MAX_PARTIAL_INJECTION_SIZE)
    return 1;
  if (y > MAX_PARTIAL_INJECTION_SIZE)
    return 1;
  if (this->reverse[y] == -1)
    return 2;
  if (this->forward[x] == -1)
    return 2;
  if (this->forward[x] != this->reverse[y])
    return 3;
#endif // PARTIAL_INJECTION_DEBUG

  this->forward[x] = -1;
  this->reverse[y] = -1;
  this->size--;

  return 0;
}

// Treat the pair as a partial injection containing this one pair.
// Compose the two producing a new partial injection, however if
// the domain and range dont overlap, then add the pair in straight.

int partial_injection::mulpair (int x, int y)
{
#ifdef PARTIAL_INJECTION_DEBUG
  if (x > MAX_PARTIAL_INJECTION_SIZE) 
    return 1;
  if (y > MAX_PARTIAL_INJECTION_SIZE)
    return 1;
  if (this->reverse[y] != -1)
    return 2;
#endif // PARTIAL_INJECTION_DEBUG

  int t = this->reverse[x];
  if (t == -1) {
    t = x;
    this->size++;
  }

  this->forward[t] = y;
  this->reverse[y] = t;

  return 0;
}

void partial_injection::show()
{
  this->show (stdout);
  return;
}

void partial_injection::show (FILE *f)
{
  printf ("{ ");
  for (int i = 0; i < MAX_PARTIAL_INJECTION_SIZE ; ++i)
    if (this->forward[i] != -1) 
      printf("(%d,%d) ", i, this->forward[i]);
  printf ("}\n");
  return;
}

void test_partial_injection ()
{
  cout << "about to test the partial injection class\n";
  partial_injection fred;
  fred.empty();
  for (int i = 0; i < 20; ++i) 
    fred.addpair (i, i + 1);
  fred.show ();
  for (int i = 0; i < 20; ++i) 
    fred.mulpair (i, i + 1);
  fred.show ();
  fred.identity (20);
  fred.show ();
  return;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

/*

Algorithm to determine approximate regularities:
================================================

Broadly speaking the intention here is to determine, from a boundary 
representation solid model, a set of isometries each of which map the 
solid approximately onto itself. However the notion of approximation
is not a predefined idea, but rather to be designed to be compatible 
with the intuitive concept that corresponds to a human assertion that 
two objects are similar.


The structure of an ACIS boundary model:
----------------------------------------

The data that the algorithm begins with is an ACIS boundary model.
An ACIS model includes a description of a collection of points, curves 
and surfaces in space which satisfy certain conditions, together with 
some combinatorial information about the way these geometric entities 
are related to each other. These points, curves and surfaces that occur 
in the boundary model will be called vertices, edges and faces 
respectively.

The vertices, edges and faces satisfy some constrants on the way in 
which they intersect. The intersection of any two entities is either
empty, or equal to some entity in the collection.  Further, if the 
intersection of two distinct faces is not empty then it is an edge 
or vertex, it is never another face. The intersection of two edges
is likewise empty, or a vertex. No two distinct vertices intersect.

An edge of finite length is either a loop, such as a circle, having
no end points, or it is similar to a line segment, having two end
points. In the latter case, the end points must be vertices in the 
collection. Each face is of finite area and may have natural edges
like a piece of paper with holes cut in it. These edges are called
the boundary of the face. The boundary of each face is a collection 
of loops. Each loop must be the union of a number of edges and 
vertices from the collection of entities. 

The ACIS solid modeling library gives a geometric description of
each of the vertices, edges and faces. For a vertex this just means
specifying its location. For edges and faces a geometric curve or
surface must be specified. This is done in two stages, firstly an
underlying, or carrier, geometry is described, and then a section
of this carrier is indicated. For example a circular arc can be 
specified by first describing the circle, and then indicating the
end points of the arc. This leaves two posibilities, so it must
also be stated which of the two possible pieces is intended. The 
situation for faces is more complicated again, however the principle 
intention is the same.

The ACIS library also makes available the information indicating 
which vertices are the end points of which edges, and which edges 
are in the boundary of which faces. This information is available 
in each direction as a constant time operation. 

The information telling which vertices are the end points of which
edges explicitely describes a network in which each non-loop edge 
connects two vertices. It is also assumed that each edge is in 
the boundary of exactly two faces. In this way an edge can be said
to connect two faces as well. Thus we have two networks defined
by the edges. One in which the edges connect vertices and the other
in which the edges connect faces. The former graph is called the
vertex graph, and the latter graph is called the face graph of
the boundary representation.


Our constraints on the boundary models:
---------------------------------------

We are not intending to handle every type of object that ACIS is
capable of modeling. In particular we can limit the situation to
handling models which define a single solid object, with faces
whose carrier surfaces are planes, cylinders, cones, spheres and
tori. In ACIS the cylinder type is actually a special case of 
cone, and a distinction will not always be drawn between these
two types of surface. Faces are the primary data, taken from the
laser scan of the component being modeled. So the shape of an
edge is determined by the intersection of the two faces that
it connects. Thus no explicit limit on the carrier geometry of
an edge is specified.


The nature of the output from the algorithm:
---------------------------------------------

Given a description of an object as an exact shape in space there
may be non trivial rigid motions of the object that would bring it 
into exact correspondence with itself. Such a motion is called an
isometric symmetry of the the object, and generically the object
is said to be symmetric. In principle we can invisage a process
that, given an exact description of an object, will return a 
description of the set of all isometric symmetries of the object.

Although an object may have an infinite number of symmetries 
the collections of rigid motions involved are not completely
arbitrary. Each complete symmetry collection for any finite
object must be the same as that for some prism, pyramid,
platonic solid, cone, cylinder or sphere. Thus, for example, 
describing a generic object with the same symmetry is a compact 
method for indicating the symmetry of the given object.

As stated, this type of symmetry requires infinite precision, 
while in practice we will be dealing with rough numerical data. 
One might suppose then that these symmetries would simply be very 
rare. But this misses most of the picture. In working with physical
components we can never measure them exactly, but the concept of
symmetry still exists. The reason is that it is not demanded that
the incidence be exact. We have a concept of approximately 
symmetric. The symmetry is accepted if the identity is
verified up to a specified precision.



---------------------------------------------------------------------


In acis, planes, cylinders, cones spheres and tori each divide
space up into exactly two components.

Each face on the boundary is a part of one of the above surfaces.

Suppose that the surface of the object is a connected combinatorial
manifold. 

Now space is cut up into sections by the various carrier 
geometries. The object itself is defined as the union of
a bunch of these pieces. 

Now, if two objects include the same pieces then they are
the same object.

---------------------------------------------------------------------


We have a graph, which is the face graph of the boundary representation.
In this graph we want to do synchronised backtracking tree growing. So we 
need to be able to attach and detach leaves on each tree, and keep track
of which nodes correspond in the two trees. We can do this with a tree
of pairs. A tree of pairs is like a pair of trees, except that it 
automatically enforces an isomorphism between the two trees, and records
directly what that isomorphism is. 

The operations need to be fast, wich is part of the reason for restricting
them to attach and detach of leaves. In this way the allowable tree operations
are constant time. It should be possible to organise all the other stuff
to use these operations. Now the second thing is that we need to be able
to check quickly whether or not we are currently using a node in the graph
as a leaf in (independently) either tree. We don't have control over the
graph nodes, unless we are prepared to go through and build a whole new
one, which is an option, however another approach is to use integer
arrays and direct references into them.

The algorithm begins with an explicit array F[n] of faces. This can 
be obtained more or less directly from ACIS, and its generation is 
not seen as being part of the algorithm. So in particular we know how 
many faces there are. We assign enough memory, an operation that is
taken to be constant time, for two arrays L and R of size n. The arrays
are initialised to -1, which is a linear operation in n. One array
for each tree. L[i] records the face in tree R that face i in L tree
is paired with in tree R, and vice versa. 

A node in the tree of pairs of faces is expressed by a pair (l,r) of
indices into the face array. Now in the face graph, F[l] and F[r]
will both have a number of neighbours, we can get a list of these,
and check to see which ones have not been used yet. Assuming an
upper limit to the number of edges a face has, this is a constant
time operation. However, complexity analysis should, perhaps, take
this upper limit into consideration as a seperate parameter.

So we have a left list of faces to try and a right list of faces
to try. We now need to know what possible pairings of these to try.
In the version of the algorithm that assumes exact topology we see
that the number of unused neighbours must be equal in each tree.
So we could stop at this point if they are not. This sort of 
situation could in principle arise if loops where made in one
tree, but not in the other. However, perhaps this is unlikely
when dealing with a search that is controlled by the behaviour
of a boundary representation in space.

Two faces may be connected to each other by more than one
edge. So we need to take this into account. What we are doing
implicitly is to produce a matching of the edges of the faces.
But we want to stick with the faces as the geometrical data.

In the ACIS data structures, the boundary of a face is a collection
of loops. Given that we are trying to extend from a given face pair
(l,r), we want to match up that part of the boundary that has not
already been used. The full matching of the boundary must first
match loops, and then match elements in the loop in cyclic 
order. Due to the geometric nature of this, there are strong 
constraints.

If any faces that bound face l, are already used, then we must 
have the same condition for face r, and used faces that bound
l must already be paired with used faces that bound r. This 
must occur in a manner that respects the loop structure as
well. In principle we could use the available isometries to
figure out which bounding faces can be matched correctly.

We should memorise the generation of collections of isometries
that match each pair of faces.



ACIS will give us a box that contains an object .. eg a face.


To begin with we want a fixed point. Given several isometries
we can find a mutual fixed point. If we are dealing with a
linear regularity, then we only need to find a fixed line.


With n^2 processing can create both a n by n array giving the
number of direct connections between the faces, and a collection
of n arrays, of storage length n, which store directly the list
of faces that are adjacent to a given face. This enables us
to quickly get the combinatorial information of both varieties
with a fast constant time operation.

Can we produce the information about combinatorial local match?
Ie, a bit array indicating if the immediate environemnt
of two faces is compatible. Let's see ... as we do the initial
n^2 scan, keep an array A[1..n] for each face, store in A[i] the
number of faces with i connections to this face.


Given a network of n faces, the number of possible combinatorial
symetries is n!, however for faces embedded in space (regardless
of whether they form the boundary of a solid object) is much less.
In principle, fixing, for example, four faces that have centroid
that are not coplanar fixes exactly the isometry, which implies
in turn all the other face matchings.


Suppose that the set of isometries that are symmetries of all
the faces contains just the identity.

A set of faces is said to be dependent if the set of isometries
that map each face to itself contains only the identity. Now the
empty set is independent. Further if we have an independent set
then the deleting elements can only increase, not decrease the
set of isometries that keep each face fixed. 

Given a set of faces, we have the collection of isometries that
map every face to itself. Call a subcollection dependent if the
only symmetries that map each element of the subcollection to
itself is the set that maps the whole to itself. Thus there is
at least one dependent set, the set as a whole. Now what we 
want is a minimal dependent set. Prefereably we would like to
find a minimal dependent set that also has the minimum number
of faces for a dependent set.

Suppose that we have four faces, consider the simplex formed from
their centroids. If this is a non degenerate simplex, then the
faces are a dependent set.

Lemma: if the set of isometries that map every face to itself
is the identity, then there exists a set of four faces that
are dependent.

Proof:

Consider the convex hull of the centroids of the faces. If this
is nondegenerate, then there must be a set of four vertices,
whose correpsonding faces are the required faces (and quite
robustly too).

If the convex hull is two dimensional, then there must be
a set of three faces that prevent all movement, except for
a reflection in the plane, now if every single face is
mirror symmetric in this plane, then this would be in the
set of symmetries that preserve all faces, but by assumption
it is not, and so there must be a fourth face that when
added prevents all motion.

If the convex hull is one dimensional, then there must be
two faces that force the line between their centroids to
be preserved. This leaves rotations about this line, 
and mirror reflections with planes that contain the
line. 

It is possible to create a collection of faces whose centroids
are collinear, and such that no three of them will prevent all
rotations, but the collection of all of them will. For example
polygonal disks of orders ... 2*3*5 3*5*7 5*7*2 7*2*3, Any 
three of these have an overlap of one prime factor, and so 
this will be a common rotational symmetry. However, we still
have to make up a solid object, and we need to do it so that
all the centroids are on the line. With the surfaces that 
we are using this would apear to be difficult at least. 
Further, we are limiting, in principle, rotational symmetries
to less than 8, and the smallest of the above is 30 fold.

For mirror planes, if there is one face without mirror symmetry
then include that, but if all have mirror symmetry, pick one
that uses a mirror symmetry that is not shared by all the 
elments of the three already picked, if there isn't one, then
this mirror symmetry would be in the set that fixes all the
faces, so we can find one.

The completes an approximate demonstration that you can
find a set of four faces that fix an object that has 
only the identity that fixes all the faces.

end-of-proof

This means that once you know where all four of these 
faces go, then you have the isometry fixed down to 
exactly one isometry, which greatly improves the
performance of the rest of the algorithm.



If we have two planar faces, then either they are parallel
or the set of isometries that fix them is the identity.
We can't make a polyhedron out of purely planar faces,
so for polyhdera we find that there must be an appropriate
four faces.

If we at least one planar face, then we the set of symmetries are 
translations parallel to the plane, and rotations perpendicular. 
If we have even one cylinder, cone, sphere or torus, we narrow the 
axis of rotation down to one. Each of these auxilary surfaces has a 
point or line that must be preserved, this fixes the possible
rotations to a single axis. Further all centres, and apexes must be 
on this line, or the set of isometries is reduced to only the 
identity. In which case we can find four surfaces that maximally fix 
the object. Otherwise we are back to a similar case for the centroids 
on a line. Except in cases of a high finite degree of rotational 
symmetry we must have four faces that maximally fix the object.

Similar sorts of things happen, but even more restrictively, for 
having at least one of the other surfaces. Spheres, tori and cones 
fix a point, and a line, cones fix a line. Adding one or two surfaces 
with dissimilar symmetry tends to stop the whole object from moving
once you have fixed them, or to reduce the motions down to the minimum 
possible, except for high order finite degree rotational symmetries.


How to find such a set:
-----------------------

If you have a polyhedron, compute the convex hull, or for that matter 
go through and find two that have a maximal centroid distance, and 
then look for something that maximises the area of the triangle, and 
then something that maximises the area of the tetrahedron. This is 
an n^2 scan, and two linear scans.

In fact you could use this technique anyway, either you get a good 
bet tetrahedron, or you show that the symmetries are restriced to 
a reflection, that can be tried explicitly, or rotations about a 
single axis, that can be looked for. Or you might find that all 
the centroids are at one point.

A sphere with a cylindrical hole in it can have this property, 
but it is a pretty restricted type of object.

NOTE: the regularities of the object will also be regularities of 
the convex closure of the set of all centroids of the faces, although 
it might be a good idea to weight these with the area of the face and 
such like. 

Ok, now we have four `registration' faces. Connect these up with a 
minimal tree in the face graph.  When we try to find the symmetries 
of the object as a whole, we only search along this tree. 

We need to find a second tree that looks like the first. At worst 
we have n places to try the root of the tree. It's hard to tell exactly
how many options will be available after that, but its going to be fairly 
limited, except where we are talking about a high degree of symmetry.

Given that we are looking at an explicit list of the isometric 
symmetries, the qeustion is how much mucking about to find each 
of these.

Note: Sugihara finds out if two polyhedra are congruent in n log n 
time, ie, in our context, what he finds is one symmetry of the polyhedron
in n log n time. Sugihara uses an edge based partitioning scheme, looking 
at the length, the angle between the two faces that meet at that edge, 
and the angle between the edges at either end of the edge. Something
similar might be done with area and number of faces.

*/

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

void algorism ()
{
  test_partial_injection();
  return;
}

void closeness (SimpleList<Symbol const> *sl)
{
  // Comapres surfaces of the body with each other using some
  // measure.
  Symbol const *s = sl->first ();
  while (s) {
    if (sil_symbol_is (s, "acis")) {
      BODY *main_body = (BODY *)sil_get_symbol_data (s);
      deconstruct (main_body);
    }
    s = sl->next ();
  }
  return;
}

SIL_PLUGIN ()
{
  sil_register_tool (&algorism, "Algorism", "0.1",
		     "Bruce I Mills", "2000 Bruce I Mills",
		     "Test of partial injection class (obsolete)",
		     "acis", "acis");
  sil_register_tool (&closeness, "Closeness", "0.1",
		     "Bruce I Mills", "2000 Bruce I Mills",
		     "Deconstruct body and compare faces",
		     "acis", "acis");
  return; 
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
