// distort.cc: distort ACIS model
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: distort.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (distort, "$Id: distort.cc 1 2005-08-10 23:35:07Z frank $");

#include "acis/acis.hxx"

#include "list.hxx"
#include "interface.hxx"
#include "xmath.hxx"
#include "acis/light.hxx"
#include "acis/silapi.hxx"
#include "base/base.hxx"

#include <lop_husk/api/lop_api.hxx>

double distort_length = 0.1;
double distort_angle = 1.0;

#define PM ((rand () < (RAND_MAX >> 1))?-1.0:1.0)
#define RAND (rand () / (double) RAND_MAX)
#define PMRAND (PM * RAND)
#define RANDDIST(d) (d * PMRAND)
#define RANDVAL(a,d) (a + RANDDIST(d))

class DistortWindow;

List<DistortWindow> distort_windows;

class DistortWindow : public Gtk::Window {
public:
  DistortWindow (BODY *b, string const &n);
  ~DistortWindow ();
  void update ();
  void close ();
private:
  void cancel ();
  void move_face (FACE *f);
  void tweak_face (FACE *f);
  void distort ();
  string sld_name;
  BODY *B;
  Gtk::VBox vbox;
  Gtk::HBox titlebox, atolbox, ltolbox, bbox;
  Gtk::Label titlelabel, title2label;
  Gtk::Label l_atol, l_ltol;
  Gtk::Adjustment a_ltol;
  Gtk::SpinButton s_ltol;
  Gtk::Adjustment a_atol;
  Gtk::SpinButton s_atol;
  Gtk::Button b_distort, b_cancel;
};

DistortWindow::DistortWindow (BODY *b, string const &n)
  : Gtk::Window (GTK_WINDOW_TOPLEVEL),
	 vbox (),
	 titlebox (true),
	 atolbox (true),
	 ltolbox (true),
	 bbox (true),
	 titlelabel ("Distort BODY:"),
	 title2label (),
	 l_atol ("Angle Tolerance:"),
	 l_ltol ("Length Tolerance:"),
	 a_ltol (distort_length, 0.0, 1000.0, 0.1, 1.0, 0.0),
	 s_ltol (a_ltol, 0.1, 4),
	 a_atol (distort_angle, 0.0, 180.0, 0.1, 1.0, 0.0),
	 s_atol (a_atol, 0.1, 4),
	 b_distort ("Distort"),
	 b_cancel ("Cancel")
{
  B = b;
  sld_name = n;

  title2label.set (n.c_str ());
  title2label.set_justify (GTK_JUSTIFY_LEFT);
  titlelabel.set_justify (GTK_JUSTIFY_RIGHT);

  l_atol.set_justify (GTK_JUSTIFY_LEFT);
  l_ltol.set_justify (GTK_JUSTIFY_LEFT);

  vbox.pack_start (titlebox, false, true, 2);
  titlebox.pack_start (titlelabel, true, true, 0);
  titlebox.pack_start (title2label, true, true, 0);

  vbox.pack_start (atolbox, false, true, 2);
  atolbox.pack_start (l_atol, false, true, 0);
  atolbox.pack_start (s_atol, false, true, 0);

  vbox.pack_start (ltolbox, false, true, 2);
  ltolbox.pack_start (l_ltol, false, true, 0);
  ltolbox.pack_start (s_ltol, false, true, 0);

  vbox.pack_start (bbox, false, true, 2);
  bbox.pack_start (b_distort, false, true, 0);
  bbox.pack_start (b_cancel, false, true, 0);
  
  add (vbox);

  b_distort.clicked.connect (slot (this, &DistortWindow::distort));
  b_cancel.clicked.connect (slot (this, &DistortWindow::cancel));

  show_all ();

  return;
}

DistortWindow::~DistortWindow ()
{
  ListNode<DistortWindow> *n = distort_windows.find (this);
  if (n) {
    n->unlink ();
    delete n;
  }
  return;
}

void DistortWindow::update ()
{
  const Symbol *s = sil_find_symbol (sld_name.c_str ());
  if (!s) {
    ListNode<DistortWindow> *n = distort_windows.find (this);
    if (n) {
      n->unlink ();
      delete n;
    }
    hide_all ();
    destroy ();
  }
  return;
}

void DistortWindow::close ()
{
  ListNode<DistortWindow> *n = distort_windows.find (this);
  if (n) {
    n->unlink ();
    delete n;
  }
  hide_all ();
  destroy ();
  return;
}

void DistortWindow::cancel ()
{
  close ();
  return;
}

void DistortWindow::move_face (FACE *f)
{
  FACE *fs[1] = { f };

  vector t (PMRAND, PMRAND, PMRAND);
  while (t.len () < sil_get_epsilon ())
    t = vector (PMRAND, PMRAND, PMRAND);
  t /= t.len ();
  t *= RANDDIST(distort_length);

  vector r (PMRAND, PMRAND, PMRAND);
  while (r.len () < sil_get_epsilon ())
    r= vector (PMRAND, PMRAND, PMRAND);
  r /= r.len ();
  double angle = RANDDIST(distort_angle / 180.0 * PI);

  sil_acis_check (api_move_faces (1, fs, translate_transf (t) *
				  rotate_transf (angle, r),
				  position (-10000.0, -10000.0, -10000.0),
				  position (10000.0, 10000.0, 10000.0)),
		  "distort", "api_move_faces");

  return;
}

void DistortWindow::tweak_face (FACE *f)
{
  int reverse[1] = { 0 };

  SURFACE *S = f->geometry ();

  if (is_SPHERE (S)) {
    SPHERE *N = (SPHERE *)S;
    N->set_radius (RANDVAL(N->radius (), distort_length));
  } else if (is_CONE(S)) {
    CONE *N = (CONE *)S;
    if (N->radius_ratio () == 1.0) {
      if (fabs (N->sine_angle ()) < sil_get_epsilon ()) {
	N->set_major_axis (RANDVAL(N->major_axis ().len (), distort_length) *
			   N->major_axis () / N->major_axis () . len ());
      }
    }
  } else if (is_TORUS(S)) {
    TORUS *N = (TORUS *)S;
    N->set_major_radius (RANDVAL (N->major_radius (), distort_length));
    N->set_minor_radius (RANDVAL (N->minor_radius (), distort_length));
  }

  FACE *fs[1] = { f };
  SURFACE *ss[1] = { S };
  sil_acis_check (api_tweak_faces (1, fs, ss, reverse,
				   position (-10000.0, -10000.0, -10000.0),
				   position (10000.0, 10000.0, 10000.0)),
		  "distort", "api_tweak_faces");

  return;
}

void DistortWindow::distort ()
{
  b_distort.set_sensitive (false);
  b_cancel.set_sensitive (false);

  distort_angle = s_atol.get_value_as_float ();
  distort_length = s_ltol.get_value_as_float ();

  sil_progress_bar (true, true);
  sil_progress_bar_text ("Distort");

  ENTITY_LIST fcs;
  sil_acis_check (api_get_faces (B, fcs), "distort", "api_get_faces");
  
  double prog = 0.0;
  sil_progress_bar_percentage (prog);
  double step = 0.65 / fcs.count ();
  if (sil_cancel_request ()) {
    sil_progress_bar (false);
    return;
  }
  
  sil_acis_check (api_initialize_local_ops (),
		  "distort", "api_initialize_local_ops");

  bool c = true;
  fcs.init ();
  FACE *f = (FACE *)fcs.next ();
  while (f && c) {
    sil_progress_bar_percentage (prog += step);
    move_face (f);
    if (sil_cancel_request ())
      c = false;
    f = (FACE *)fcs.next ();
  }

  if (c) {
    ENTITY_LIST fcs2;
    sil_acis_check (api_get_faces (B, fcs2), "distort", "api_get_faces");
    step = 0.35 / fcs2.count ();

    fcs2.init ();
    f = (FACE *)fcs2.next ();
    while (f && c) {
      sil_progress_bar_percentage (prog += step);
      tweak_face (f);
      if (sil_cancel_request ())
	c = false;
      f = (FACE *)fcs2.next ();
    }
  }

  sil_acis_check (api_terminate_local_ops (),
		  "distort", "api_initialize_local_ops");

  sil_refacet ();
  sil_progress_bar_percentage (1.0);

  sil_progress_bar (false);

  close ();

  return;
}

void distort_symbol (SimpleList<Symbol const> *sl)
{
  Symbol const *s = sl->first ();
  while (s) {
    const char *t = sil_get_symbol_type (s);
    if (strcmp(t, "acis") == 0) {
      ENTITY *e = static_cast<ENTITY *> (sil_get_symbol_data (s));
      if (is_BODY (e)) {
	DistortWindow *fw = manage (new DistortWindow 
				    (static_cast<BODY *> (e),
				     static_cast<string const> 
				     (sil_get_symbol_name (s))));
	distort_windows.add_first (fw);
      }
    }
    s = sl->next ();
  }
  return;
}

void update_handler ()
{
  ListNode<DistortWindow> *l = distort_windows.first ();
  while (l) {
    ListNode<DistortWindow> *nn = l->next ();
    l->data () -> update ();
    l = nn;
  }
  return;
}

SIL_PLUGIN ()
{
  sil_register_tool (&distort_symbol, "Distort", "0.2",
		     "Frank C Langbein", "2000-2001 Frank C Langbein",
		     "Distort ACIS model.", "acis", "acis");
  sil_register_update_handler (&update_handler);
  return;
}

SIL_CLOSE_PLUGIN ()
{
  sil_deregister_update_handler (&update_handler);
  ListNode<DistortWindow> *l = distort_windows.first ();
  while (l) {
    ListNode<DistortWindow> *nn = l->next ();
    l->data () -> close ();
    l = nn;
  }
  return;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
