// show.cc: show information about kernel symbols
//
// Copyright (C) 2000-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: show.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (show, "$Id: show.cc 1 2005-08-10 23:35:07Z frank $");

#include "interface.hxx"

#include "list.hxx"
#include "kernel/kernel.hxx"
#include "kernel/analyse.hxx"
#include "kernel/combinatorics.hxx"

class AnalyseWindow;

List<AnalyseWindow> analyse_windows;

void solid_hilite (bool hi, Analyse *a)
{
  if (a->name () == a->hilite_name ()) {
    sil_hilite (hi, a->hilite_name () . c_str (), a->solid ());
  } else {
    XListNode<Lump> *n = a->solid () -> lumps () . first ();
    while (n) {
      if (n->data () -> reference ())
	sil_hilite (hi, a->hilite_name () . c_str (),
		    n->data () -> reference ());
      n = n->next ();
    }
  }
  sil_redraw ();
  return;
}

void reg_hilite (bool hi, Regularity *r)
{
  r->hilite (hi);
  sil_redraw ();
  return;
}

class AnalyseWindow : public Gtk::Window {
public:
  AnalyseWindow (Analyse *a, char const *n);
  ~AnalyseWindow ();
  void update ();
  void update_analyse ();
  void close ();
  void add_selected (Regularity *r) {
    selected.add_first (r);
  }
  void del_selected (Regularity *r) {
    selected.del (r);
  }
private:
  void add_regs (List<Regularity> *rl);
  void add_model_info ();
  void reg_item (Regularity *r, Gtk::Tree *t);
  void hilite_all (List<Regularity> *rl, bool t) const;
  Analyse *analyse;
  SimpleList<Regularity> selected;
  Gtk::VBox vbox;
  Gtk::ScrolledWindow swin;
  Gtk::Tree *tree;
  Gtk::HBox hbox;
  Gtk::Button b_update;
  Gtk::Button b_close;
  int rows;
  char *name;
};

AnalyseWindow::AnalyseWindow (Analyse *a, char const *n) 
  : Gtk::Window (GTK_WINDOW_TOPLEVEL),
	 vbox (),
	 swin (),
	 hbox (),
	 b_update ("Update"),
	 b_close ("Close")
{
  analyse = a;
  take (analyse);

  name = new char[strlen(n) + 1];
  strcpy (name, n);

  set_title ("SiL Analyse Results");
  set_usize (450, 500);

  add (vbox);
  
  tree = manage (new Gtk::Tree ());
  tree->set_selection_mode (GTK_SELECTION_MULTIPLE);

  rows = 0;

  add_model_info ();

  SimpleList< List<Regularity> > *regs = analyse->regularities ();
  List<Regularity> *rl = regs->first ();
  while (rl) {
    add_regs (rl);
    rl = regs->next ();
  }

  swin.add_with_viewport (*tree);
  vbox.pack_start (swin, true, true, 2);

  hbox.pack_start (b_update, true, true, 2);
  hbox.pack_start (b_close, true, true, 2);
  vbox.pack_start (hbox, false, true, 2);
  
  b_update.clicked.connect (slot (this, &AnalyseWindow::update_analyse));
  b_close.clicked.connect (slot (this, &AnalyseWindow::close));

  show_all ();
}

AnalyseWindow::~AnalyseWindow()
{
  if (name) {
    delete []name;
    name = NULL;
  }

  ListNode<AnalyseWindow> *n = analyse_windows.find (this);
  if (n) {
    n->unlink ();
    delete n;
  }

  if (analyse) {
    SimpleList< List<Regularity> > *regs = analyse->regularities ();
    List<Regularity> *rl = regs->first ();
    while (rl) {
      hilite_all (rl, false);
      rl = regs->next ();
    }
    drop (analyse);
    analyse = NULL;
    sil_redraw ();
  }
}

void AnalyseWindow::add_regs (List<Regularity> *rl)
{
  if (rl) {
    ListNode<Regularity> *r = rl->first ();
    if (r) {
      char str[1024];
      snprintf (str, 1023, "%s [%ld]", r->data () -> name () . c_str (),
		rl->length ());
      Gtk::TreeItem *item = manage (new Gtk::TreeItem (str));
      tree->append (*item);
      ++rows;
      item->show ();
      Gtk::Tree *tr = manage (new Gtk::Tree ());
      tr->set_selection_mode (GTK_SELECTION_MULTIPLE);
      tr->set_view_mode (GTK_TREE_VIEW_ITEM);
      item->set_subtree (*tr);
      reg_item (r->data (), tr);
      r = r->next ();
      while (r) {
	reg_item (r->data (), tr);
	r = r->next ();
      }
    }
  }
  return;
}

void AnalyseWindow::reg_item (Regularity *r, Gtk::Tree *t)
{
  string const text = r->description ();
  Gtk::TreeItem *item = manage (new Gtk::TreeItem (text.c_str ()));
  item->select.connect (SigC::bind<bool,Regularity*>
			(SigC::slot (&reg_hilite), true, r));
  item->deselect.connect (SigC::bind<bool,Regularity*>
			  (SigC::slot (&reg_hilite), false, r));
  t->append (*item);
  ++rows;
  item->show ();
  if (r->child_regularities () &&
      r->child_regularities () -> length () > 0) {
    Gtk::Tree *subtree = manage (new Gtk::Tree ());
    subtree->set_selection_mode (GTK_SELECTION_MULTIPLE);
    subtree->set_view_mode (GTK_TREE_VIEW_ITEM);
    item->set_subtree (*subtree);
    Regularity *rr = r->child_regularities () -> first ();
    while (rr) {
      reg_item (rr, subtree);
      rr = r->child_regularities () -> next ();
    }
  }
  return;
}

void AnalyseWindow::add_model_info ()
{
  gchar str[1024];
  if (analyse->name () == analyse->hilite_name ()) {
    snprintf (str, 1023, "Analyser info for %s", analyse->name () . c_str ());
  } else {
    snprintf (str, 1023, "Analyser info for %s (%s)",
	      analyse->name () . c_str (),
	      analyse->hilite_name () . c_str ());
  }
  Gtk::TreeItem *item = manage (new Gtk::TreeItem (str));
  item->select.connect (SigC::bind<bool,Analyse*>
			(SigC::slot (&solid_hilite),
			 true, analyse));
  item->deselect.connect (SigC::bind<bool,Analyse*>
			  (SigC::slot (&solid_hilite),
			   false, analyse));
  tree->append (*item);
  ++rows;
  item->show ();
  Gtk::Tree *tr = manage (new Gtk::Tree ());
  tr->set_selection_mode (GTK_SELECTION_MULTIPLE);
  tr->set_view_mode (GTK_TREE_VIEW_ITEM);
  item->set_subtree (*tr);

  Solid *s = analyse->solid ();

  snprintf (str, 1023, "Lumps: %ld", s->lumps () . length());
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Shells: %ld", s->shells () . length());
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Faces: %ld", s->faces () . length());
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Loops: %ld", s->loops () . length());
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Edges: %ld", s->edges () . length());
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Vertices: %ld", s->vertices () . length());
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Max length: %.4g", analyse->maximum_length ());
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Average length: %.4g", analyse->average_length ());
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Total time: %0ldm %0ld.%03ds",
	    analyse->time_secs () [0] / 60, 
	    analyse->time_secs () [0] % 60, 
	    analyse->time_msecs () [0]);
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Scalar time: %0ldm %0ld.%03ds",
	    analyse->time_secs () [1] / 60, 
	    analyse->time_secs () [1] % 60, 
	    analyse->time_msecs () [1]);
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Parallel direction time: %0ldm %0ld.%03ds",
	    analyse->time_secs () [2] / 60, 
	    analyse->time_secs () [2] % 60, 
	    analyse->time_msecs () [2]);
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Direction plane time: %0ldm %0ld.%03ds",
	    analyse->time_secs () [3] / 60, 
	    analyse->time_secs () [3] % 60, 
	    analyse->time_msecs () [3]);
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Planar angle-regular time: %0ldm %0ld.%03ds",
	    analyse->time_secs () [4] / 60, 
	    analyse->time_secs () [4] % 60, 
	    analyse->time_msecs () [4]);
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Direction cone time: %0ldm %0ld.%03ds",
	    analyse->time_secs () [5] / 60, 
	    analyse->time_secs () [5] % 60, 
	    analyse->time_msecs () [5]);
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Conical angle-regular time: %0ldm %0ld.%03ds",
	    analyse->time_secs () [6] / 60, 
	    analyse->time_secs () [6] % 60, 
	    analyse->time_msecs () [6]);
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Axis intersection time: %0ldm %0ld.%03ds",
	    analyse->time_secs () [7] / 60, 
	    analyse->time_secs () [7] % 60, 
	    analyse->time_msecs () [7]);
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Axis position time: %0ldm %0ld.%03ds",
	    analyse->time_secs () [8] / 60, 
	    analyse->time_secs () [8] % 60, 
	    analyse->time_msecs () [8]);
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Position time: %0ldm %0ld.%03ds",
	    analyse->time_secs () [9] / 60, 
	    analyse->time_secs () [9] % 60, 
	    analyse->time_msecs () [9]);
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Loop time: %0ldm %0ld.%03ds",
	    analyse->time_secs () [10] / 60, 
	    analyse->time_secs () [10] % 60, 
	    analyse->time_msecs () [10]);
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  snprintf (str, 1023, "Geometry type time: %0ldm %0ld.%03ds",
	    analyse->time_secs () [11] / 60, 
	    analyse->time_secs () [11] % 60, 
	    analyse->time_msecs () [11]);
  item = manage (new Gtk::TreeItem (str));
  tr->append (*item);
  ++rows;
  item->show ();

  return;
}

void AnalyseWindow::hilite_all (List<Regularity> *rl, bool t) const 
{
  ListNode<Regularity> *r = rl->first ();
  while (r) {
    r->data () -> hilite (t);
    r = r->next ();
  }
  return;
}

void AnalyseWindow::close ()
{
  hide_all ();
  destroy ();
  return;
}

void AnalyseWindow::update_analyse ()
{
  sil_progress_bar (true, false);

  sil_progress_bar_text ("Unhilighting regularities");
  SimpleList< List<Regularity> > *regs = analyse->regularities ();
  List<Regularity> *rl = regs->first ();
  while (rl) {
    hilite_all (rl, false);
    rl = regs->next ();
  }
  sil_redraw ();

  analyse->eval ();
  sil_msg ("analyse", "Updated analysis of " +
	   static_cast<string>(analyse->name ()) + ".");

  sil_progress_bar_text ("Updating regularities list");
  tree->clear_items (0, rows);
  selected.clear ();
  rows = 0;
  add_model_info ();
  rl = regs->first ();
  while (rl) {
    add_regs (rl);
    rl = regs->next ();
  }

  sil_progress_bar (false);
  return;
}

void AnalyseWindow::update ()
{
  const Symbol *s = sil_find_symbol (name);
  if (s) {
    SimpleList< List<Regularity> > *regs = analyse->regularities ();
    List<Regularity> *rl = regs->first ();
    while (rl) {
      hilite_all (rl, false);
      rl = regs->next ();
    }
    sil_redraw ();
    tree->clear_items (0, rows);
    selected.clear ();
    rows = 0;
    add_model_info ();
    rl = regs->first ();
    while (rl) {
      add_regs (rl);
      rl = regs->next ();
    }
  } else {
    if (name) {
      delete []name;
      name = NULL;
    }
    if (analyse) {
      SimpleList< List<Regularity> > *regs = analyse->regularities ();
      List<Regularity> *rl = regs->first ();
      while (rl) {
	hilite_all (rl, false);
	rl = regs->next ();
      }
      drop (analyse);
      analyse = NULL;
      sil_redraw ();
    }
    ListNode<AnalyseWindow> *n = analyse_windows.find (this);
    if (n) {
      n->unlink ();
      delete n;
    }
    hide_all ();
    destroy ();
  }
  return;
}

void show (SimpleList<Symbol const> *sl)
{
  Symbol const *s = sl->first ();
  while (s) {
    const char *t = sil_get_symbol_type (s);
    if (strcmp (t, "kernel") == 0) {
      Entity *e = (Entity *)sil_get_symbol_data (s);
      if (e->is (SIL_ANALYSE_TYPE)) {
	AnalyseWindow *aw = manage 
	  (new AnalyseWindow ((Analyse *)e,
			      sil_get_symbol_name (s)));
	analyse_windows.add_first (aw);
      }
    }
    s = sl->next ();
  }
  return;
}

void update_handler ()
{
  ListNode<AnalyseWindow> *n = analyse_windows.first ();
  while (n) {
    ListNode<AnalyseWindow> *nn = n->next ();
    n->data () -> update ();
    n = nn;
  }
  return;
}

SIL_PLUGIN ()
{
  sil_register_tool (&show, "Show", "0.1.1",
		     "Frank C Langbein", "2000 Frank C Langbein",
		     "Display information about a kernel symbol.",
		     "kernel", "kernel");
  sil_register_update_handler (&update_handler);
  return;
}

SIL_CLOSE_PLUGIN ()
{
  sil_deregister_update_handler (&update_handler);
  ListNode<AnalyseWindow> *n = analyse_windows.first ();
  while (n) {
    ListNode<AnalyseWindow> *nn = n->next ();
    n->data () -> close ();
    n = nn;
  }
  return;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
