// simplify.cc: simplify analyser results
//
// Copyright (C) 2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: simplify.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (simplify, "$Id: simplify.cc 1 2005-08-10 23:35:07Z frank $");

#include "list.hxx"
#include "interface.hxx"
#include "xmath.hxx"
#include "kernel/kernel.hxx"
#include "kernel/analyse.hxx"
#include "kernel/combinatorics.hxx"

double max_length = 0.2;
double max_angle = 2.0;

class SimplifyWindow;

List<SimplifyWindow> simplify_windows;

class SimplifyWindow : public Gtk::Window {
public:
  SimplifyWindow (Analyse *a, string const &n);
  ~SimplifyWindow ();
  void update ();
  void close ();
private:
  void cancel ();
  void check_subregs (Regularity *r);
  void simplify ();
  string name;
  Analyse *A;
  Gtk::VBox vbox;
  Gtk::HBox titlebox, atolbox, ltolbox, bbox;
  Gtk::Label titlelabel, title2label;
  Gtk::Label l_atol, l_ltol;
  Gtk::Adjustment a_ltol;
  Gtk::SpinButton s_ltol;
  Gtk::Adjustment a_atol;
  Gtk::SpinButton s_atol;
  Gtk::Button b_simplify, b_cancel;
};

SimplifyWindow::SimplifyWindow (Analyse *a, string const &n)
  : Gtk::Window (GTK_WINDOW_TOPLEVEL),
	 vbox (),
	 titlebox (true),
	 atolbox (true),
	 ltolbox (true),
	 bbox (true),
	 titlelabel ("Simplify Regularities:"),
	 title2label (),
	 l_atol ("Maximum Angle Tolerance:"),
	 l_ltol ("Maximum Length Tolerance:"),
	 a_ltol (max_length, 0.0, 1000.0, 0.1, 1.0, 0.0),
	 s_ltol (a_ltol, 0.1, 4),
	 a_atol (max_angle, 0.0, 180.0, 0.1, 1.0, 0.0),
	 s_atol (a_atol, 0.1, 4),
	 b_simplify ("Simplify"),
	 b_cancel ("Cancel")
{
  A = a;
  name = n;

  title2label.set (n.c_str ());
  title2label.set_justify (GTK_JUSTIFY_LEFT);
  titlelabel.set_justify (GTK_JUSTIFY_RIGHT);

  l_atol.set_justify (GTK_JUSTIFY_LEFT);
  l_ltol.set_justify (GTK_JUSTIFY_LEFT);

  vbox.pack_start (titlebox, false, true, 2);
  titlebox.pack_start (titlelabel, true, true, 0);
  titlebox.pack_start (title2label, true, true, 0);

  vbox.pack_start (atolbox, false, true, 2);
  atolbox.pack_start (l_atol, false, true, 0);
  atolbox.pack_start (s_atol, false, true, 0);

  vbox.pack_start (ltolbox, false, true, 2);
  ltolbox.pack_start (l_ltol, false, true, 0);
  ltolbox.pack_start (s_ltol, false, true, 0);

  vbox.pack_start (bbox, false, true, 2);
  bbox.pack_start (b_simplify, false, true, 0);
  bbox.pack_start (b_cancel, false, true, 0);
  
  add (vbox);

  b_simplify.clicked.connect (slot (this, &SimplifyWindow::simplify));
  b_cancel.clicked.connect (slot (this, &SimplifyWindow::cancel));

  show_all ();

  return;
}

SimplifyWindow::~SimplifyWindow ()
{
  ListNode<SimplifyWindow> *n = simplify_windows.find (this);
  if (n) {
    n->unlink ();
    delete n;
  }
  return;
}

void SimplifyWindow::update ()
{
  const Symbol *s = sil_find_symbol (name.c_str ());
  if (!s) {
    ListNode<SimplifyWindow> *n = simplify_windows.find (this);
    if (n) {
      n->unlink ();
      delete n;
    }
    hide_all ();
    destroy ();
  }
  return;
}

void SimplifyWindow::close ()
{
  ListNode<SimplifyWindow> *n = simplify_windows.find (this);
  if (n) {
    n->unlink ();
    delete n;
  }
  hide_all ();
  destroy ();
  return;
}

void SimplifyWindow::cancel ()
{
  close ();
  return;
}

void SimplifyWindow::check_subregs (Regularity *r)
{
  if (r->sub_regularities ()) {
    Real max_angle_rad = max_angle / 180.0 * PI;
    Regularity *sr = r->sub_regularities () -> first ();
    while (sr) {
      Regularity *srn = r->sub_regularities () -> next ();
      switch (sr->type ()) {
      case SIL_SPECIALVALUE_TYPE:
	// NOTE: We do not simplify special values at the moment
	break;
      case SIL_PARALLELDIRECTION_TYPE:
      case SIL_ORTHOGONALDIRECTION_TYPE:
	if (sr->tolerance () > max_angle_rad) {
	  r->sub_regularities () -> del (sr);
	  delete sr;
	}
	break;
      default:
	sil_msg_warn ("Simplify", "unknown sub-regularity " + 
		      sr->type_id ());
      }
      sr = srn;
    }
    if (r->sub_regularities () -> length () == 0)
      r->delete_sub_regularities ();
  }
  return;
}

void SimplifyWindow::simplify ()
{
  b_simplify.set_sensitive (false);
  b_cancel.set_sensitive (false);

  max_angle = s_atol.get_value_as_float ();
  max_length = s_ltol.get_value_as_float ();

  Real max_angle_rad = max_angle / 180.0 * PI;

  sil_progress_bar (true, true);
  sil_progress_bar_text ("Simplify");

  unsigned long count = 0;
  unsigned long all = A->regularities () -> length ();
  List<Regularity> *rs = A->regularities () -> first ();
  while (rs) {
    sil_progress_bar_percentage (static_cast<Real> (count++) /
				 static_cast<Real> (all));
    ListNode<Regularity> *rn = rs->first ();
    sil_progress_bar_text (rn->data () -> type_id ());
    switch (rn->data () -> type ()) {
    case SIL_CLUSTER_TYPE:
      sil_progress_bar_text (static_cast<Cluster*> (rn->data ()) -> 
			     value () -> type_id ());
      switch (static_cast<Cluster*> (rn->data ()) -> value () -> type ()) {
      case SIL_DIRECTION_TYPE:
      case SIL_ANGLE_TYPE:
      case SIL_DIRECTIONCONE_TYPE:
      case SIL_DIRECTIONPLANE_TYPE:
	// Simplify angular tolerance cluster hierarchy
	rn = rs->first ();
	while (rn) {
	  ListNode<Regularity> *rnn = rn->next ();
	  Cluster *c = static_cast<Cluster*> (rn->data ());
	  if (c->tolerance () > max_angle_rad) {
	    rn->unlink ();
	    if (c->sub_clusters ()) {
	      Cluster *sc;
	      while ((sc = c->sub_clusters () -> unlink_first ())) {
		rs->add_last (sc);
		sc->parent_cluster () = NULL;
		if (!rnn)
		  rnn = rs->last ();
	      }
	    }
	    A->removed_regularities () -> add_first (c);
	    delete rn;
	  } else {
	    check_subregs (c);
	  }
	  rn = rnn;
	}
	break;
      case SIL_LENGTH_TYPE:
      case SIL_INTERSECTIONPOSITION_TYPE:
      case SIL_ALIGNEDAXES_TYPE:
      case SIL_POSITION_TYPE:
      case SIL_PARTIALLYPLANEEQUAL_TYPE:
      case SIL_PARTIALLYLINEEQUAL_TYPE:
      case SIL_POINTCIRCLE_TYPE:
	// Simplify length tolerance cluster hierarchy
	rn = rs->first ();
	while (rn) {
	  ListNode<Regularity> *rnn = rn->next ();
	  Cluster *c = static_cast<Cluster*> (rn->data ());
	  if (c->tolerance () > max_length) {
	    rn->unlink ();
	    if (c->sub_clusters ()) {
	      Cluster *sc;
	      while ((sc = c->sub_clusters () -> unlink_first ())) {
		rs->add_last (sc);
		sc->parent_cluster () = NULL;
		if (!rnn)
		  rnn = rs->last ();
	      }
	    }
	    A->removed_regularities () -> add_first (c);
	    delete rn;
	  } else {
	    check_subregs (c);
	  }
	  rn = rnn;
	}
	break;
      case SIL_LOOPPOLYGON_TYPE:
	// NOTE: We do not simplify this
	break;
      default:
	sil_msg_warn ("Simplify", "unknown cluster regularity for " + 
		      static_cast<Cluster*> (rn->data ()) -> value () ->
		      type_id ());
      }
      break;
    case SIL_SPECIALRATIO_TYPE:
      // NOTE: We do not simplify this
      break;
    case SIL_DIRECTIONANGLE_TYPE:
    case SIL_PLANARANGLEREGULAR_TYPE:
    case SIL_CONICALANGLEREGULAR_TYPE:
    case SIL_REGULARPOINTCIRCLE_TYPE:
      // Angular tolerance filter.
      rn = rs->first ();
      while (rn) {
	ListNode<Regularity> *rnn = rn->next ();
	Regularity *r = static_cast<Regularity*> (rn->data ());
	if (r->tolerance () > max_angle_rad) {
	  rn->unlink ();
	  A->removed_regularities () -> add_first (r);
	  delete rn;
	} else {
	  check_subregs (r);
	}
	rn = rnn;
      }
      break;
    case SIL_REGULARPOINTPOSITION_TYPE:
    case SIL_POINTLINE_TYPE:
      // Length tolerance filter.
      rn = rs->first ();
      while (rn) {
	ListNode<Regularity> *rnn = rn->next ();
	Regularity *r = static_cast<Regularity*> (rn->data ());
	if (r->tolerance () > max_length) {
	  rn->unlink ();
	  A->removed_regularities () -> add_first (r);
	  delete rn;
	} else {
	  check_subregs (r);
	}
	rn = rnn;
      }
      break;
    default:
      sil_msg_warn ("Simplify", "unknown regularity " + rn->data () -> 
		    type_id ());
      break;
    }
    List<Regularity> *rs_next = A->regularities () -> next ();
    if (rs->length () == 0) {
      A->regularities () -> del (rs);
      delete rs;
    }
    rs = rs_next;
  }

  sil_updated ();
  sil_progress_bar_percentage (1.0);
  sil_progress_bar (false);
  close ();

  return;
}

void simplify_symbol (SimpleList<Symbol const> *sl)
{
  Symbol const *s = sl->first ();
  while (s) {
    const char *t = sil_get_symbol_type (s);
    if (strcmp(t, "kernel") == 0) {
      Entity *e = static_cast<Entity*> (sil_get_symbol_data (s));
      if (e->is (SIL_ANALYSE_TYPE)) {
	SimplifyWindow *fw = manage (new SimplifyWindow 
				     (static_cast<Analyse*> (e),
				      static_cast<string const> 
				      (sil_get_symbol_name (s))));
	simplify_windows.add_first (fw);
      }
    }
    s = sl->next ();
  }
  return;
}

void update_handler ()
{
  ListNode<SimplifyWindow> *l = simplify_windows.first ();
  while (l) {
    ListNode<SimplifyWindow> *nn = l->next ();
    l->data () -> update ();
    l = nn;
  }
  return;
}

SIL_PLUGIN ()
{
  sil_register_tool (&simplify_symbol, "Simplify Regs", "0.1",
		     "Frank C Langbein", "2001 Frank C Langbein",
		     "Simplify regularities.", "kernel", "kernel");
  sil_register_update_handler (&update_handler);
  return;
}

SIL_CLOSE_PLUGIN ()
{
  sil_deregister_update_handler (&update_handler);
  ListNode<SimplifyWindow> *l = simplify_windows.first ();
  while (l) {
    ListNode<SimplifyWindow> *nn = l->next ();
    l->data () -> close ();
    l = nn;
  }
  return;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
