// renderer.cxx: Renderer class implementation.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: renderer.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (renderer, "$Id: renderer.cc 1 2005-08-10 23:35:07Z frank $");

#include "app.hxx"

#include "interface.hxx"

#include "list.hxx"
#include "renderer.hxx"
#include "xmath.hxx"
#include "modules.hxx"
#include "symbols.hxx"
#include "base/basetypes.hxx"
#include "base/base.hxx"
#include "base/base_module.hxx"

extern double render_box_min;
extern double render_box_max;
extern double render_clear_red;
extern double render_clear_green;
extern double render_clear_blue;

bool Renderer::gl_lights[8] = {
  false, false, false, false, false, false, false, false
};

inline void check_opengl (string const &error)
{
  GLenum glerror = glGetError ();
  if (glerror != GL_NO_ERROR)
    sil_msg_error ("OpenGL", error + ": " + static_cast<string const>
		   ((char const *)gluErrorString (glerror)));
  return;
}

void View::init ()
{
  c_objx = objx;
  c_objy = objy;
  c_objz = objz;
  c_eyeaz = eyeaz;
  c_eyeel = eyeel;
  c_ex = ex;
  c_ey = ey;
  c_ez = ez;
  // lookat and up are used for the camera position in the eye
  // movement mode. They depend on the elevation and the azimuth of
  // the camera direction. up will always point straight up for the
  // initialization.
  double azmrad = torad (eyeaz - 180.0);
  double azmsin = sin (azmrad);
  double azmcos = cos (azmrad);
  double elrad = torad (eyeel);
  double elcos = cos (elrad);
  double elmrad = torad (eyeel + 90.0);
  double elmcos = cos (elmrad);
  lookat[0] = azmsin * elcos;
  lookat[1] = sin (elrad);
  lookat[2] = azmcos * elcos;
  up[0] = azmsin * elmcos;
  up[1] = sin (elmrad);
  up[2] = azmcos * elmcos;
  return;
}

// OpenGL view initialization.
void View::show ()
{
  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();

  gluLookAt (c_ex, c_ey, c_ez,
	     c_ex + lookat[0], c_ey + lookat[1], c_ez + lookat[2],
	     up[0], up[1], up[2]);

  glCallList (2);

  glRotatef (c_objz, 0.0, 0.0, -1.0);
  glRotatef (c_objx, 1.0, 0.0, 0.0);
  glRotatef (c_objy, 0.0, 0.0, -1.0);

  check_opengl ("show view");

  return;
}

// Change view angle.
void View::change_eye_dir (int dx, int dy)
{
  static GLfloat m[3], dyy, dxx;
  ncrossprod (up, lookat, m);
  dyy = dy * movesens;
  dxx = dx * movesens;
  lookat[0] -= dxx * m[0] + dyy * up[0];
  lookat[1] -= dxx * m[1] + dyy * up[1];
  lookat[2] -= dxx * m[2] + dyy * up[2];
  normalize (lookat);
  up[0] += dyy * lookat[0];
  up[1] += dyy * lookat[1];
  up[2] += dyy * lookat[2];
  normalize (up); 
  return;
}

void View::change_obj_rot (int dx, int dy, int dz)
{
  c_objx += anglesens * dx;
  c_objy += anglesens * dy;
  c_objz += anglesens * dz;
  return;
}

void View::change_eye_pos (int dx, int dy)
{
  // In eye movement mode change the camera position along the camera
  // direction as indicated by the dy parameter...
  c_ex -= distsens * dy * lookat[0];
  c_ey -= distsens * dy * lookat[1];
  c_ez -= distsens * dy * lookat[2];
  // ...and twist the camera as indicated by the dx parameter.
  GLfloat m[3];
  double drad = torad (anglesens * dx);
  double dcos = cos (drad);
  double dsin = sin (drad);
  ncrossprod (lookat, up, m);
  up[0] = up[0] * dcos + m[0] * dsin;
  up[1] = up[1] * dcos + m[1] * dsin;
  up[2] = up[2] * dcos + m[2] * dsin;
  normalize (up);
  return;
}

void Projection::show (GLint width, GLint height)
{
  glViewport (0, 0, width, height);   

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();

  double asp = (double)width / (double)height;

  if (mode == proj_perspective)
    gluPerspective (angle, aspect * (GLdouble) asp, near, far);
  else
    glOrtho (left, right, bottom, top, near, far);

  check_opengl ("show projection");

  return;
}

Renderer::Renderer ()
{
  view = new View;
  proj = new Projection;

  sld = true;
  drawsubparts = false;
  localviewer = false;

  rb_x_min = rb_y_min = rb_z_min = render_box_min;
  rb_x_max = rb_y_max = rb_z_max = render_box_max;

  pointsize = 1.0;
  point_size_range[0] = 1.0;
  point_size_range[1] = 10.0;
  linewidth = 1.0;
  line_width_range[0] = 1.0;
  line_width_range[1] = 10.0;

  surface_tolerance = 0.0;
  normal_tolerance = 15.00;
  triang_mode = SIL_TRIANG_FRINGE_1;
  grid_mode = SIL_GRID_INTERIOR;
  max_lines = 0 ;
  min_lines = 0;
  max_line_length = 5.0;
  aspect_ratio = 0.0;
  adjust_mode = SIL_ADJUST_NON_GRID;
}

Renderer::~Renderer()
{
  if (view)
    delete view;
  if(proj)
    delete proj;
}

void Renderer::set_refinement (double surf_tol,
			       double normal_tol,
			       SiL_TriangMode t_mode,
			       SiL_GridMode g_mode,
			       int max_l,
			       int min_l,
			       double max_ll,
			       double ar,
			       SiL_AdjustMode am)
{
  surface_tolerance = surf_tol;
  normal_tolerance = normal_tol;
  triang_mode = t_mode;
  grid_mode = g_mode;
  max_lines = max_l;
  min_lines = min_l;
  max_line_length = max_ll;
  aspect_ratio = ar;
  adjust_mode = am;
  return;
}

void Renderer::refinement (double &surf_tol,
			   double &normal_tol,
			   SiL_TriangMode &t_mode,
			   SiL_GridMode &g_mode,
			   int &max_l,
			   int &min_l,
			   double &max_ll,
			   double &ar,
			   SiL_AdjustMode &am) const
{
  surf_tol = surface_tolerance;
  normal_tol = normal_tolerance;
  t_mode = triang_mode;
  g_mode = grid_mode;
  max_l = max_lines;
  min_l = min_lines;
  max_ll = max_line_length;
  ar = aspect_ratio;
  am = adjust_mode;
  return;
}

void Renderer::init ()
{
  // Get limits
  glGetFloatv (GL_POINT_SIZE_RANGE, point_size_range);
  glGetFloatv (GL_LINE_WIDTH_RANGE, line_width_range);

  // Lighting
  glEnable (GL_LIGHTING);
  static GLfloat ambientlight[4] = {0.2, 0.2, 0.2, 1.0 };
  glLightModelfv (GL_LIGHT_MODEL_AMBIENT, ambientlight);  
  if (localviewer)
    glLightModeli (GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
  else
    glLightModeli (GL_LIGHT_MODEL_LOCAL_VIEWER, GL_FALSE);
  glLightModeli (GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

  // Coloring
  glShadeModel (GL_SMOOTH);

  // Blending
  glEnable (GL_BLEND);
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  // Anitaliasing
  glEnable (GL_LINE_SMOOTH);
  glHint (GL_LINE_SMOOTH_HINT, GL_DONT_CARE);

  // Depth test
  glEnable (GL_DEPTH_TEST);
  glDepthRange (0.0, 1.0);
  glDepthFunc (GL_LEQUAL);

  // Renormalize normals.
  glEnable (GL_NORMALIZE);

  // Drawing
  glEnable (GL_LINE_STIPPLE);
  glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
  glFrontFace (GL_CCW);

  // Clearing
  glClearColor (render_clear_red, render_clear_green, render_clear_blue, 0.0);
  glClearDepth (1.0);
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Adding default light source
  void *ptr = App::instance () -> modules -> get_ptr ("base");
  if (ptr) {
    void (*p) (void);
    p = (void (*)(void))dlsym (ptr, "sil_base_create_default_light__Fv");
    const char *e = dlerror ();
    if (!e) {
      p ();
    } else {
      sil_msg_warn ("render", 
		    "base module incomaptible for creating default light sources");
    }
  } else {
    sil_msg_warn ("render", 
		  "can't create a default light source as the base module is not available");
  }

  check_opengl ("setting OpenGL paramters");
  
  view->init();

  return;
}

void Renderer::size (GLint width, GLint height)
{
  proj->show (width, height);
  return;
}

void Renderer::draw ()
{
  App::instance () -> modules -> draw_init ();
  glPointSize (pointsize);
  glLineWidth (linewidth);
  if (glIsList (1)) 
    glDeleteLists (1, 1);
  glNewList (1, GL_COMPILE);
  const Symbol *s = sil_first_symbol ();
  while (s) {
    if (sil_symbol_is_visible (s))
      s->draw ();
    s = sil_next_symbol ();
  }
  glEndList ();
  check_opengl ("creatig model list");
  if (glIsList (2)) 
    glDeleteLists (2, 1);
  glNewList (2, GL_COMPILE);
  for (int k = 0; k < 8; ++k)
    Renderer::gl_lights[k] = false;
  glDisable (GL_LIGHT0);
  glDisable (GL_LIGHT1);
  glDisable (GL_LIGHT2);
  glDisable (GL_LIGHT3);
  glDisable (GL_LIGHT4);
  glDisable (GL_LIGHT5);
  glDisable (GL_LIGHT6);
  glDisable (GL_LIGHT7);
  s = sil_first_symbol ();
  while (s) {
    if (sil_symbol_is_visible (s))
      s->draw_fixed ();
    s = sil_next_symbol ();
  }
  glEndList ();
  check_opengl ("creating light list");
  return;
}

void Renderer::show ()
{
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  view->show ();
  glCallList (1);
  glFlush ();
  check_opengl ("show error");
  return;
}

void Renderer::facet ()
{
  App::instance () -> modules -> facet_init ();
  const Symbol *s = sil_first_symbol ();
  while (s) {
    if (sil_symbol_is_visible (s))
      s->facet ();
    s = sil_next_symbol ();
  }
  return;
}

/// Local Variables:
/// exe-comm: "./sil"
/// compile-command: "make"
/// End:
