// renderer.hxx: Renderer class.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: renderer.hxx,v 1.24.2.1 2001/08/16 16:19:05 scmfcl Exp $

#ifndef RENDERER_HXX
#define RENDERER_HXX 1

#include "interface.hxx"

extern double sens_angle;
extern double sens_move;
extern double sens_dist;

class View {
public:
  inline View (double a = 0.0, double e = 0.0, 
	       double x = 0.0, double y = 0.0 , double z = 100.0,
	       double ox = 0.0, double oy = 0.0, double oz = 0.0) {
    objx = ox;
    objy = oy;
    objz = oz;
    eyeel = e;
    eyeaz = a;
    ex = x;
    ey =  y;
    ez = z;
    anglesens = sens_angle;
    movesens = sens_move;
    distsens = sens_dist;
    return;
  };
  void init ();
  void show ();
  void change_eye_dir (int dx, int dy);
  void change_eye_pos (int dx, int dy);
  void change_obj_rot (int dx, int dy, int dz);
  inline void set_angle_sens (double as) {
    anglesens = as;
    return;
  };
  inline void set_move_sens (double ms) {
    movesens = ms;
    return;
  };
  inline double angle_sens () const {
    return anglesens;
  };
  inline double move_sens () const {
    return movesens;
  };
  inline void set_dist_sens (double ds) {
    distsens = ds;
  };
  inline double dist_sens () const {
    return distsens;
  };
private:
  double eyeel;
  double eyeaz;
  double ex;
  double ey;
  double ez;
  double objx;
  double objy;
  double objz;
  double c_eyeel;
  double c_eyeaz;
  double c_ex;
  double c_ey;
  double c_ez;
  double c_objx;
  double c_objy;
  double c_objz;
  GLfloat lookat[3];
  GLfloat up[3];
  // Movement
  double anglesens;
  double movesens;
  double distsens;
  // Rotation distance on lookat
};

typedef enum {
  proj_perspective,
  proj_orthographic
} ProjectionMode;

class Projection {
public:
  inline Projection () {
    set (60.0, 1.0, -100.0, 100.0, -100.0, 100.0, 0.1, 1000.0, false);
  };
  inline void set (double an, double as,
		   double l, double r, double b, double t, double n,
		   double f, bool ortho) {
    if (ortho)
      mode = proj_orthographic;
    else
      mode = proj_perspective;
    angle = an;
    aspect = as;
    left = l;
    right = r;
    bottom = b;
    top = t;
    near = n;
    far = f;
    return;
  };
  void show (GLint width, GLint height);
  inline void get (double &an, double &as, double &l, double &r,
		   double &b, double &t,
		   double &n, double &f, bool &ortho) const {
    if (mode == proj_orthographic)
      ortho = true;
    else
      ortho = false;
    an = angle;
    as = aspect;
    l = left;
    r = right;
    t = top;
    b = bottom;
    n = near;
    f = far;
    return;
  };
private:
  ProjectionMode mode;
  double near;
  double far;
  double angle;
  double aspect;
  double left;
  double right;
  double top;
  double bottom;
};

class Renderer {
public:
  Renderer ();
  ~Renderer ();
  void init ();
  void size (GLint width, GLint height);
  void facet ();
  void draw ();
  void show ();
  inline void set_solid (bool b) { 
    sld = b;
    return; 
  };
  inline bool solid () const { 
    return sld;
  };
  void set_local_viewer (bool b) {
    localviewer = b;
    return;
  };
  inline bool local_viewer () const { 
    return localviewer;
  };
  inline void set_draw_sub_parts (bool f) {
    drawsubparts = f;
    return;
  };
  inline bool draw_sub_parts () const {
    return drawsubparts;
  };
  void set_refinement (double surf_tol,
		       double normal_tol,
		       SiL_TriangMode t_mode,
		       SiL_GridMode g_mode,
		       int max_l,
		       int min_l,
		       double max_ll,
		       double ar,
		       SiL_AdjustMode am);
  void refinement (double &surf_tol,
		   double &normal_tol,
		   SiL_TriangMode &t_mode,
		   SiL_GridMode &g_mode,
		   int &max_l,
		   int &min_l,
		   double &max_ll,
		   double &ar,
		   SiL_AdjustMode &am) const;
  inline void init_view () {
    view->init ();
    return;
  };
  inline void change_eye_dir (int dx, int dy) {
    view->change_eye_dir (dx, dy);
    return;
  };
  inline void change_eye_pos (int dx, int dy) {
    view->change_eye_pos (dx, dy);
    return;
  };
  inline void change_obj_rot (int dx, int dy, int dz) {
    view->change_obj_rot (dx, dy, dz);
    return;
  };
  inline void set_projection (double an, double as,
			      double l, double r, double b, double t, double n,
			      double f, bool o) {
    proj->set (an, as, l, r, b, t, n, f, o);
    return;
  };
  inline void projection (double &an, double &as, double &l, double &r,
			  double &b, double &t,
			  double &n, double &f, bool &ortho) const {
    proj->get (an, as, l, r, b, t, n, f, ortho);
    return;
  };
  inline double point_size () const {
    return pointsize;
  };
  inline double max_point_size () const {
    return point_size_range[1];
  };
  inline double min_point_size () const {
    return point_size_range[0];
  };
  inline double line_width () const {
    return linewidth;
  };
  inline double max_line_width () const {
    return line_width_range[1];
  };
  inline double min_line_width () const {
    return line_width_range[0];
  };
  inline void set_point_size (double s) {
    if (s > (double)point_size_range[1])
      pointsize = (double)point_size_range[1];
    else if (s < (double)point_size_range[0])
      pointsize = (double)point_size_range[0];
    else
      pointsize = s;
    return;
  };
  inline void set_line_width (double s) {
    if (s > (double)line_width_range[1])
      linewidth = (double)line_width_range[1];
    else if (s < (double)line_width_range[0])
      linewidth = (double)line_width_range[0];
    else
      linewidth = s;
    return;
  };
  inline void set_render_box (double a, double b, double c, 
			      double d, double e, double f) {
    rb_x_min = a;
    rb_x_max = b;
    rb_y_min = c;
    rb_y_max = d;
    rb_z_min = e;
    rb_z_max = f;
    return;
  };
  inline void render_box (double &a, double &b, double &c, 
			  double &d, double &e, double &f) const {
    a = rb_x_min;
    b = rb_x_max;
    c = rb_y_min;
    d = rb_y_max;
    e = rb_z_min;
    f = rb_z_max;
    return;
  };
  static bool gl_lights[8];
private:
  // Render parameters
  bool drawsubparts;
  bool sld;
  // Render box
  double rb_x_min, rb_x_max;
  double rb_y_min, rb_y_max;
  double rb_z_min, rb_z_max;
  // OpenGL parameters
  double pointsize, linewidth;
  GLfloat point_size_range[2], line_width_range[2];
  bool localviewer;
  // OpenGL Display parameters
  View *view;
  Projection *proj;
  // refinement
  double surface_tolerance;
  double normal_tolerance;
  SiL_TriangMode triang_mode;
  SiL_GridMode grid_mode;
  int max_lines;
  int min_lines;
  double max_line_length;
  double aspect_ratio;
  SiL_AdjustMode adjust_mode;
};

#endif // !RENDERER_HXX

/// Local Variables:
/// exe-comm: "./sil"
/// compile-command: "make"
/// End:
