// points.hxx: Points class.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: points.hxx,v 1.19.2.1 2001/08/16 16:19:17 scmfcl Exp $

#ifndef POINTS_HXX
#define POINTS_HXX 1

#include "material.hxx"

class RangePoints;

// General Scanner type class.
class Scanner {
public:
  Scanner () { 
    mat = NULL;
  };
  virtual ~Scanner () { 
    if (mat)
      delete mat;
  };
  virtual string const &type () const {
    return tid;
  };
  bool is (string const &s) const {
    return (this->type () == s);
  };
  Material *material () const {
    return mat;
  };
  void set_material (Material *m) {
    if (mat)
      delete mat;
    mat = m;
    return;
  };
private:
  static string const tid;
  Material *mat;
};

// The RangePoint class represents a single point in the point cloud from the
// scanner.
class RangePoint : public Scanner {
public:
  RangePoint () : Scanner () {
    vx = vy = vz = 0.0;
    nxt = NULL;
  };
  ~RangePoint () { 
  };
  RangePoint (double xx, double yy, double zz) : Scanner () {
    vx = xx;
    vy = yy;
    vz = zz;
    nxt = NULL;
  };
  string const &type () const {
    return tid;
  };
  void set (double xx, double yy, double zz) {
    vx = xx;
    vy = yy;
    vz = zz;
    return;
  };
  void set_normal (double xx, double yy, double zz) {
    vnx = xx;
    vny = yy;
    vnz = zz;
    return;
  };
  void get (double &xx, double &yy, double &zz) {
    xx = vx;
    yy = vy;
    zz = vz;
    return;
  };
  double get (short c) {
    if (c == 0)
      return vx;
    if (c == 1)
	return vy;
    if (c == 2)
      return vz;
    if (c == 3)
      return vnx;
    if (c == 4)
      return vny;
    return vnz;
  }
  double x () const {
    return vx;
  };
  double &x () {
    return vx;
  };
  double y () const {
    return vy;
  };
  double &y () {
    return vy;
  };
  double z () const {
    return vz;
  };
  double &z () {
    return vz;
  };
  double nx () const {
    return vnx;
  }
  double ny () const {
    return vny;
  }
  double nz () const {
    return vnz;
  }
protected:
  RangePoint (double xx, double yy, double zz, RangePoint *n) : Scanner () {
    vx = xx;
    vy = yy;
    vz = zz;
    nxt = n;
  };
  RangePoint *next () const {
    return nxt;
  };
  RangePoint *&next () {
    return nxt;
  };
  friend class RangePoints;
  friend void sil_cluster (RangePoints *pts, double dist);
private:
  static string const tid;
  double vx, vy, vz, vnx, vny, vnz;
  RangePoint *nxt;
};

// The RangePoints class represents point clouds from the scanner.
class RangePoints : public Scanner {
public:
  RangePoints ();
  ~RangePoints ();
  string const &type () const {
    return tid;
  }
  void add (double x, double y, double z);
  void add (RangePoint *p);
  void add_check (double x, double y, double z);
  bool del (RangePoint *p);
  void unite (RangePoints *rp);
  unsigned long num_points () const {
    return count;
  }
  RangePoint *first ();
  RangePoint *next ();
  bool first (double &x, double &y, double &z);
  bool next (double &x, double &y, double &z); 
  RangePoint *replace_points (RangePoint *p = NULL);
  void set_translation (double x, double y, double z) {
    tx = x;
    ty = y;
    tz = z;
    return;
  }
  void translation (double &x, double &y, double &z) const {
    x = tx;
    y = ty;
    z = tz;
    return;
  }
  void set_rotation (double x, double y, double z) {
    rp = x;
    rt = y;
    rs = z;
    return;
  }
  void rotation (double &x, double &y, double &z) const {
    x = rp;
    y = rt;
    z = rs;
    return;
  }
  void apply_transformation (bool normals = false);
  void apply_transformation (double m1, double m2, double m3,
			     double m4, double m5, double m6,
			     double m7, double m8, double m9,
			     double t1, double t2, double t3,
			     bool normals = false);
  void centre ();
  bool save (string const &name);
private:
  static string const tid;
  RangePoint *head;
  RangePoint *window;
  unsigned long count;
  double tx, ty, tz, rp, rt, rs;
};

// The RangeMesh is used for the reconstruction process and represents a
// triangulates point cloud.
class RangeMesh : public Scanner {
public:
  RangeMesh (RangePoints *rp = NULL, unsigned long nt = 0);
  RangeMesh (unsigned long np, unsigned long nt);
  ~RangeMesh ();
  string const &type () const {
    return tid;
  }
  unsigned long num_points () const {
    return npts;
  }
  double *x () const {
    return x_pts;
  }
  double *y () const {
    return y_pts;
  }
  double *z () const {
    return z_pts;
  }
  unsigned long num_triangles () const {
    return ntrgs;
  }
  unsigned long *triangle_v1 () const {
    return trg_v1;
  }
  unsigned long *triangle_v2 () const {
    return trg_v2;
  }
  unsigned long *triangle_v3 () const {
    return trg_v3;
  }
  void init_ids ();
  unsigned short *ids () const {
    return id;
  }
  unsigned short &max_id () {
    return m_id;
  }
  unsigned short max_id () const {
    return m_id;
  }
  void init_normals ();
  double *normal_x () const {
    return n_x;
  }
  double *normal_y () const {
    return n_y;
  }
  double *normal_z () const {
    return n_z;
  }
  void init_project ();
  double *project_x () const {
    return px_pts;
  }
  double *project_y () const {
    return py_pts;
  }
  double *project_z () const {
    return pz_pts;
  }
  void init_neighbours ();
  unsigned long *neighbour1 () const {
    return nb1;
  }
  unsigned long *neighbour2 () const {
    return nb2;
  }
  unsigned long *neighbour3 () const {
    return nb3;
  }
  SimpleList<string> &region () {
    return reg;
  }
  SimpleList<string> &classify () {
    return clas;
  }
private:
  unsigned long find_f_triangle (unsigned long v,
				 SimpleList<unsigned long> *vtrg1,
				 SimpleList<unsigned long> *vtrg2,
				 SimpleList<unsigned long> *vtrg3,
				 unsigned long l, unsigned long &n) const;
  unsigned long find_b_triangle (unsigned long v,
				 SimpleList<unsigned long> *vtrg1,
				 SimpleList<unsigned long> *vtrg2,
				 SimpleList<unsigned long> *vtrg3,
				 unsigned long l, unsigned long &n) const;
  static string const tid;
  unsigned long npts, ntrgs;
  double *x_pts, *y_pts, *z_pts, *px_pts, *py_pts, *pz_pts, *n_x, *n_y, *n_z;
  unsigned short *id, m_id;
  unsigned long *trg_v1, *trg_v2, *trg_v3, *nb1, *nb2, *nb3;
  SimpleList<string> reg, clas;
};

#endif // !POINTS_HXX

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
